# $Id: lib.py,v 1.23 2006-03-03 13:42:02 devel Exp $
# Utility functions.
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

import os, sys, stat, string, fnmatch

#-------------------------------------------------------------------------------
# Exceptions.
#-------------------------------------------------------------------------------

import traceback

# Oops is for being thrown as an exception.
# (This would've been named "Exception" but Python predefines it.)
class Oops:
    
    def __init__( s, text="" ):
        s.text = text[:]
        
    def __repr__( s ):
        return s.text

def PrintException():
    """ Print call stack when an exception occurs. """
    # Put PrintException() after an "except:" clause.
    traceback.print_exc( file=sys.stderr )

#-------------------------------------------------------------------------------
# Basic functions.
#-------------------------------------------------------------------------------

def Unique( list ):
    """ Remove duplicates out of a list. """
    # Build a dictionary as duplicate keys will overwrite each other.
    d = { }
    for x in list:
        d[str(x)] = 1
    return d.keys()

def List2String( list, separator="" ):
    s = ""
    for x in list:
        s = s + x + separator
    return s

def ListDeleteIf( list, Callback ):
    """ Remove elements from a list that match a criteria. """
    # Callback is passed every element which should return True
    # if the element should be deleted.
    
    mfd = []   # marked for deletion
    
    i = 0
    for elem in list:
        if Callback( elem ):
            mfd.append( i )
        i += 1

    # This is a little tricky: when one element is deleted,
    # all the remaining indexes become 1 too great and so on.
    i = 0
    deletions = 0
    for idx in mfd:
        del list[idx-deletions]
        deletions += 1

def DictDeleteIf( dict, Callback ):
    """ Remove key/value pairs from a dictionary that match a criteria. """
    # Callback is passed every key which should return True
    # if the key/value pair should be deleted.
    
    mfd = []   # marked for deletion
    
    for key in dict.keys():
        if Callback( key ):
            mfd.append( key )

    for key in mfd:
        del dict[key]

#-------------------------------------------------------------------------------
# String functions.
#-------------------------------------------------------------------------------

def IfEol( c ):
    """ Return True if end-of-line. """
    return c == "\n"

def Chomp( s ):
    """ Discard last char if it's newline """
    if s[-1] == "\n":
        s = s[0:-1]
    return s

def IfQuoted( text ):
    """ Return true if string is quoted. """
    return len(text) >= 2 and text[0] in "\"'" and text[-1] in "\"'"

def Unquote( text ):
    """ Removes the 2 surrounding quote chars out of a string. """
    # PRECOND: String is quoted (no checks).
    return text[1:-1]    # chop first and last char

def UnquoteWords( wordList ):
    """ Remove surrounding quote chars in a list of strings. """
    i = 0
    for word in wordList:
        if IfQuoted( word ):
            wordList[i] = Unquote( word )
        i += 1
    return wordList

def SplitWords( wordStr, keepOuterQuotes=True ):
    """ Split a string into a list but does not split quoted words. """
    wordList = []   # result
    word = ""       # accumulates a word char-by-char
    quote = None    # holds { "'", '"', None }
    whitespace = " \t\n"
    i = 0
    while i < len(wordStr):
        c = wordStr[i]
        # Escaped char?
        if c == "\\":
            word += c
            if i+1 < len(wordStr):
                word += wordStr[i+1]
                i += 1
        # Quote char?
        elif c == '"' or c == "'":
            # Beginning quote char?
            if not quote:
                if keepOuterQuotes:
                    word += c
                # Remember which kind of quote char to determine final quote char.
                quote = c
                # A pair of the same quote char?
                if i+1 < len(wordStr) and wordStr[i+1] == wordStr[i]:
                    word += c
                    i += 1
            else:
                # A pair of the same quote char?
                if i+1 < len(wordStr) and wordStr[i+1] == wordStr[i]:
                    if keepOuterQuotes:
                        word += c
                    word += c
                    if c == quote:   # to handle '""Hello World""'
                        quote = None
                    i += 1
                # Ending quote char?
                elif c == quote:
                    if keepOuterQuotes:
                        word += c
                    quote = None
                # Different kind of quote char than beginning one.
                else:
                    word += c
        # Quoting?
        elif quote:
            # Append this char before the space case below.
            word += c
        # Whitespace?
        elif c in whitespace:
            # Append accumulated word (unless empty) to list.
            if word:
                wordList.append( word )
                word = ""
        else:
            word += c
        # Continue to next char.
        i += 1

    # Append final word (unless empty) to list.
    if word:
        wordList.append( word )
        
    return wordList

def IfHaveEachChar( set1, set2 ):
    """ Return true if each char in set1 exists in set2. """
    for c1 in set1:
        if not c1 in set2:
            return False
    return True

def AddIntoParentheses( parentheses, addition ):
    """ Helper to add comma-separated sub-strings inside a parenthesized string. """
    # Eg: "(abc)" can become "(abc,def)"
    if parentheses == "":
        parentheses = "(" + addition + ")"
    else:
        parentheses = parentheses[:-1] + "," + addition + ")"
    return parentheses

#-------------------------------------------------------------------------------
# File I/O.
#-------------------------------------------------------------------------------

def ReadLine( fd ):
    """ Generator function to read lines from a file descriptor.
        Returns the line excluding the newline char. """
    remainder = ""
    while True:
        piece = os.read( fd, 16384 )
        # Continue with remainder (that didn't form a complete line).
        piece = remainder + piece
        if not piece:
            return
        remainder = "" # consumed
        # Is the last line a complete line?
        if piece[-1] != "\n":
            # Is the incomplete line the only line?
            # If so, wait for more input.
            if piece.find( "\n" ) == -1:
                remainder = piece
                continue
            else:
                # Extract out the incomplete last line into "remainder".
                i = piece.rfind( "\n" )
                try:
                    remainder = piece[i+1:]
                except:
                    remainder = ""
                piece = piece[:i]
        # Now yield all of the complete lines.
        lineList = piece.split( "\n" )
        if not lineList[-1]:
            del lineList[-1]    # split() will make the last element an empty line
        for line in lineList:
            yield line[:-1]    # minus newline char
    return

def WriteBuffer( fd, buf ):
    """ Write every byte of a buffer to a file descriptor.
        This is necessary for pipes, sockets. """
    buf2 = buf[:]
    while len(buf2):
        n = os.write( fd, buf2 )
        buf2 = buf2[n:]

def IsExecutableFile( fname ):
    """ True if is a file and is executable (returns false for directories). """
    status = os.stat( fname )
    mode = status[0]
    if (mode & stat.S_IEXEC) and not (mode & stat.S_IDIR):
        return True
    return False

def FileModificationTime( pname ):
    """ Get a file's modification time.  Exception if file doesn't exist. """
    t = os.stat( pname )
    return t[stat.ST_MTIME]

#-------------------------------------------------------------------------------
# Pathnames.
#-------------------------------------------------------------------------------

def FileExt( fname ):
    """ Return a file's extension/suffix (including '.') or None. """
    try:
        i = fname.rindex( "." )
        return fname[i:]
    except:
        return None

def PrettyDirName( dirname ):
    """ Return dirname with a trailing slash. """
    try:
        if dirname[-1] != "/":
            dirname += "/"
        else:
            # In case another trailing slash was blindly appended.
            if dirname[-2:] == "//":
                dirname = dirname[:-1]
    except:
        PrintException()
    return dirname

def AbsPath( fname, cwd=None ):
    """ Return an absolute path. """
    if not cwd:
        cwd = os.getcwd()
    if IfCygwin():
        # Change to cygwin naming convention: /cygdrive/c/cwd/fname
        if len(fname) > 2 and fname[1] == ":":   # DOS drive letter?
            return AbsPathDos2Cygwin( fname )
        else:
            fname = fname.replace( "\\", "/" )
            return os.path.join( cwd, fname )
    else:
        if fname[0] == "/":
            return fname
        else:
            return os.path.join( cwd, fname )

def AbsPathDos2Cygwin( path ):
    path = "/cygdrive/" + path[0] + "/" + path[2:]
    path = path.replace( "\\", "/" )
    return path.replace( "//", "/" )

def TransformPath( path ):
    """ Cope with differences in OS file names. """
    if not IfCygwin():
        return path
    else:
        # cygwin
        savedPath = path
        try:
            if path.find( "/cygdrive" ) == 0:
                # /cygdrive/c --> c:\
                driveLetter = path[10]
                path = driveLetter + ":" + path[11:]
                path = path.replace( "/", "\\" )    # to MSDOG dir separator
            return path
        except:
            return savedPath

def TransformPathList( pathList ):
    """ Cope with differences in OS file names. """
    i = 0
    while i < len(pathList):
        pathList[i] = TransformPath( pathList[i] )
        i += 1

def Basename( fname ):
    # os.path.basename doesn't work right on cygwin.
    try:
        i1 = fname.rfind( "/" )
        i2 = fname.rfind( "\\" )
        if i1 == -1 and i2 == -1:
            return fname
        if i1 > i2:
            return fname[i1+1:]
        else:
            return fname[i2+1:]
    except:
        return fname

def MatchFileglob( fname, glob, caseSens=False ):
    """ Return true if fname matches glob pattern. """
    # This wrapper exists because the behavior of fnmatch.fnmatch()
    # is inconsistent across platforms (it may or may not be case-sensitive).
    # fnmatchcase() is supposed to be always case-sensitive.
    if not caseSens:
        fname = fname.lower()
        glob  = glob.lower()
    return fnmatch.fnmatchcase( fname, glob )    # fnmatchcase() is always case-sensitive

#-------------------------------------------------------------------------------
# Process execution.
#-------------------------------------------------------------------------------

def RunCmd( fname ):
    """ Run a command, return its stdout as a string. """
    pipe = os.popen( fname + " 2>&1" )
    out = ""
    while 1:
        line = pipe.readline()
        if not line:
            break
        out = out + line
    pipe.close()
    return out

def RunProgram( fname, child, argTuple ):
    """ Run a program. """
    # If child is false, the Python process is replaced (exec).
    if IfCygwin():
        os.spawnv( os.P_NOWAIT, fname, argTuple )
        return
    if child:
        pid = os.fork()
        if pid != 0:
            # This is the parent.
            return
    try:
        if fname[0] == '/':
            os.execv( fname, argTuple )  # full pathname
        else:
            os.execvp( fname, argTuple ) # filename via PATH
    except:
        PrintWarning( "RunProgram(" + fname + ") failed" )
    # exec() failed.
    if child:
        # Stop child process.
        os._exit(0)  # not exit(0)

#-------------------------------------------------------------------------------
# Operating-system specific.
#-------------------------------------------------------------------------------

# Be careful of ordering calls to these when determining which OS.
# IfUnix() is very generic and somewhat of a catch-all.

def IfCygwin():
    return sys.platform.find( "cygwin" ) != -1

# Mac OS X
def IfDarwin():
    return sys.platform.find( "darwin" ) != -1

def IfUnix():
    # python/lib/module-os.html:
    # The following names have currently been registered:
    # 'posix', 'nt', 'mac', 'os2', 'ce', 'java', 'riscos'
    return os.name == "posix"
        
#-------------------------------------------------------------------------------
# Finder class.
#-------------------------------------------------------------------------------

import re

class OopsBadRegex(Oops):
    def __init__( s, text="" ):
        Oops.__init__( s, text )

class Finder:
    """ Class to facilitate text searching using various matching methods. """

    def __init__( s, pattern, caseSens=False, exact=False, regex=False, glob=False, negative=False ):

        # Construct according to search method.
        # - method: regex
        # - method: file glob matching
        # - method: basic string matching
        # - method: exact
        # - case-sensitive (all methods)
        # - negative: IfMatch() returns True not pattern does not match
        s.patternOrig = pattern[:]      # copy
        s.pattern     = pattern
        s.caseSens    = caseSens        # true if case-sensitive
        s.regex       = regex
        s.exact       = exact
        s.glob        = glob
        s.negative    = negative
        s.basic       = not (exact or regex or glob)

        # The user is supposed to surround a multi-word pattern with quotes.
        # The surrounding quotes should be discarded.  Which leads to the question:
        # How can the user include a quote char in the pattern?
        # What if the pattern is '"Hello World"'?  The question answers itself:
        # by surrounding the pattern with two pairs of quotes at both ends.
        if IfQuoted( s.pattern ):
            s.pattern = Unquote( s.pattern )    # removes the 2 outermost quotes only

        # Compile regex.
        try:
            if s.regex:
                if s.caseSens:
                    s.regexObj = re.compile( s.pattern )
                else:
                    s.regexObj = re.compile( s.pattern, re.IGNORECASE )
        except:
            # This can happen if the user enters an unquoted multi-word regex
            # so that only the first part of the regex reaches here.
            raise OopsBadRegex

    def IfMatch( s, haystack ):
        if not s.negative:
            return s.__IfMatch( haystack )
        else:
            return not s.__IfMatch( haystack )
            
    def __IfMatch( s, haystack ):
        """ Return true if passed haystack matches pattern (given to constructor). """
        # The match methods have varying precedence or are mutually-exclusive.
        if s.basic:
            return Find( haystack, s.pattern, s.caseSens ) != -1
        elif s.glob:
            return MatchFileglob( haystack, s.pattern, caseSens=s.caseSens )
        elif s.regex:
            return s.regexObj.search( haystack )
        elif s.exact:
            return s.pattern == haystack
        else:
            raise Oops( "Finder.IfMatch: unhandled search method" )

    def GetMatchingLines( s, haystack, fnameHaystack="" ):
        """ Return a list of line numbers that match pattern. """
        # fnameHaystack is informational (shown in error message, no file is read here).
        if s.negative:
            raise Oops( "unhandled case" )
        if s.basic:
            # This is a common case written in C++ for speed.
            return FindLines( haystack, s.pattern, s.caseSens )
        else:
            # Search thru file before searching line-by-line (usually saves time).
            # Rare case: if a data file has a src file ext, Python will raise TypeError
            # while parsing the args for Find() (written in C++ at Utils.cc).
            matchingLines = []
            try:
                somewhere = s.IfMatch( haystack )
            except:
                somewhere = False
                if haystack.find( "\0" ):
                    PrintWarning( "Finder: found null char in " + fnameHaystack )
                else:
                    raise
            if somewhere:
                # Search line-by-line.
                lineNum = 1
                lineList = haystack.split( "\n" )
                for line in lineList:    # line is a string but lineNum is an int
                    if s.IfMatch( line ):
                        matchingLines.append( lineNum )
                    lineNum += 1
            return matchingLines            

    def GetPattern( s ):
        """ Returns specified pattern. """
        return s.patternOrig    # not s.pattern which could be modified internally

    def IfGlob( s ):
        return s.glob

    def IfRegex( s ):
        return s.regex

    def IfCaseSens( s ):
        return s.caseSens

    def GetModesString( s ):
        """ Returns finder's modes as an informational/printable string. """
        modes = ""
        if s.caseSens:
            modes = AddIntoParentheses( modes, "case-sensitive" )
        if s.regex:
            modes = AddIntoParentheses( modes, "regex" )
        if s.glob:
            modes = AddIntoParentheses( modes, "glob" )
        if s.exact:
            modes = AddIntoParentheses( modes, "exact" )
        if s.basic:
            modes = AddIntoParentheses( modes, "basic" )
        return modes

#-------------------------------------------------------------------------------
# Nest class.
#-------------------------------------------------------------------------------

class Nest:
    # Helper class to nest complementary actions.
    # IfProceed() returns True when the depth indicates an operation should be executed.
    # action1 has no meaning here, its complement is action2 (actions have meaning to caller).
    # action1=True increments depth, action1=False decrements depth.
    # count(post):  action1:   action2:   proceed:  outer-level:
    # +1            True       False      True      True
    # +2            True       False      False     False (nested)
    # +1            False      True       False     False (nested)
    # 0             False      True       True      True

    def __init__( s, nestFirst=True ):
        # nestFirst tells IfProceed() to bypass changing depth
        # the first time it is called.  Eg, the first time a GUI is enabled
        # should be ignored, thereafter, enable/disable GUI operations will nest.
        s.depth     = 0
        s.first     = True
        s.nestFirst = nestFirst

    def IfProceed( s, action1 ):

        # First time called?
        if s.first:
            s.first = False
            # Should the first call affect depth?
            if not s.nestFirst:
                return True     # leave depth alone, tell caller to proceed
        
        if not action1:
            s.depth -= 1
            if s.depth != -1:
                return False    # at inner level
        else:
            s.depth += 1
            if s.depth != 0:
                return False    # at inner level
            
        return True             # at outer level
# $Id: Import.py,v 1.4 2006-02-28 03:56:41 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

import copy, sys, os, os.path, stat, time, re
import traceback
import thread, threading
from string import *
# $Id: Utils.py,v 1.1 2006-02-25 03:56:48 devel Exp $
# Python functions implemented in C++.
# Author:   Jim Brooks  http://www.jimbrooks.org
# License:  GNU General Public License (GPL)
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2006
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Utility functions implemented in C++.
#-------------------------------------------------------------------------------

import pycpp_utils

def Find( haystack, needle, caseSens=False ):
    """ Case-insensitive string search (implemented in C++). """
    # Similar to string.find().
    # If haystack contains needle, returns index where needle is,
    # else returns -1 as no match.
    return pycpp_utils.Utils_Find( haystack, needle, caseSens )

def FindLines( haystack, needle, caseSens=False ):
    """ Returns a list of matching line numbers (implemented in C++). """
    return pycpp_utils.Utils_FindLines( haystack, needle, caseSens )
# $Id: DequeChar.py,v 1.6 2006-02-25 03:56:48 devel Exp $
# Python DequeChar class built over underlying C++ STL deque<char>.
# Author:   Jim Brooks  http://www.jimbrooks.org
# License:  GNU General Public License (GPL)
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Class DequeChar.
#-------------------------------------------------------------------------------

import pycpp_dequechar

class DequeChar:
    """ DequeChar is based on C++ STL deque<char> """

    def __init__( s, chars="" ):
        pycpp_dequechar.DequeChar_New( s )
        pycpp_dequechar.DequeChar_Set( s, chars )

    def __del__( s ):
        # An exception will normally occur during program exit
        # because modules including pycpp_dequechar are gone.
        try:
            pycpp_dequechar.DequeChar_Delete( s )
        except:
            pass

    def PopHead( s, n=1 ):
        """ Remove n chars as a string from head.
            Raises exception if not all n chars are available. """
        return pycpp_dequechar.DequeChar_PopHead( s, n )

    def PopTail( s, n=1 ):
        """ Remove n chars as a string from tail.
            Raises exception if not all n chars are available. """
        return pycpp_dequechar.DequeChar_PopTail( s, n )
        
    def PeekHead( s, n=1 ):
        """ Peek n chars as a string at head without removing them.
            Raises exception if not all n chars are available. """
        return pycpp_dequechar.DequeChar_PeekHead( s, n )

    def PeekTail( s, n=1 ):
        """ Peek n chars as a string at tail without removing them.
            Raises exception if not all n chars are available. """
        return pycpp_dequechar.DequeChar_PeekTail( s, n )

    def PushHead( s, chars ):
        """ Enqueue a string at head. """
        pycpp_dequechar.DequeChar_PushHead( s, chars )

    def PushTail( s, chars ):
        """ Enqueue a string at tail. """
        pycpp_dequechar.DequeChar_PushTail( s, chars )

    def IfEmpty( s ):
        """ True if deque is empty. """
        return pycpp_dequechar.DequeChar_IfEmpty( s )

    def Size( s ):
        """ Return amount of elements in deque. """
        return pycpp_dequechar.DequeChar_Size( s )

    def Set( s, ibuf ):
        """ Set the contents of deque from a string.
            To clear the deque, pass an empty string. """
        pycpp_dequechar.DequeChar_Set( s, ibuf )

    def Get( s ):
        """ Return contents of deque as a string. """
        return pycpp_dequechar.DequeChar_Get( s )

    def __setitem__( s, idx, c ):
        """ Assign to a char by index in deque. """
        # A negative index within range is allowed.
        pycpp_dequechar.DequeChar_SetItem( s, idx, c )

    def __getitem__( s, idx ):
        """ Return a char by index from deque. """
        # A negative index within range is allowed.
        return pycpp_dequechar.DequeChar_GetItem( s, idx )

    def Index( s, sought, idxStart=0 ):
        """ Return index of substring in deque starting at an index. """
        # Raises exception if no match, sought string is empty,
        # or index plus sought string is past end of deque.
        return pycpp_dequechar.DequeChar_Index( s, sought, idxStart )
# $Id: GreedyDict.py,v 1.5 2006-02-25 03:56:48 devel Exp $
# Python GreedyDict built over underlying C++ GreedMap template class.
# Author:   Jim Brooks  http://www.jimbrooks.org
# License:  GNU General Public License (GPL)
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Class GreedyDict.
#-------------------------------------------------------------------------------

import pycpp_greedydict

class GreedyDict:
    """ Dictionary that greedily finds the longest matching key. """

    def __init__( s, hintKeySize=15 ):
        pycpp_greedydict.GreedyDict_New( s, hintKeySize )
        return

    # INTERFACE
    def __del__( s ):
        # An exception will normally occur during program exit
        # because modules including pycpp_greeydict are gone.
        try:
            pycpp_greedydict.GreedyDict_Delete( s )
        except:
            pass
        return

    # INTERFACE
    def __setitem__( s, key, val ):
        """ Insert a key,value into the dictionary. """
        pycpp_greedydict.GreedyDict_Insert( s, key, val )

    # INTERFACE
    def __getitem__( s, key ):
        """ Return a value from the dictionary. """
        # This match do an approximate match (see GreedyMap.hh).
        (key2,val) = pycpp_greedydict.GreedyDict_Find( s, key )
        return val

    # INTERFACE
    def Find( s, key ):
        """ Return (key,value) from the dictionary.
            The returned key is the actual matching key which
            may or may not exactly be the same as the passed key. """
        return pycpp_greedydict.GreedyDict_Find( s, key )
# $Id: misc.py,v 1.27 2006-03-05 15:56:57 devel Exp $
# Misc functions for pypersrc.
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Functions for source files, programming languages.
#-------------------------------------------------------------------------------

# Programming languages recognized by the  -l arg.
progLangList = [ "c", "c++", "python", "java", "perl", "forth" ]

progLangExtDict = { ".c"    : "c",
                    ".h"    : "c",
                    ".cc"   : "c++",
                    ".C"    : "c++",
                    ".cpp"  : "c++",
                    ".cxx"  : "c++",
                    ".c++"  : "c++",
                    ".H"    : "c++",
                    ".hh"   : "c++",
                    ".hxx"  : "c++",
                    ".h++"  : "c++",
                    ".inl"  : "c++",
                    ".java" : "java",
                    ".jav"  : "java",
                    ".j"    : "java",
                    ".py"   : "python",
                    ".pyw"  : "python",
                    ".pl"   : "perl",
                    ".pm"   : "perl",
                    ".ph"   : "perl",
                    ".4th"  : "forth",
                    ".fth"  : "forth",
                    ".seq"  : "forth",
                    ".fs"   : "forth", # GForth
                    }

def WhichKindOfSrcFile( pname ):
    """ Return which programming language a source file is for.
        Eg, returns 'c' for main.c.  Returns empty string if unknown. """
    (root,ext) = os.path.splitext( pname )
    if ext and progLangExtDict.has_key( ext ):
        return progLangExtDict[ext]
    else:
        return ""

def ReadSrcFile( fname ):
    """ Return string containing a source file or raise IOError exception. """
    f = open( fname, "r" )    # may raise exception
    buf = f.read()
    buf = buf.replace( "\r", "" )    # remove MSDOS CR chars
    f.close()
    return buf

def SrcFile2TokenList( fname ):
    """ Compute a token list from a source file. """
    # Read the source code file.
    # Lex file to build a token list.
    # Parser will build a second token list from the first.
    # Then load the text widget from the second.
    try:
        text = ReadSrcFile( fname )
    except IOError:
        PrintWarning( "SrcFile2TokenList: source file " + fname + " can't be read, continuing" )
        return ([],False)
    text = text.expandtabs( st.tabWidth )
    tokenList = st.lex.LexFile( fname, text )
    tokenList = st.parser.ParseFile( fname, tokenList )
    return (tokenList,True)

#-------------------------------------------------------------------------------
# Printing messages.
#-------------------------------------------------------------------------------

# "Warning" is defined by Python BTW.

def Print( *args ):
    """ Only print a message if verbose. """
    if st.verbose:
        __Print( sys.stdout, *args )

def PrintError( *args ):
    """ Print an error message. """
    st.error = True
    __Print( sys.stderr, "ERROR: ", *args )
    try:
        if st.gui.IfReady():
            st.gui.statusbar.PrintError( *args )
    except:
        pass

def PrintWarning( *args ):
    """ Print a warning message. """
    st.warning = True
    __Print( sys.stderr, "WARNING: ", *args )
    try:
        if st.gui.IfReady():
            st.gui.statusbar.PrintWarning( *args )
    except:
        pass

# PRIVATE
def __Print( file, *args ):
    """ Print to a file and the log. """
    try:
        logText = ""
        for a in args:
            astr = str(a)
            logText += astr
            file.write( astr )
            file.flush()
        if astr[-1] != "\n":
            logText += "\n"
            file.write( "\n" )
            file.flush()
        st.log.Write( logText )
    except:
        PrintException()

#-------------------------------------------------------------------------------
# Class Log.
#-------------------------------------------------------------------------------

class Log:
    # The usual methods are thread-safe.
    
    VERBOSE = -1
    NORMAL  = 0
    URGENT  = 1
    
    def __init__( s ):
        s.buf        = ""
        s.enabled    = True
        s.level      = Log.NORMAL
        s.splashLine = ""
        s.mutex      = threading.RLock()  # reentrant lock

    # INTERFACE
    def Clear( s ):
        s.mutex.acquire()
        s.buf = ""
        s.buf += s.splashLine
        s.mutex.release()

    # PRIVATE
    def __Write( s, text, level=NORMAL ):  # level=Log.NORMAL won't compile
        # PRECOND: Mutex was acquired.
        if s.enabled and level >= s.level:
            if text[-1] == "\n":
                s.buf += text
            else:
                s.buf += text + "\n"

    # INTERFACE
    def Write( s, text, level=NORMAL ):  # level=Log.NORMAL won't compile
        try:
            s.mutex.acquire()
            s.__Write( text, level )
        finally:
            s.mutex.release()
        
    # INTERFACE
    def WriteList( s, textList, level=NORMAL, prologue="" ):
        """ Write a list of log messages. """
        try:
            s.mutex.acquire()
            for text in textList:
                s.__Write( prologue + text, level )
        finally:
            s.mutex.release()

    # INTERFACE
    def Read( s ):
        """ Returns a copy of the log. """
        # Return a copy to prevent inteference by later logging.
        s.mutex.acquire()
        buf2 = s.buf[:]
        s.mutex.release()
        return buf2

    # INTERFACE
    def Enable( flag=True ):
        # Pass False to disable logging.
        # Returns the old flag so that it can be restored.
        s.mutex.acquire()
        old = s.enabled
        s.enabled = flag
        s.mutex.release()
        return old
        
    # INTERFACE
    def SetLogLevel( s, level ):
        s.level = level
        
    # INTERFACE
    def SetSplashLine( s, line ):
        """ Sets the first line shown in the log. """
        s.splashLine = line
        s.Write( s.splashLine )

#-------------------------------------------------------------------------------
# Easter eggs.
#-------------------------------------------------------------------------------

def Underdog():
    text = \
    "------------------------------------------------------------------------------\n"\
    " Look!  Up in the sky!  It's a bird!  It's a plane! \n"\
    " No, it's Underdog! \n"\
    "                                    ,.+----.\n"\
    "                               ,*''         I\n"\
    "              ____           /         Mbp. dP          __\n"\
    "          ,o@QBBBBBbm,_     ; ,d       /~`QMP'\\     _odBBBBbo,\n"\
    "          OBBROYALBBBBBBBBmOBBBM      |    |   |,*~dBBBBBBBBBBBb \n"\
    "          OBBBBBBBBBBqBBBBBBBBBP      |  0 |0 ,'  dBBBBBBBBBBBBBb\n"\
    "          '*BBBP'     ^OBBBP'          \\__/  ~    OBBBBBBBBBBBBBP\n"\
    "      /                     :           __.       `0BBBBBBBBBBP|\n"\
    "                    .-.     `          ~,A.         `0BBBBBBBBP |\n"\
    "                 .-(   `x__  `\\        .B*Bb          `'?OOP~   ;\n"\
    "              .-(   `\\/'   `.  `~--++*\\{  `Pb           ,'     ,'\n"\
    "   /         (   `\\_/'\\__/  )     _    !`-' !`.       ,'     ,'  \n"\
    "        /     `\\_/'       ,'     L '+,_|    |-,`.__ ,'    _,','~~\\.~~~\\\n"\
    "                 \\        /\\    / ',   '+.  |/ >   `~~~~~~  ,'    :/~  \\\n"\
    "    /             `~===\\/'  `\\/'    '+.   >{  /\\        __.,: |   ;    ;\n"\
    "           /           {         ,|    `\\/  \\/  `Y~~---~   `. `--:'~~~';\n"\
    "                        `\\     ,&#|    uuu    uuu \\         `,   ;^`~~';\n"\
    "        /               ,d#b,.&###|    UUU.   UUU. `\\     __.;~~' `~~~'\n"\
    "      /              ,d###########|    UUU;   |UU:   Y~~~~\n"\
    "                 ,d############i~      `UUUUUUUUU;   I  \n"\
    "           ,d##################I        `UUUUUUU'    I     /   /\n"\
    " (#############################b.                    Im \n"\
    "    )############################b.      \\           d#b\n"\
    "   (###############################b++g+++~#b      ,d##P\n"\
    "       }####################P~ ~Y####P    J#P`~~~T'\n"\
    "      (###################P'     `YP'    ;P'   ,'\n"\
    "        )###############P'        '    ,'    ,'    /\n"\
    "       (##############P'          __+~'   /##'\n"\
    "  /            Y####P'         ;-'     /###P\n"\
    "            /   Y#P'         /'      /&##P~\n"\
    "                 |         /'      /&#P~         \n"\
    "      /          `\\______/'\\_____/'            \n"\
    "                                    /            \n"\
    "                           /\n"\
    "              /                          \n"\
    "------------------------------------------------------------------------------"

    st.log.Write( text )
    print text
# $Id: Threads.py,v 1.26 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# Notes:
# ==============================================================================

# ==============================================================================
# Thread design:
#
# Why are multiple threads used?  To fix the annoyance of the statusbar/GUI
# not being updated during long operations by running them in another thread.
#
# pypersrc can run 1 or 2 threads using the Python "threading" module.
# Most GUI functions aren't thread-safe, so the technique of top/bottom halfs is used.
# The initial main thread is dedicated to the GUI (it can call any GUI method).
# Long operations (taking more than a few seconds) are run by a temporary second thread
# which is forbidden to call GUI methods except a few that are thread-safe.
#
# The second thread is created as an object of the Python class threading.
# It executes as a top-half and stores its results as members.
# The first thread polls the second from a GUI idle function.
# When the second thread expires, its threading object persists as a zombie.
# The first thread "harvests" the results by calling the threading object's BottomHalf()
# (even though the second thread of execution it was meant for has expired!).
# 
# Skeleton class for thread objects based on the idea of top/bottom halfs.
# The run method (required by the threading class) is what invokes the top-half.
# Threads.Spawn() creates a Thread* object and passes a set of args to it
# as if it were a function.  How the args are passed to the top-half
# is somewhat convoluted: the Thread* constructor stores them, and the run() method
# passes them to the TopHalf() method.
#
# class ThreadDerived(ThreadBase):
#    
#    def __init__( s, *argsTopHalf ):
#       ThreadBase.__init__( s, argsTopHalf )
#        
#    def TopHalf( s [,arg1,arg2,..] ):
#        # This is executed by the spawned thread.
#        ...
#        # returning terminates the thread
#
#    def BottomHalf( s ):
#        # This is called and run by the main thread (not this thread) at an idle function.
#
# Remember: Bottom-half code usually shouldn't contain any calls to the GUI.
# Calling st.gui.Method() is a bug unless it is thread-safe (such as Statusbar.Print()).
#
# ==============================================================================

#-------------------------------------------------------------------------------
# Class Mutex.
#-------------------------------------------------------------------------------

class MutexBase:
    """ Helper class providing a wrapper for executing a function protected by a mutex. """
    # Example:
    #     class Class(MutexBase):
    #         def CritSect( s, x, y):
    #         def Whatever( s ):
    #             s.MutexCall( s.CritSect, x, y )
    #                          ^ pass as self.method (not just method)

    def __init__( s ):
        s.mutex = threading.RLock()    # reentrant lock

    def MutexCall( s, method, *args ):
        """ Protect with a mutex a call to a class method having a critical section. """
        # This may look as if it could return before releasing the mutex,
        # but the finally clause will always execute in any case.
        try:
            s.mutex.acquire()
            return method( *args )    # in Python, functions always return something (None by default)
        finally:
            s.mutex.release()    # always executes (even if try returns)

#-------------------------------------------------------------------------------
# Class Threads (unrelated to ThreadBase).
#-------------------------------------------------------------------------------

class Threads:
    # Manages the second thread.
    # Threads are spawned by pass their class object to Spawn().
    # __Harvest() is called by the main thread when GUI is idle,
    # and executes the bottom-half of an expired thread object
    # (does something with the results the thread computed).

    def __init__( s, argList ): # argList: cmd-line args (unused)
        # PRECOND: GUI was started.

        # Points to the second thread which is an object of class Thread*.
        s.thread = None

    # INTERFACE
    def Spawn( s, ThreadClass, *argsTopHalf ):
        # Spawn a second thread.
        # Passes args to thread'ss top-half when thread is activated.
        # ThreadClass must be a pypersrc Thread* class (derived from threading class).

        if (s.thread is not None) and (s.thread.isAlive()):
            # For simplicity, limit pypersrc to a main thread and one temporary second thread.
            # s.thread may point to an expired object.
            PrintWarning( "thread still busy, try again" )
        else:
            # Call class constructor, instantiate a "threading" object.
            s.thread = ThreadClass( *argsTopHalf )  # construct threading object
            s.thread.start()                        # schedule second thread
            s.__ScheduleHarvest()                   # schedule GUI idle func (for main thread)

    # PRIVATE
    def __Harvest( s ):
        # This is called by the main thread when the GUI idle.
        # A second thread exists.  Is it still running?
        if s.thread is not None:
            if s.thread.isAlive():
                s.__ScheduleHarvest()    # still running, reschedule
            else:
                # Clear s.thread before (not after) calling BottomHalf()
                # as it could spawn a thread which sets s.thread.
                thread_ = s.thread
                s.thread = None
                thread_.BottomHalf()    # call bottom half

    # PRIVATE
    def __ScheduleHarvest( s ):
        st.gui.ScheduleTimerFunc( 250, s.__Harvest )

#-------------------------------------------------------------------------------
# Class ThreadsBase.
#-------------------------------------------------------------------------------

class ThreadBase(threading.Thread):
    
    def __init__( s, *argsTopHalf ):
        # This is the thread constructor.  The thread really isn't running yet.
        # Store the passed args which are actually for TopHalf(), not this.
        s.argsTopHalf = argsTopHalf
        threading.Thread.__init__( s )

    def run( s ):
        """ This is where the thread begins by threading.start(). """
        # Pass the stored args to TopHalf().
        s.TopHalf( *s.argsTopHalf )

    def TopHalf( s ):
        """ Thread executes this.  Returning causes it to terminate. """
        # Derived class defines TopHalf()'s args.
        raise Oops( "abstract method" )

    def BottomHalf( s ):
        """ This is called and run by the main thread (not this thread) at an idle function. """
        raise Oops( "abstract method" )

#-------------------------------------------------------------------------------
# Class ThreadGatherTag
#-------------------------------------------------------------------------------

class ThreadGatherTag(ThreadBase):
    
    def __init__( s, *argsTopHalf ):
        ThreadBase.__init__( s, *argsTopHalf )
        s.srcFileList = []
        
    def TopHalf( s, argList ):  # argList: cmd-line args
        # Gather and tag source files.
        s.srcFileList = st.gather.Gather( argList )
        if s.srcFileList:
            st.gui.statusbar.Print( "Tagging source files...", timeout=0 )
            st.tags.TagFiles( s.srcFileList )
            # ...the rest is deferred to BottomHalf()...
        else:
            st.gui.statusbar.Print( "You need to pass names of source files or a directory (nothing to browse)." )

    def BottomHalf( s ):
        """ This is called and run by the main thread (not this thread) at an idle function. """
        st.gui.SetFileList( s.srcFileList )
        st.gui.statusbar.Print( st.GetSplashText() )
        if not s.srcFileList:
            text = "No src files were gathered.\nPass names of dirs/files on cmd-line."
            PrintWarning( text )

#-------------------------------------------------------------------------------
# Class ThreadRefresh
#-------------------------------------------------------------------------------

class ThreadRefresh(ThreadBase):
    
    def __init__( s, *argsTopHalf ):
        ThreadBase.__init__( s, *argsTopHalf )
        
    def TopHalf( s ):
        # Regather src files to determine if refresh is necessary.
        st.gui.statusbar.Print( "Refreshing...", timeout=0, now=True )
        st.log.Write( "refresh: start" )
        (s.change, s.modifiedList, s.removedList, s.addedList, s.unchangedList) = st.gather.Refresh()
        if s.change:
            st.refreshCnt += 1    # only increment if a change
            st.log.Write( "refresh: source files changed" )
            if s.modifiedList:
                st.log.WriteList( s.modifiedList, prologue="refresh: modified file: " )
            if s.removedList:
                st.log.WriteList( s.removedList, prologue="refresh: removed file: " )
            if s.addedList:
                st.log.WriteList( s.addedList, prologue="refresh: added file: " )

    def BottomHalf( s ):
        """ This is called and run by the main thread (not this thread) at an idle function. """        
        # Did src files change?
        if s.change:
            # Refresh.
            st.gui.Enable( False )
            try:
                s.Refresh()
            except:
                PrintException()
            st.gui.Enable( True )
        else:
            # No refresh necessary.
            st.log.Write( "refresh: no src files changed" )
        st.log.Write( "refresh: done" )

    def Refresh( s ):
        """ Refresh after src files has changed. """
        # PRECOND: Src files changed
        # PRECOND: GUI is disable.

        # Refresh tags.
        st.gui.statusbar.Print( "Refreshing tags...", timeout=0 )
        freeTagFileList = s.modifiedList + s.removedList
        tagFileList     = s.modifiedList + s.addedList
        st.tags.FreeTags( freeTagFileList )
        if tagFileList:
            st.log.Write( "refresh: tagging modified/added files" )
            st.tags.TagFiles( tagFileList  )
        else:
            st.log.Write( "refresh: tags = " + str(st.tags.TagCnt()) )
        # Refresh history.
        st.gui.resultsHistory.Refresh( s.change, s.addedList, s.modifiedList, s.removedList )
        # Refresh source file listbox.
        srcFileList = s.modifiedList + s.unchangedList + s.addedList
        srcFileList.sort()
        st.gui.SetFileList( srcFileList )
        # Flush affected src files out of cache.
        for fname in s.modifiedList + s.removedList:
            st.gui.cacheSrcFile.Flush( fname )
        # Refresh text widgets showing affected src files.
        for textWid in st.gui.notebookSrcFile.GetTextWidList():
            textWid.Refresh( s.change, s.addedList, s.modifiedList, s.removedList )
        # Refresh the results text widget which may be showing stale results.
        # Subtle: ResultsTextWid.Refresh() can cause a second thread to spawn
        # (eg global search) which will work as this is run by the main thread.
        st.gui.GetResultsTextWid().Refresh( s.change, s.addedList, s.modifiedList, s.removedList )
        st.gui.statusbar.Print( "Refreshing...", timeout=1000 )   # timeout in one second

#-------------------------------------------------------------------------------
# Class ThreadFindText
#-------------------------------------------------------------------------------

class ThreadFindText(ThreadBase):
    
    def __init__( s, *argsTopHalf ):
        ThreadBase.__init__( s, *argsTopHalf )
        
    def TopHalf( s, command, finder, srcFileList ):
        # Top-half:
        # This produces HTML-like results in a text widget.
        # The user can click on a hyperlink to jump to the match.
        # This builds a token list consisting of two types of tuples:
        # (eTokenTag,(PhonyTag))
        # (eTokenText,text)

        tokenList = []
        pattern = finder.GetPattern()
        local = (len(srcFileList) == 1)

        # Although searchedFileCnt is accessed by two threads, no mutex
        # is necessary as it's just a fundamental type used only for display.
        searchedFileCnt = 0
        totalFileCnt = len(srcFileList)
        def SearchUpdate():
            st.gui.statusbar.Print( "Searching for `" + pattern + "` in " \
                                    + str(searchedFileCnt) + "/" + str(totalFileCnt) + " files", \
                                    baton=True )
        st.gui.statusbar.SetUpdateCallback( SearchUpdate )
        
        for srcFile in srcFileList:
            if st.escKey:
                st.log.Write( "ThreadFindText: searching stopped by user" )
                break
            
            searchedFileCnt += 1

            # Find all matches in this src file.
            try:
                text = ReadSrcFile( srcFile )
            except IOError:
                text = ""
                PrintWarning( "ThreadFindText: can't open file " + srcFile )
            lineList = finder.GetMatchingLines( text, srcFile )
            for line in lineList:
                phonyTag = st.tags.PhonyTag( srcFile + ":" + str(line), srcFile, line )
                tokenList.append( (eTokenTag,phonyTag) )

        st.gui.statusbar.SetUpdateCallback( None )

        if len(srcFileList) == 1:
            # fnames of local searches should be unique.
            fname = "[results: " + command + " (" + srcFileList[0] + ") ]"
        else:
            fname = "[results: " + command + "]"
        finderModes = finder.GetModesString()
        if tokenList:
            tokenList.insert( 0, (eTokenText,"Found `" + pattern + "` " + finderModes + ":\n\n") )
        else:
            tokenList.insert( 0, (eTokenText,"Did not find `" + pattern + "` " + finderModes + ".\n\n") )

        # Save state for bottom-half.
        s.tokenList = tokenList
        s.fname = fname
        s.command = command
            
    def BottomHalf( s ):
        """ This is called and run by the main thread (not this thread) at an idle function. """
        contents = ContentsSearch( tokenList=s.tokenList, fname=s.fname, command=s.command )
        st.gui.GetResultsTextWid().LoadContents( contents=contents )
        st.gui.statusbar.Clear()
# $Id: Gather.py,v 1.38 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Class Gather.
#-------------------------------------------------------------------------------

class Gather(MutexBase):
    """ Gathers a list of source files to be browsed. """

    def __init__( s, argList ):
        MutexBase.__init__( s )
        
        s.ProcessArgsPre( argList )
        
        # This defines the list of source files to browse.
        # A source file's modification time is remembered
        # to skip re-tagging it by refresh if it wasn't modified.
        s.srcFileDict       = {}                 # value is file modification time
        s.srcFileListSorted = []                 # should be equivalent to sorting keys of s.srcFileDict
        s.savedArgList      = []
        s.cwd               = os.getcwd()
        s.excludeFinderList = []                 # list of Finder objects that exclude files/dirs
        
        s.ProcessArgsPost( argList )

    # INTERFACE
    def ProcessArgsPre( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True

    # INTERFACE
    def ProcessArgsPost( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        try:
            i = 0
            while i < len(argList):
                arg = argList[i]
                if arg is None:
                    i += 1
                elif arg == "-x" or arg == "-xc":
                    # -x[c] <glob> excludes files/dirs.  [c] means case-sensitive.
                    # -x should be followed by one glob (not multiple).
                    # Design note: "-x glob1 glob2" would open the door to misinterpretation.
                    # Eg "pypersrc -x glob ." means exclude glob and browse CWD (not glob2='.').
                    argList[i] = None
                    i += 1
                    # The next arg should be one glob pattern.
                    if i < len(argList):
                        pattern = argList[i]
                        caseSens = (arg == "-xc")
                        finder = Finder( pattern=pattern, glob=True, caseSens=caseSens )
                        s.excludeFinderList.append( finder )
                        argList[i] = None
                        i += 1
                    else:
                        PrintWarning( "Gather: -x <glob> is missing glob arg." )
                else:
                    i += 1
        except:
            PrintException()
        return True

    # INTERFACE
    def Gather( s, argList ):
        """ Gathers/updates a list of source files to be browsed.  List is sorted by fname. """
        # - Updates s.srcFileDict,s.srcFileListSorted.
        #   This is a dictionary of filename,modificationTime pairs.
        #   The idea is to only re-tag files that were modified.
        # - Returns a list of sources files that need to be tagged.
        #   This excludes unmodified files.
        # - This accepts cmd-line args but is passed file/directory names
        #   by the GUI code (by a file dialog).
        return s.MutexCall( s.__Gather, argList )

    # PRIVATE
    def __Gather( s, argList ):
        # Temps used across methods.
        s.passCnt = 1    # information for user, changes each pass
        s.fileCnt = 0    # information for user, changes each pass (this isn't the final src file cnt!)
        s.dirCnt  = 0    # information for user
        
        st.log.Write( "Gather: gathering source files in programming languages " + str(st.langList) )
        
        # Log file/dir args (not yet proven that they are).
        fileArgInfo = ""
        for arg in argList:
            if arg is not None:
                fileArgInfo = AddIntoParentheses( fileArgInfo, arg )
        st.log.Write( "Gather: file/dir args: " + fileArgInfo )

        s.savedArgList = argList[:]
        s.srcFileDict.clear()

        def GatherUpdateStatusbar( gather ):
            # -- executed by main thread --
            # Gather.passCnt/fileCnt are just ints, using a mutex isn't necessary.
            text = "Gathering[" + str(gather.passCnt) + "]: " + str(gather.fileCnt) + " files"
            st.gui.statusbar.Print( text, baton=True )

        # First pass.
        # Build fileList from args.
        # Append args which are filenames.
        # For args which are directories, recurse into them, append filenames.
        s.passCnt = 1    # changed each pass
        s.fileCnt = 0    # changed each pass
        fileList = []
        st.gui.statusbar.SetUpdateCallback( GatherUpdateStatusbar, s )   # pass this object as "s"
        for arg in argList:
            try:
                if arg is None:
                    continue
                elif arg[0] == "-":                 # if not a hypenated cmd-line arg
                    continue
                elif os.path.isdir( arg ):          # dir?
                    dir = AbsPath( arg, s.cwd )
                    s.__GatherDir( dir, fileList )  # appends to fileList
                elif os.path.isfile( arg ):         # file?
                    fileList.append( os.path.normpath(arg) )
                    s.fileCnt += 1
                else:
                    s.__WarnUnrecognizedArg( arg )
            except:
                PrintException()

        # Second pass with gathered filenames found in dirs.
        s.passCnt = 2    # changed each pass
        s.fileCnt = 0    # changed each pass
        try:
            for file in fileList:
                s.fileCnt += 1
                
                # Is this a source file to browse?
                if s.__IfSrcFileToBrowse( file ):
                    file = TransformPath( file )
                    # OPTIMIZE: Logging every file is a bottleneck!
                    #st.log.Write( "Gather: found file to browse: " + file )
                    #st.log.Write( "Gather: found file to browse: " + file, Log.VERBOSE )
                    # Store file's modification time.
                    s.srcFileDict[file] = FileModificationTime( file )
        except:
            PrintException()
        st.gui.statusbar.SetUpdateCallback( None )
        fileList = None   # superceded by s.srcFileDict

        if st.escKey:
            st.log.Write( "Gather: stopped by user" )
        else:
            st.log.Write( "Gather: gathered " + str(len(s.srcFileDict)) + " src files in " +str(s.dirCnt) + " dirs" )

        # Return sorted list of source files.
        st.gui.statusbar.Print( "Sorting gathered source files...", timeout=0 )
        s.srcFileListSorted = s.srcFileDict.keys()
        s.srcFileListSorted.sort()
        st.gui.statusbar.Clear()
        return s.srcFileListSorted

    # PRIVATE
    def __Regather( s ):
        s.Gather( s.savedArgList )

    # PRIVATE
    def __GatherDir( s, dirPathName, fileList ):
        """ Recursively gather source files in a directory. """
        # Modifies fileList in-place.

        s.dirCnt += 1
        prettyDirName = PrettyDirName(dirPathName)

        # OPTIMIZE: These are too slow.
        #st.log.Write( "Gather: recursing dir " + prettyDirName, Log.VERBOSE )
        #st.gui.statusbar.Print( "Gathering " + prettyDirName, timeout=0 )
        
        for entry in os.listdir( dirPathName ):
            # Use absolute pathnames only.  Mixing relative and absolute
            # may cause the problem of duplicate files.
            pathName = os.path.join( dirPathName, entry )
            pathName = os.path.join( s.cwd, pathName )
            if entry in st.ignoreDirList:
                st.log.Write( "Gather: ignoring dir " + PrettyDirName(pathName) )
            elif s.__IsExcluded(entry):
                # Exclude file or dir.
                st.log.Write( "Gather: excluding " + pathName )
            elif os.path.isdir( pathName ) and not st.escKey: # don't recurse if user pressed Esc
                s.__GatherDir( pathName, fileList )
            elif os.path.isfile( pathName ):
                fileList.append( os.path.normpath(pathName) )
                s.fileCnt += 1

    # INTERFACE
    def Refresh( s ):
        """ Refresh gathered source files. """
        # Returns (change=True/False, modifiedList, removedList, addedList, unchangedList)
        return s.MutexCall( s.__Refresh )

    # PRIVATE
    def __Refresh( s ):
        modifiedList  = []
        removedList   = []
        addedList     = []
        unchangedList = []
        
        # Re-gather source files, then examine any changes relative
        # to the previous gathering.  Re-gathering might add source files,
        # encounter modified ones, or omit removed ones.
        savedSrcFileDict = s.srcFileDict.copy()
        s.__Regather()    # updates s.srcFileDict

        # Determine which source files were modified, removed, or unchanged.
        for file in savedSrcFileDict.keys():
            if s.srcFileDict.has_key( file ):
                if s.srcFileDict[file] == savedSrcFileDict[file]:   # value is file modification time
                    unchangedList.append( file )
                else:
                    modifiedList.append( file )
            else:
                removedList.append( file )

        # Determine which source files were added.
        for file in s.srcFileDict.keys():
            if not savedSrcFileDict.has_key( file ):
                addedList.append( file )

        change = len(modifiedList) + len(removedList) + len(addedList) > 0
        return (change, modifiedList, removedList, addedList, unchangedList)

    # PRIVATE
    def __IfSrcFileToBrowse( s, fname ):
        """ Return true if this is a source file to browse according to its file extension. """
        try:
            ext  = FileExt( fname )       # get file extension
            lang = progLangExtDict[ext]   # if file extension is mapped to a known prog lang
            if lang in st.langList:       # if prog lang is supposed to be browsed
                return True
            else:
                return False
        except:
            return False

    # INTERFACE
    def IfSrcFile( s, fname ):
        """ Return true if this is a source file that was gathered. """
        try:
            s.mutex.acquire()
            retval = s.srcFileDict.has_key( fname )
            return retval
        finally:
            s.mutex.release()   # this executes always, even if try returns

    # INTERFACE
    def GetSrcFileList( s ):
        """ Return a list of source files that were gathered. """
        try:
            s.mutex.acquire()
            retval = s.srcFileListSorted[:]   # return copy to avoid thread conflicts
            return retval
        finally:
            s.mutex.release()

    # PRIVATE
    def __IsExcluded( s, fname ):
        """ Return true if file/dir was excluded by -x[c] cmd-line arg. """
        for finder in s.excludeFinderList:
            if finder.IfMatch( fname ):
                return True
        return False

    # PRIVATE
    def __WarnUnrecognizedArg( s, arg ):
        text = "Gather: unrecognized cmd-line arg '" + arg + "' which isn't a file/dir."
        # Show tip only once.
        try:
            if s.showedTipSubargsAreLimited:
                pass
        except:
            s.showedTipSubargsAreLimited = True
            text += "\nTip: Don't pass multiple sub-args on cmd-line: \n" \
                    "wrong: pypersrc -l c c++    \n" \
                    "right: pypersrc -l c -l c++ \n"
        PrintWarning( text )
# $Id: Lex.py,v 1.36 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Parsing source code for function calls, variable declarations, etc is simplified
# by first scanning thru the raw source code to produce "tokens" (this is the
# idea behind lex and yacc).  A parser can parse thru tokens rather than raw text.
# In this Python implementation, a token is a tuple consisting of an enumeration
# and the text that yielded the token.
#
# Class is-a:
#
# LexBase --> LexC
#             LexPython
#             LexPerl
#             
# Class has-a:
#
# Lex -->     LexC
#             LexPython
#             LexPerl
#-------------------------------------------------------------------------------

# Enums are prefixed by "e".
# 0x00..0xff  : chars with no corresponding tokens
# 0x100..0x1ff: tokens that can be combined if adjacent
# 0x200..0x2ff: tokens that can't be combined
# combinable:
eTokenInvalid     = 0x100
eTokenWs          = 0x101   # space, tab, newline
eTokenSymbol      = 0x102   # symbol/punctuation chars
eTokenNum         = 0x103   # numeric chars
eTokenWord        = 0x104   # a word or a possible identifier
eTokenKeyword     = 0x105   # a word or a possible identifier
eTokenKeywordLoud = 0x106   # a word or a possible identifier
eTokenType        = 0x107   # data type (int,bool)
eTokenText        = 0x108   # for results from commands (not used for src files)
# non-combinable:
eTokenTag         = 0x200   # a source code tag
eTokenQuoted      = 0x201   # text inside quotes  
eTokenComment     = 0x202   # comment (language specific)

# ------------------------------------------------------------------------------
# Class Lex.
# ------------------------------------------------------------------------------

class Lex:

    def __init__( s, argList ):
        s.ProcessArgsPre( argList )
        s.lexC        = LexC()        # includes C++
        s.lexPython   = LexPython()
        s.lexJava     = LexJava()
        s.ProcessArgsPost( argList )
        return

    # INTERFACE
    def ProcessArgsPre( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        i = 0
        while i < len(argList):
            arg = argList[i]
            if arg is None:
                i += 1
            elif arg.find( "-hilite-no" ) != -1:    # no/none
                LexBase.SetHilite( "no" )           # static method
                i += 1
            elif arg.find( "-hilite-lo" ) != -1:    # hi/high
                LexBase.SetHilite( "lo" )           # static method
                i += 1
            elif arg.find( "-hilite-hi" ) != -1:    # lo/low
                LexBase.SetHilite( "hi" )           # static method
                i += 1
            else:
                i += 1
        return True

    # INTERFACE
    def ProcessArgsPost( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True

    # INTERFACE
    def LexFile( s, fname, buf="" ):
        """ Lex a source code file.  Returns a list of token tuples. """

        # If buffer wasn't passed, open file.
        if buf == "":
            try:
                buf = ReadSrcFile( fname )
            except IOError:
                PrintWarning( "Lex: can't open file " + fname )
                return []

        proglang = WhichKindOfSrcFile( fname )

        if proglang == "c" or proglang == "c++":
            return s.lexC.LexBuffer( buf )
        elif proglang == "python":
            return s.lexPython.LexBuffer( buf )
        elif proglang == "java":
            return s.lexJava.LexBuffer( buf )
        else:
            PrintWarning( "Lex: unknown source file" )
            return []

# ------------------------------------------------------------------------------
# Class LexBase.
# ------------------------------------------------------------------------------

class LexBase:
    """ Tokenizes source code for parsing. """

    # 0=none 1=low 2=high
    hilite = 2

    # Precompute token table that correlates ASCII chars to eToken enums.
    # The token table is language-specific.
    tokenTable1 = []    # more hiliting (slower)
    tokenTable2 = []    # less hiliting (faster)
    for i in range(0,0x100):
        tokenTable1.append(0)
    for i in range(33,48):
        tokenTable1[i] = eTokenText
    for i in range(48,58):
        tokenTable1[i] = eTokenWord #eTokenNum  # treat numbers as alphabetic 
    for i in range(58,65):                       # to handle alpha-numeric tags
        tokenTable1[i] = eTokenSymbol
    for i in range(65,91):
        tokenTable1[i] = eTokenWord
    for i in range(91,97):
        tokenTable1[i] = eTokenSymbol
    for i in range(97,123):
        tokenTable1[i] = eTokenWord
    for i in range(123,127):
        tokenTable1[i] = eTokenSymbol
    for c in " \t\n\r":
        tokenTable1[ord(c)] = eTokenText
    tokenTable1[ord("_")] = eTokenWord
    # tokenTable2 is for low hiliting (lacks some tokens).
    tokenTable2 = tokenTable1[:]
    for i in range(0,0x100):
        if tokenTable2[i] in [eTokenSymbol,eTokenWs]:
            tokenTable2[i] = eTokenText
    tokenTable = tokenTable2

    def __init__( s ):

        # Here's Pythonic programming:
        # A base constructor is calling a method in a derived class (impossible in C++).
        # This is cleaner than rewriting it the C++ way by passing down keywordDict arg.
        keywordDict = s.KeywordDict()

        # Holds contents of a source file.
        s.deq = DequeChar()

        # Prepare for lexing keywords.
        s.keywordMin = 1000
        s.keywordMax = 0
        for keyword in keywordDict.keys():
            keywordLen = len(keyword)
            if keywordLen < s.keywordMin:
                s.keywordMin = keywordLen
            elif keywordLen > s.keywordMax:
                s.keywordMax = keywordLen

        # Used to match keywords using "greedy" searches.
        # Eg, to find "#ifdef" instead of just "#if".
        s.greedyDictKeywords = GreedyDict( s.keywordMax )
        for (key,val) in keywordDict.items():
            s.greedyDictKeywords[key] = val
        
        return

    # INTERFACE
    def LexBuffer( s, buf ):
        """ Lex a buffer (string).  Returns a list of token tuples. """

        # Use C++ STL deque which is more suitable and faster than Python strings.
        s.deq.Set( buf )
        
        # Ensure file's last char is EOL which eliminates special cases.
        if s.deq.PeekTail() != "\n":
            s.deq.PushTail( "\n" )
            
        tokenList = []
        token     = eTokenInvalid
        tokenPrev = eTokenInvalid
        lineNum   = 0

        # No hilite is simple.
        if LexBase.hilite == 0:
            try:
                text = s.deq.Get()
                t = (token,text)
                tokenList = [ t ]
            except:
                PrintException()
            return tokenList
        
        # First pass.
        try:
            while not s.deq.IfEmpty():
                c = s.deq.PeekHead()
                # Keyword?
                text = ""
                if LexBase.hilite == 2:
                    if tokenPrev != eTokenWord:  # avoid misidentifying a word containing a keyword
                        (token,text) = s.LexKeyword( tokenPrev )
                if text:
                    pass    # (token,text) already
                # Quoted string?
                elif c == '"' or c == "'":
                    token = eTokenQuoted
                    text  = s.LexQuotedString()
                # Comment?
                elif s.IfCommentBegin():
                    # Comments are lang-specific, defer to derived methods.
                    token = eTokenComment
                    text  = s.LexComment()
                # EOL?
                elif c == "\n":
                    s.deq.PopHead()
                    lineNum += 1
                    token = eTokenText
                    text  = c
                # Token in table?
                elif LexBase.tokenTable[ord(c)]:
                    s.deq.PopHead()
                    token = LexBase.tokenTable[ord(c)]
                    text  = c
                # Uncategorized token.
                else:
                    s.deq.PopHead()
                    continue    # ignore it

                # Combine the same kind of tokens (not all kinds can be combined).
                if token == tokenPrev and s.IfCombinableToken(token):
                    tokenList[-1] = (token,tokenList[-1][1] + text)
                    continue

                # Append tuple to token list.
                t = (token,text)
                tokenList.append( t )
                tokenPrev = token
        except:
            PrintException()

        # Second pass: Find words that really are tags.
        if LexBase.hilite:
            try:
                tokenList2 = []
                while tokenList:
                    token = tokenList[0][0]
                    text  = tokenList[0][1]
                    if token == eTokenWord and pycpp_tags.IfTag( text ):
                        token = eTokenTag    # really is a tag
                    tokenList2.append( (token,text) )
                    del tokenList[0]
            except:
                PrintException()
            tokenList = tokenList2

        # Third pass: Correct problem of false keywords.
        if LexBase.hilite == 2:
            try:
                tokenList = s.CorrectFalseKeywords( tokenList )
            except:
                PrintException()

        # Fourth pass: Combine tokens (again).
        if LexBase.hilite == 1:
            try:
                prev = (-999,"")
                tokenList2 = []
                i = 0
                j = 0
                for t in tokenList:
                    if t[0] == eTokenWord or t[0] == eTokenKeyword:
                        t = (eTokenText,t[1])

                    if t[0] != prev[0]:
                        tokenList2.append( t )
                        j += 1
                    else:
                        tokenList2[j-1] = (t[0], tokenList2[j-1][1] + t[1])
                    prev = t
                    i += 1
                tokenList = tokenList2
            except:
                PrintException()

        # Done.
        return tokenList

    # INTERFACE/STATIC
    def SetHilite( level ):  # self omitted
        # This is a Python static method -- eg LexBase.SetHilite("lo").
        if level in ["no","none"]:
            LexBase.tokenTable = LexBase.tokenTable2
            LexBase.hilite = 0
        elif level in ["lo","low"]:
            LexBase.tokenTable = LexBase.tokenTable2
            LexBase.hilite = 1
        elif level in ["hi","high"]:
            LexBase.tokenTable = LexBase.tokenTable1
            LexBase.hilite = 2
    SetHilite = staticmethod(SetHilite)

    # PROTECTED
    def IfCombinableToken( s, token ):
        return 0x100 <= token <= 0x1ff

    # PROTECTED
    def CorrectFalseKeywords( s, tokenList ):
        """ Eliminate false keywords.  Returns new tokenList. """
        # This is necessary because text such as "g_return_if_fail"
        # would be lexed incorrectly because "return" is a C keyword.
        # This method corrects such errors.
        # TODO For FORTH this would be different (in derived class).
        try:
            tokenList2 = []
            keywordList     = [eTokenKeyword,eTokenType]
            keywordLikeList = [eTokenWord,eTokenNum]   # tokens similar to keyword tokens
            while tokenList:
                token = tokenList[0][0]
                if not token in keywordLikeList and not token in keywordList:
                    tokenList2.append( tokenList[0] )
                    del tokenList[0]
                    continue                    
                # keywordlike, keyword, keywordlike
                if  tokenList[0][0] in keywordLikeList \
                and tokenList[1][0] in keywordList \
                and tokenList[2][0] in keywordLikeList:
                    token = tokenList[0][0]    # first is preferred, third may be number when first is alpha
                    text = tokenList[0][1] + tokenList[1][1] + tokenList[2][1]
                    tokenList2.append( (token,text) )
                    del tokenList[0:3]
                    continue
                # keywordlike, keyword
                if  tokenList[0][0] in keywordLikeList \
                and tokenList[1][0] in keywordList:
                    token = tokenList[0][0]
                    text = tokenList[0][1] + tokenList[1][1]
                    tokenList2.append( (token,text) )
                    del tokenList[0:2]
                    continue
                # keyword, keywordlike
                if  tokenList[0][0] in keywordList \
                and tokenList[1][0] in keywordLikeList:
                    token = tokenList[1][0]
                    text = tokenList[0][1] + tokenList[1][1]
                    tokenList2.append( (token,text) )
                    del tokenList[0:2]
                    continue
                # no mistaken keywords
                tokenList2.append( tokenList[0] )
                del tokenList[0]
        except:
            PrintException()
        return tokenList2
        
    # PROTECTED
    def LexQuotedString( s ):
        """ Lex a quoted string in source code. """

        quote = s.deq.PopHead()
        text = quote

        # Find the ending quote char.
        while not s.deq.IfEmpty():
            c = s.deq.PopHead()
            text += c
            # End quote?
            if c == quote:
                return text
            # Escaped char?
            elif c == "\\":
                # TODO handle more complex escaped cases.
                if not s.deq.IfEmpty():
                    text += s.deq.PopHead()
                
        # String wasn't terminated.
        return text

    # PROTECTED
    def LexKeyword( s, tokenPrev ):
        """ Lex a possible keyword.  Returns (token,keyword) or (None,None). """
        # Greedy match.
        try:
            try:
                text = s.deq.PeekHead( s.keywordMax )
            except:
                text = s.deq.PeekHead( s.deq.Size() )  # peeked past EOF
            (key,val) = s.greedyDictKeywords.Find( text )
            s.deq.PopHead( len(key) )
            return (val,key)
        except:
            return (None,None)

    # PROTECTED
    def LexDelim( s, delim ):
        """ Lex text til an end delimiter.
            Returns string including the end delimiter. """
        idx = s.deq.Index( delim )  # may raise exception 
        return s.deq.PopHead( idx + len(delim) )  # from beginning til end delimiter

# ------------------------------------------------------------------------------
# Class LexC (C/C++).
# ------------------------------------------------------------------------------

class LexC(LexBase):
    """ Lex class for C and C++. """

    def __init__( s ):
        LexBase.__init__( s )
        return

    # PROTECTED
    def KeywordDict( s ):
        return \
        { "return"            : eTokenKeyword,
          "if"                : eTokenKeyword,
          "else"              : eTokenKeyword,
          "return"            : eTokenKeyword,
          "for"               : eTokenKeyword,
          "while"             : eTokenKeyword,
          "do"                : eTokenKeyword,
          "continue"          : eTokenKeyword,
          "break"             : eTokenKeyword,
          "switch"            : eTokenKeyword,
          "case"              : eTokenKeyword,
          "default"           : eTokenKeyword,
          "sizeof"            : eTokenKeyword,
          "goto"              : eTokenKeywordLoud,
          "NULL"              : eTokenKeywordLoud,
          "asm"               : eTokenKeyword,
          # C pre-processsor
          "#define"           : eTokenKeyword,
          "#ifdef"            : eTokenKeyword,
          "#ifndef"           : eTokenKeyword,
          "#if"               : eTokenKeyword,
          "#else"             : eTokenKeyword,
          "#elif"             : eTokenKeyword,
          "#endif"            : eTokenKeyword,
          #  
          "#\tdefine"         : eTokenKeyword,    # not ideal but would be
          "#\tifdef"          : eTokenKeyword,    # slow/difficult as a special-case
          "#\tifndef"         : eTokenKeyword,
          "#\tif"             : eTokenKeyword,
          "#\telse"           : eTokenKeyword,
          "#\telif"           : eTokenKeyword,
          "#\tendif"          : eTokenKeyword,
          #  
          "#\t\tdefine"       : eTokenKeyword,
          "#\t\tifdef"        : eTokenKeyword,
          "#\t\tifndef"       : eTokenKeyword,
          "#\t\tif"           : eTokenKeyword,
          "#\t\telse"         : eTokenKeyword,
          "#\t\telif"         : eTokenKeyword,
          "#\t\tendif"        : eTokenKeyword,
          #  
          "#   define"        : eTokenKeyword,
          "#   ifdef"         : eTokenKeyword,
          "#   ifndef"        : eTokenKeyword,
          "#   if"            : eTokenKeyword,
          "#   else"          : eTokenKeyword,
          "#   elif"          : eTokenKeyword,
          "#   endif"         : eTokenKeyword,
          #  
          "#       define"    : eTokenKeyword,    # ok, that's enough
          "#       ifdef"     : eTokenKeyword,
          "#       ifndef"    : eTokenKeyword,
          "#       if"        : eTokenKeyword,
          "#       else"      : eTokenKeyword,
          "#       elif"      : eTokenKeyword,
          "#       endif"     : eTokenKeyword,
          #
          "#include"          : eTokenKeyword,
          "#undef"            : eTokenKeyword,
          "#error"            : eTokenKeyword,
          "#pragma"           : eTokenKeyword,
          "defined"           : eTokenKeyword,
          # C++
          "catch"             : eTokenKeyword,
          "delete"            : eTokenKeyword,
          "using"             : eTokenKeyword,
          "namespace"         : eTokenKeyword,
          "new"               : eTokenKeyword,
          "operator"          : eTokenKeyword,
          "private"           : eTokenKeyword,
          "public"            : eTokenKeyword,
          "this"              : eTokenKeyword,
          "throw"             : eTokenKeyword,
          "try"               : eTokenKeyword,
          "typeid"            : eTokenKeyword,
          "typename"          : eTokenKeyword,
          "static_cast"       : eTokenKeywordLoud,
          "const_cast"        : eTokenKeywordLoud,
          "reinterpret_cast"  : eTokenKeywordLoud,
          # C types
          "void"              : eTokenType,
          "char"              : eTokenType,
          "short"             : eTokenType,
          "int"               : eTokenType,
          "long"              : eTokenType,
          "float"             : eTokenType,
          "double"            : eTokenType,
          "const"             : eTokenType,
          "volatile"          : eTokenType,
          "register"          : eTokenType,
          "static"            : eTokenType,
          "extern"            : eTokenType,
          "signed"            : eTokenType,
          "unsigned"          : eTokenType,
          "struct"            : eTokenType,
          "union"             : eTokenType,
          "typedef"           : eTokenType,
          "enum"              : eTokenType,
          # C++ types
          "bool"              : eTokenType,
          "class"             : eTokenType,
          "explicit"          : eTokenType,
          "export"            : eTokenType,
          "friend"            : eTokenType,
          "mutable"           : eTokenType,
          "protected"         : eTokenType,
          "template"          : eTokenType,
          "virtual"           : eTokenType,
          "inline"            : eTokenType,
          # C++ STL types
          "string"            : eTokenType,
          "vector"            : eTokenType,
          "deque"             : eTokenType,
          "set"               : eTokenType,
          "multiset"          : eTokenType,
          "map"               : eTokenType,
          "multimap"          : eTokenType,
          "iterator"          : eTokenType,
          "const_iterator"    : eTokenType,
          "size_type"         : eTokenType,
        }
    
    # PROTECTED
    def IfCommentBegin( s ):
        """ Return True if this is the start of a comment. """
        try:
            text = s.deq.PeekHead( 2 )
            if text[0] != "/":
                return False
            if text[1] == "*":  # C comment?
                return True
            if text[1] == "/":  # C++ comment?
                return True
            return False
        except:
            return False

    # PROTECTED
    def LexComment( s ):
        """ Lex a comment.
            @pre deq is pointing to the start of a comment. """
        text = s.deq.PeekHead( 2 )
        if text == "/*":
            return s.LexDelim( "*/" )
        elif text == "//":
            return s.LexDelim( "\n" )
        else:
            raise Oops

# ------------------------------------------------------------------------------
# Class LexJava.
# ------------------------------------------------------------------------------

class LexJava(LexC):

    def __init__( s ):
        LexC.__init__( s )
        return

    # PROTECTED
    def KeywordDict( s ):
        return \
        { "abstract"          : eTokenKeyword,   
          "break"             : eTokenKeyword,
          "case"              : eTokenKeyword,
          "catch"             : eTokenKeyword,
          "class"             : eTokenKeyword,
          "continue"          : eTokenKeyword,
          "default"           : eTokenKeyword,
          "do"                : eTokenKeyword,
          "else"              : eTokenKeyword,
          "extends"           : eTokenKeyword,
          "false"             : eTokenKeyword,
          "finally"           : eTokenKeywordLoud,
          "for"               : eTokenKeyword,
          "if"                : eTokenKeyword,
          "implements"        : eTokenKeyword,
          "import"            : eTokenKeyword,
          "instanceof"        : eTokenKeyword,
          "interface"         : eTokenKeyword,
          "native"            : eTokenKeyword,
          "new"               : eTokenKeyword,
          "null"              : eTokenKeywordLoud,
          "package"           : eTokenKeyword,
          "return"            : eTokenKeyword,
          "strictfp"          : eTokenKeyword,
          "super"             : eTokenKeyword,
          "switch"            : eTokenKeyword,
          "synchronized"      : eTokenKeyword,
          "this"              : eTokenKeyword,
          "throw"             : eTokenKeywordLoud,
          "throws"            : eTokenKeyword,
          "transient"         : eTokenKeyword,
          "true"              : eTokenKeyword,
          "try"               : eTokenKeyword,
          "while"             : eTokenKeyword,
          # types
          "private"           : eTokenType,
          "protected"         : eTokenType,
          "public"            : eTokenType,
          "static"            : eTokenType,
          "final"             : eTokenType,
          "boolean"           : eTokenType,
          "byte"              : eTokenType,
          "char"              : eTokenType,
          "short"             : eTokenType,
          "int"               : eTokenType,
          "long"              : eTokenType,
          "float"             : eTokenType,
          "double"            : eTokenType,
          "const"             : eTokenType,
          "void"              : eTokenType,
          "volatile"          : eTokenType,
          "String"            : eTokenType,
        }

# ------------------------------------------------------------------------------
# Class LexScript.
# ------------------------------------------------------------------------------

class LexScript(LexBase):
    """ Base class for languages that use # for comments. """
    
    def __init__( s ):
        LexBase.__init__( s )
        return

    # PROTECTED
    def IfCommentBegin( s ):
        """ Return True if this is the start of a comment. """
        try:
            text = s.deq.PeekHead( 1 )
            return text[0] == "#"
        except:
            return False

    # PROTECTED
    def LexComment( s ):
        """ Lex a comment.
            @pre deq is pointing to the start of a comment. """
        return s.LexDelim( "\n" )

# ------------------------------------------------------------------------------
# Class LexPython.
# ------------------------------------------------------------------------------

class LexPython(LexScript):

    def __init__( s ):
        LexBase.__init__( s )
        return

    # PROTECTED
    def KeywordDict( s ):
        return \
        { "return"          : eTokenKeyword,
          "and"             : eTokenKeyword,
          "assert"          : eTokenKeyword,
          "break"           : eTokenKeyword,
          "class"           : eTokenKeyword,
          "continue"        : eTokenKeyword,
          "def"             : eTokenKeyword,
          "del"             : eTokenKeyword,
          "elif"            : eTokenKeyword,
          "else"            : eTokenKeyword,
          "except"          : eTokenKeyword,
          "exec"            : eTokenKeyword,
          "finally"         : eTokenKeyword,
          "for"             : eTokenKeyword,
          "from"            : eTokenKeyword,
          "global"          : eTokenKeyword,
          "if"              : eTokenKeyword,
          "import"          : eTokenKeyword,
          "in"              : eTokenKeyword,
          "is"              : eTokenKeyword,
          "lambda"          : eTokenKeyword,
          "not"             : eTokenKeyword,
          "or"              : eTokenKeyword,
          "pass"            : eTokenKeyword,
          "print"           : eTokenKeyword,
          "raise"           : eTokenKeywordLoud,
          "return"          : eTokenKeyword,
          "try"             : eTokenKeyword,
          "while"           : eTokenKeyword,
        }
# $Id: Parse.py,v 1.15 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Class Parse.
#-------------------------------------------------------------------------------

# Parse modes are set by bitwise combinations.  The parser can be in multiple modes.
# Src files are filtered by setting the parse modes of Parser object.
eParseDefault   = 0
eParseComments  = 1
eParseFunctions = 2
eParseData      = 4
parseModesNameDict = { eParseComments  : "comment",
                       eParseFunctions : "function",
                       eParseData      : "data" }
# Bitwise combination of parse modes that filter source code.
filterParseModes = eParseComments | eParseFunctions | eParseData

class Parse:
    """ Parse tokenized source code. """
    # See also Lex.py.

    def __init__( s, argList ):
        s.ProcessArgsPre( argList )
        s.modes = 0                          # bitwise combination
        s.ProcessArgsPost( argList )
        return

    # INTERFACE
    def ProcessArgsPre( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True
    
    # INTERFACE
    def ProcessArgsPost( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True

    # INTERFACE
    def IfFiltering( s ):
        """ Return True if the parser is filtering source code elements. """
        return s.modes & filterParseModes

    # INTERFACE
    def GetModesString( s, parseModes=None ):
        """ Return modes as a printable/informational string. """
        # Eg builds "functions,comments" or "default".
        if parseModes is None:
            parseModes = s.modes
        modesStr = ""
        for modeBit in parseModesNameDict.keys():
            if modeBit & parseModes:
                modesStr += parseModesNameDict[modeBit] + ","
        if modesStr:
            modesStr = modesStr[0:-1]    # chop final ","
        else:
            modesStr = "default"
        return "(modes=" + modesStr + ")"

    # INTERFACE
    def GetModes( s ):
        """ Return modes (bitwise int) the parser is current in. """
        return s.modes

    # INTERFACE
    def SetModes( s, modes ):
        """ Set the modes the parser will be in.  modes is a list of eParse* enums. """
        # modes should be bitwise int combination.
        s.modes = s.modes

    # INTERFACE
    def SetDefaultMode( s ):
        """ Set the parser default mode. """
        s.modes = 0

    # INTERFACE
    def AddMode( s, mode ):
        """ Put the parser into an additional mode. """
        s.modes = s.modes | mode

    # INTERFACE
    def RemoveMode( s, mode ):
        """ Take the parser out of a mode. """
        s.modes = s.modes & ~mode

    # INTERFACE
    def ParseFile( s, fname, tokenList ):
        """ Returns a second token list from the passed one. """
        # For example, to digest comments, the second token list would omit
        # most of the original tokens.

        # Show source file verbatim (not digested)?
        if not s.modes:
            return tokenList

        # Digest comments, functions, etc into a dictionary of lineNums/tokens.
        # Then sort and merge the digestions.
        # The line numbers are scaled, because, for example, a hyperlink tag
        # is a token as is the text token for its newline, but both are on the same line,
        # so the line number is scaled and the newline is given a line number one higher.
        digestedTokenDict = { }

        if eParseComments & s.modes:
            s.__DigestComments( fname, tokenList, digestedTokenDict )
        
        if    (eParseFunctions & s.modes) \
           or (eParseData      & s.modes):
            s.__DigestFunctionsOrData( fname, tokenList, digestedTokenDict )

        mergedTokenList = []
        keyList = digestedTokenDict.keys()
        keyList.sort()
        for key in keyList:
            mergedTokenList.append( digestedTokenDict[key] )
        return mergedTokenList

    # PRIVATE
    def __DigestComments( s, fname, tokenList, digestedTokenDict ):
        """ Adds comments to a dictionary of lineNums/tokens. """
        lineNum = 1
        for token in tokenList:
            if token[0] == eTokenComment:
                # See ParseFile() why line numbers are scaled.
                digestedTokenDict[lineNum*10+9] = (token[0], Chomp(token[1])+"\n") # token
            lineNum += token[1].count( "\n" )

    # PRIVATE
    def __DigestFunctionsOrData( s, fname, tokenList, digestedTokenDict ):
        """ Adds tags of functions or data to a dictionary of lineNums/tokens. """
        try:
            # Show a terse representation: show just a function on every line.
            # Get list of tags in this module.
            tagList = st.tags.GetModuleTags( fname )
            if tagList:
                # Build a dictionary of tags keyed by line number
                # in order to sort the tags by line number.
                tagDict = { }
                for tag in tagList:
                    tagDict[ tag[TAG_LINE] ] = tag

                # Build token list using tags sorted by line number.
                keyList = tagDict.keys()
                keyList.sort()
                for key in keyList:
                    tag = tagDict[key]
                    lineNumScaled = tag[TAG_LINE]*10
                    # Is this tag for a function/method?
                    if (eParseFunctions & s.modes) and (tag[TAG_TYPE] in ["function","method"]):
                        # See ParseFile() why line numbers are scaled.
                        digestedTokenDict[lineNumScaled] = (eTokenTag,tag[TAG_NAME]) # a token
                        # An empty line between helps the user to click.
                        digestedTokenDict[lineNumScaled+1] = (eTokenText,"()\n\n")
                    # Is this a tag for data?
                    if (eParseData & s.modes) and not (tag[TAG_TYPE] in ["function","method"]):
                        digestedTokenDict[lineNumScaled]   = (eTokenText,tag[TAG_TYPE] + " ")
                        digestedTokenDict[lineNumScaled+1] = (eTokenTag,tag[TAG_NAME])
                        digestedTokenDict[lineNumScaled+2] = (eTokenText,"\n\n")
        except:
            PrintException()
# $Id: Tags.py,v 1.36 2006-03-05 05:02:34 devel Exp $
# pypersrc.
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Class Tags.
#-------------------------------------------------------------------------------

# Import module exported from C++ component.
import pycpp_tags

#-------------------------------------------------------------------------------
# A source code tag (not to be confused with a Tk text tag) is a tuple
# consisting of (name,type,file,line) that originates from Exuberant ctags.
# Phony tags are produced by pypersrc itself which are shown as "results"
# in a text widget.  Eg the results from a text search, shown as hyperlinks,
# are phony tags.
#-------------------------------------------------------------------------------

# Tag tuple indexes.
TAG_NAME = 0
TAG_TYPE = 1
TAG_FILE = 2
TAG_LINE = 3

class Tags:
    """ This Python class uses the integrated Exuberant ctags and
        C++ components to tag source code. """

    def __init__( s, argList ):
        s.ProcessArgsPre( argList )
        s.mutex = threading.RLock()   # reentrant lock
        
        # ctags uses the env var TMPDIR.
        tmpdir = "/tmp"
        if os.environ.has_key( "TMPDIR" ):
            tmpdir = os.environ["TMPDIR"]
        else:
            # $HOME/tmp exists?
            if os.environ.has_key( "HOME" ):
                dir = os.environ["HOME"] + "/tmp"
                if os.path.isdir( dir ):
                    tmpdir = dir
        os.environ["TMPDIR"] = tmpdir
        
        s.ProcessArgsPost( argList )

    def ProcessArgsPre( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True # NOP

    def ProcessArgsPost( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True # NOP

    def TagStatusbarUpdate( s ):
        try:
            s.mutex.acquire()    # protect s.taggedFileCnt
            st.gui.statusbar.Print( "Tagging: " + str(s.TagCnt()) + " tags in " \
                                    + str(s.taggedFileCnt) + "/" + str(s.srcFileCnt) + " files", \
                                    baton=True )
        finally:
            s.mutex.release()
    
    def TagFiles( s, fileList ):
        """ Pass filenames to Exuberant ctags.
            fileList --> Exuberant ctags --> tag lines --> C++ multimap. """
        st.log.Write( "Tags: starting" )
        s.FreeTags( fileList )    # free tags before re-tagging due to refresh
        st.gui.statusbar.SetUpdateCallback( s.TagStatusbarUpdate )
        s.taggedFileCnt = 0
        s.srcFileCnt = len(fileList)
        for file in fileList:
            if st.escKey:
                st.log.Write( "Tags: stopped by user" )
                break
            pycpp_tags.TagFile( file )
            s.mutex.acquire()    # protect s.taggedFileCnt
            s.taggedFileCnt += 1
            s.mutex.release()
        st.log.Write( "Tags: total = " + str(s.TagCnt()) )
        st.gui.statusbar.SetUpdateCallback( None )
        return

    def TagCnt( s ):
        """ Return amount of tags. """
        return pycpp_tags.TagCnt()

    def FreeTags( s, fileList ):
        """ Free tags belonging to source code files. """
        for file in fileList:
            pycpp_tags.FreeTags( file )
        return

    def IfTagRef( s, tagName, fileName, line ):
        """ Return true if this is a reference to a tag. """
        # For example, a reference to a tag should be drawn as a hyperlink.
        # But a tag should be drawn normally on the line where it is defined.
        return pycpp_tags.IfTagRef( tagName, fileName, line )

    def FindTag( s, tagName, tagType=None, tagFile=None, caseSens=True, partial=False, regex=False ):
        """ Return a list of matching tags or None. """
        # Search for a tag can be constrained by tag's type and/or tag's file.

        # Match tags by name.
        tagList = pycpp_tags.FindTag( tagName, caseSens, partial, regex )
        
        # Constrain by tag type (partial matching).
        if tagType:
            tagList2 = []
            for tag in tagList:
                thisTagType = tag[TAG_TYPE]
                if thisTagType.find( tagType ) != -1:
                    tagList2.append( tag )
            tagList = tagList2
        
        # Constrain by tag file (fileglob matching).
        if tagFile:
            tagList2 = []
            for tag in tagList:
                thisTagFile = tag[TAG_FILE]
                if MatchFileglob( thisTagFile, tagFile ):
                    tagList2.append( tag )
            tagList = tagList2
        
        return tagList

    def GetModuleTags( s, moduleName ):
        """ Return a list of tags in a module (source file). """
        return pycpp_tags.GetModuleTags( moduleName )

    def ChooseTagCmp( s, choice1, choice2 ):
        # sort() comparison callback.  Used by ChooseTag().
        score1 = choice1[0]
        score2 = choice2[0]
        if score1 < score2:
            return 1          # 1 (not -1) so that highest score will be first element
        elif score1 == score2:
            return 0
        else:
            return -1

    def ChooseTag( s, clickedTagName, fnameBrowse ):
        """ Return a good and bad list of matching tags. """
        # This is called when the user clicks a hyperlink to a src code tag.
        # In the good list, the best match is the first element
        # and the least likely is last.
        # The bad list shows tags which are probably bad matches
        # such as a matching tag in a different programming language file. """
        # For tags that have the same name, the best matching tag would be:
        # - defined in the same file the user is browsing
        # - or the same kind of programming language
        # - or its pathname is nearest the file the user is browsing
        # The tags are ranked by accumulating a score:
        # 100000 : defined in the same file the user is browsing
        #  10000 : the same kind of programming language
        # -10000 : different kind of programming language
        #  <1000 : pathname is nearest the file the user is browsing
        #          (count of matching chars in pathname)
        # Then the good list is sorted by scores.
        # TODO:
        # - Rank higher methods in the same class.
        # - Rank higher a tag closer to where the user clicked.
        
        tagList = s.FindTag( clickedTagName )
        if not tagList:
            return ([],[])
        elif len(tagList) == 1:
            return (tagList,[]) # only one tag matches
        goodList = []
        badList  = []
        for tag in tagList:
            score = 0
            fname = tag[2]
            if fname == fnameBrowse:
                score += 100000
            if WhichKindOfSrcFile( fname ) == WhichKindOfSrcFile( fnameBrowse ):
                score += 10000
            else:
                score -= 10000
            score += s.CountMatchingChars( fname, fnameBrowse )
            t = (score,tag)
            if score > 0:
                goodList.append( t )
            else:
                badList.append( t )
                
        # Sort the good list so that the tag with highest score
        # is the first element (to be shown at top of results text widget).
        goodList.sort( s.ChooseTagCmp )
        
        # Derive tag lists from good/bad lists (sans scores).
        goodTagList = []
        for choice in goodList:
            tag = choice[1]
            goodTagList.append( tag )
        badTagList = []
        for choice in badList:
            tag = choice[1]
            badTagList.append( tag )
        return (goodTagList,badTagList)

    def CountMatchingChars( s, x, y ):
        """ From the left, count chars in two strings that match. """
        n = 0
        i = 0
        while i < len(x) and i < len(y):
            if x[i] == y[i]:
                n += 1
            i += 1
        return n
                
    def PhonyTag( s, tagName, fileName, lineNum ):
        """ Return a new phony tag. """
        return (tagName,"phony",fileName,lineNum)

    def IfPhonyTagInToken( s, token ):
        try:
            # Could be a better way.
            return token[0] == eTokenTag and token[1][TAG_TYPE] == "phony"
        except:
            return False
# $Id: Command.py,v 1.27 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# Notes:    For consistency, commands with a file arg take it last.
# ==============================================================================

# See GuiText.py for mouse button behavior for text widgets.

helpText = \
"Keyboard/mouse:                                                                  \n" \
"---------------                                                                  \n" \
"button 1|>1   load into new|same page                                            \n" \
"F1            history of command results                                         \n" \
"F2            show results from last command                                     \n" \
"Ctrl-W        close page                                                         \n" \
"refresh       click this after you modify a source file                          \n" \
"up|down       previous|next command in history (Command entry widget)            \n" \
"                                                                                 \n" \
"Commands (basic):                                                                \n" \
"-----------------                                                                \n" \
"h[elp]|?                         help screen                                     \n" \
"n                                goto line                                       \n" \
"b[cr] <file>                     browse source file                              \n" \
"f[cr] <text> [[cr] [~]<file>]..  find text in source file(s)                     \n" \
"g[cr] <text>                     find text in all source files                   \n" \
"t[cre] <tag>|* [type]|* [file]   find a source code tag                          \n" \
"                                                                                 \n" \
"Matching:                                                                        \n" \
"[c] case-sensitive, [r] regex, [e] exact, [*] match any, [~] negative            \n" \
"For filename matching, globbing is implied if pattern contains a '*?[]' char.    \n" \
"                                                                                 \n" \
"Examples:                                                                        \n" \
"b debug              list src files whose names match 'debug'                    \n" \
"b *.c                lists C src files                                           \n" \
"t *                  lists all tags                                              \n" \
"t * func             lists tags of all functions                                 \n" \
"t Get func           lists tags of all functions named Get (partial matching)    \n" \
"t Get func *.c       lists tags of all functions in all C src files              \n" \
"t Get * *.c          lists all tags partially matching Get in all C src files    \n" \
"t * * *.c            lists all tags in all C src files                           \n" \
"f error              search for 'error' in current file                          \n" \
"f error *            search for 'error' in all files                             \n" \
"g error              same as 'f error *'                                         \n" \
"fc Error             case-sensitive search for 'Error'                           \n" \
"fc Error *.c         likewise but only search in C files                         \n" \
"fc Error c *.c       likewise but file matching is case-sensitive, excludes *.C  \n" \
'f Error *.cc *.py    include C++ and Python file in global search                \n' \
'f Error ~*.cc ~*.py  exclude C++ and Python files in global search               \n' \
"fr return.*false     search for regular expression                               \n" \
'f "Hello World"      surround with quotes a pattern that is multiple words       \n' \
'f ""Hello World""    repeat leading or trailing quote to include it in pattern   \n' \
"                                                                                 \n" \
"Commands (advanced):                                                             \n" \
"--------------------                                                             \n" \
"class                show class diagram (TODO)                                   \n" \
"trace                show chains of function calls (TODO)                        \n" \
"config               edit configuration (TODO)                                   \n" \
"                                                                                 \n" \
"Commands (misc):                                                                 \n" \
"----------------                                                                 \n" \
"hilite no|lo|hi      level of syntax highlighting (affects loading speed)        \n" \
"                                                                                 \n" \
"Commands (debug):                                                                \n" \
"-----------------                                                                \n" \
"log                  show log                                                    \n" \
"log [c]lear          clear log                                                   \n" \
"log save <file>      save log to file                                            \n" \
"                                                                                 \n" \
"Command-line arguments:                                                          \n" \
"-----------------------                                                          \n" \
"[dir..] [files..]    src files and/or directories to browse                      \n" \
"-x[c] <glob>         exclude files [c: case-sensitive]                           \n" \
"                                                                                 \n" \
"Program updates available at:                                                    \n" \
"http://www.jimbrooks.org/hypersrc/                                               \n"

#-------------------------------------------------------------------------------
# Class Command.
#-------------------------------------------------------------------------------

class Command:

    # Internal commands (as if the user entered them in the Command entry).
    # Internal and user commands are stored in the command history.
    INTERNAL_COMMAND_CLICK_TAG    = "[click_tag]"
    INTERNAL_COMMAND_SPLASH       = "[splash]"

    def __init__( s, argList ):
        s.commandList    = []          # list of user commands (for command history)
        s.commandListIdx = 0           # command recalled in command history

    # INTERFACE
    def Recall( s, next=False, prev=False ):
        """ Recall the next or previous command.  Returns empty string if no commands. """
        if len(s.commandList) == 0:
            return ""
        else:
            idx = s.commandListIdx
            # Use negative indexing.
            if next:
                idx += 1
                if idx >= 0:
                    idx = -1
            else:
                idx -= 1
                if idx < -len(s.commandList):
                    idx = -len(s.commandList)
        command = s.commandList[idx]    # shouldn't throw
        s.commandListIdx = idx          # commit index
        return command

    # INTERFACE
    def Do( s, command, resultsTextWid=None ):
        """ Process command given by user (or internal command). """
        try:
            if not command:
                return

            # s.resultsTextWid is to be used henceforth and in subroutines.
            if resultsTextWid is None:
                s.resultsTextWid = st.gui.GetResultsTextWid()
            else:
                s.resultsTextWid = resultsTextWid
            
            # Remember this command.
            # Even a bad command is remembered so that
            # the user has a chance to the recall and fix it.
            s.commandList.append( command )
            s.commandListIdx = 0     # this resets command history recall

            # Only commandWords[0] should always be present.
            # Accessing commandWords[>0] could cause an exception!
            commandWords = SplitWords( command )    # handles quoted words

            # Goto line number?
            if commandWords[0].isdigit():
                line = int(commandWords[0])  # convert string from user to int
                s.__GotoLine( line )
            # Browse source file?
            elif commandWords[0] in [ "b", "br", "bc" ]:
                s.__BrowseSrcFile( command, commandWords )
            # Find text?
            elif commandWords[0][0] in "fg":
                s.__FindText( command, commandWords )
            # Find tag?
            elif commandWords[0] in [ "t", "tc", "tr", "tcr", "trc", "te" ]:
                s.__FindTag( command, commandWords )
            # Syntax hiliting/highlighting.
            elif commandWords[0] == "hilite":
                hiliteLevel = commandWords[1]
                LexBase.SetHilite( hiliteLevel )  # call Python static method
                st.gui.FlushSrcFiles()
                st.gui.ReloadSrcFiles()
            # Help?
            elif commandWords[0] in [ "h", "help", "?" ]:
                tokenList = [ (eTokenText,helpText) ]
                contents = ContentsResults( tokenList=tokenList, fname="[results: help]", command=command )
                s.resultsTextWid.LoadContents( contents )
            # Log command?
            elif commandWords[0] == "log":
                s.__Log( commandWords )
            # (internal command)
            elif commandWords[0] == s.INTERNAL_COMMAND_CLICK_TAG:
                # Re-click tag for refresh.
                del s.commandList[-1]    # don't store internal commands in history
                clickedTagName = commandWords[1]
                clickedFname   = commandWords[2]
                st.gui.TagClick( clickedTagName, clickedFname )
            #  (internal command)
            elif commandWords[0] == s.INTERNAL_COMMAND_SPLASH:
                # Splash screen shows help instructions.
                del s.commandList[-1]    # don't store internal commands in history
                text = "pypersrc " + st.version + "\n\n" + helpText
                fname = "[results: splash]"
                s.__ShowTempResults( fname, text )
            # Easter egg?
            elif commandWords[0] in [ "easter", "egg", "easteregg", "underdog" ]:
                Underdog()
                s.Do( "log" )
            # Unknown command.
            else:
                st.gui.statusbar.Print( "Unknown or unimplemented command.  Type 'h' for help." )
                return

        except:
            PrintException()

    # PRIVATE
    def __ShowTempResults( s, fname, text ):
        """ Show temp results that won't be saved in history. """
        tokenList = [ (eTokenText,text) ]
        contents = ContentsResults( tokenList=tokenList, fname=fname, command=None )
        contents.saveInHistory = False   # forget
        s.resultsTextWid.LoadContents( contents )

    # PRIVATE
    def __Log( s, commandWords ):
        """ Execute a log command. """
        # commandWords[>0] could be absent and access will cause exception.
        
        # Clear log?
        try:
            if commandWords[1] in [ "clear", "c" ]:
                st.log.Clear()
                s.resultsTextWid.LoadLog()
                return
        except:
            pass

        # Save log to file?
        try:
            if commandWords[0] == "log" and commandWords[1] == "save":
                if len(commandWords) != 3:
                    st.gui.statusbar.Print( "One argument is required." )
                else:
                    try:
                        fname = commandWords[2]
                        fd = os.open( fname, os.O_CREAT|os.O_TRUNC|os.O_WRONLY, 0600 )
                        os.write( fd, st.log.Read() )
                        os.close( fd )
                        text = "Saved log into file " + AbsPath( fname )
                        st.gui.statusbar.Print( text )
                        st.log.Write( text )
                    except:
                        PrintException()
                        text = "Failed to write log to file '" + fname + "'"
                        st.gui.statusbar.Print( text )
                        st.log.Write( text )
                return
        except:
            pass

        # Show log in text widget.
        s.resultsTextWid.LoadLog()
        
    # PRIVATE
    def __GotoLine( s, line ):
        """ Goto line number in current text widget. """
        textWid = st.gui.notebookSrcFile.GetCurrentTextWid()
        if textWid:
            st.log.Write( "Command: goto line " + str(line) + " in " + textWid.GetFname() )
            textWid.GotoLine( line=line, hiliteLine=True )
        
    # PRIVATE
    def __FindText( s, command, commandWords ):
        """ Find text in all source files. """

        # Determine scope of search.
        # Use logical AND with length to ensure element index is valid.
        # Treat "g <text> <file>" as constrained search for convenience.
        searchOne  = (command[0][0] == "f") and (len(commandWords) == 2)
        searchAll  = (command[0][0] == "g" and len(commandWords) == 2) \
                  or (command[0][0] == "f" and len(commandWords) == 3 and commandWords[2] == "*")
        searchSome = not searchAll and len(commandWords) > 2
        if int(searchOne) + int(searchAll) + int(searchSome) != 1:
            raise Oops( "FindText: confused about scope of search!" )

        # Gather srcFileList according to scope of search.
        if searchOne:
            (srcFileList,errorText) = s.__FindTextOne( command, commandWords )
        elif searchAll:
            (srcFileList,errorText) = s.__FindTextAll( command, commandWords )
        elif searchSome:
            (srcFileList,errorText) = s.__FindTextSome( command, commandWords )
        else:
            raise Oops( "FindText: unhandled case!" )

        # srcFileList now should have list of files to search.
        if srcFileList:
            # Search in a second thread as searching can be slow.
            pattern = commandWords[1]
            modifiers = commandWords[0][1:]
            finderText = s.__Finder( pattern, modifiers, findFile=False )
            st.threads.Spawn( ThreadFindText, command, finderText, srcFileList )
        elif errorText:
            st.gui.statusbar.Print( errorText )

    # PRIVATE
    def __FindTextOne( s, command, commandWords ):
        """ Search one file.  Returns (srcFileList,errorText). """
        
        # But is there yet a current text widget (notebook empty)?
        textWid = st.gui.notebookSrcFile.GetCurrentTextWid()
        if textWid:
            fname = textWid.contents.fname
            srcFileList = [ fname ]
            errorText = ""
        else:
            srcFileList = []
            errorText = "Text widget is empty, nothing to find locally."
        return (srcFileList,errorText)

    # PRIVATE
    def __FindTextAll( s, command, commandWords ):
        """ Search all files.  Returns (srcFileList,errorText). """
        if commandWords[0] == "g" and len(commandWords) != 2:
            srcFileList = []
            errorText = "g command: requires one arg."
        else:
            srcFileList = st.gather.GetSrcFileList()
            errorText = ""
        return (srcFileList,errorText)

    # PRIVATE
    def __FindTextSome( s, command, commandWords ):
        """ Search in a constrained subset of src files:
            f <text> [cr] [~]<file>...
            Returns (srcFileList,errorText). """
        
        # Build list of Finder objects (one for each file pattern the user specified).
        finderFileList = []
        modifiersFile = ""     # modifiers for file pattern, in separate word ('~' prefixes the pattern)
        anyNegative = False    # if negative match was specified one or more times
        i = 2                  # index of first file pattern (one or more)
        while i < len(commandWords):
            if len(commandWords[i]) <= 2 and IfHaveEachChar( commandWords[i], "cr" ):
                # This is a pattern modifier (case-sensitive and/or regex).
                modifiersFile = commandWords[i]
            else:
                # This is a pattern for a file.  It could have a negative match prefix ('~').
                patternFile = commandWords[i]
                negative = (patternFile[0] == "~")    # negative match?
                if negative:
                    anyNegative = True
                    patternFile = patternFile[1:]    # chop '~'
                finderFile = s.__Finder( patternFile, modifiersFile, negative=negative, findFile=True )
                finderFileList.append( finderFile )
            i += 1
        if not finderFileList:
            st.gui.statusbar.Print( "f command: unrecognized args, try quoting pattern, or type 'help'." )
            return
        
        # Build list of matching src files.
        srcFileList = []
        for fname in st.gather.GetSrcFileList():
            # Test if this src file matches pattern(s).
            if not anyNegative:
                # Multiple non-negative matches imply logical OR.
                # Eg, "f text *.cc *.py" means C++ and Python files.
                for finderFile in finderFileList:
                    if finderFile.IfMatch( fname ):
                        srcFileList.append( fname )
                        break
            else:
                # Multiple negative matches imply logical AND.
                # Eg "f text ~*.cc ~*.py" means all files excluding C++ and Python.
                mismatch = False
                for finderFile in finderFileList:
                    # This can be confusing: IfMatch() returns True if fname does NOT match
                    # (negative matching is being used), then "not" inverts.
                    if not finderFile.IfMatch( fname ):
                        mismatch = True
                        break
                if not mismatch:    # if all negative matches were true (logical AND)
                    srcFileList.append( fname )
                        
        if srcFileList:
            errorText = ""
        else:
            errorText = "The file pattern doesn't match any filenames (nothing to search)."
        return (srcFileList,errorText)

    # PRIVATE
    def __FindTag( s, command, commandWords ):
        """ Find a source code tag. """
        # "t <name> <type> <file>"  constrain by name, type, file
        # "t <name> * <file>"       constrain by name, file (* means match any and a syntactic placeholder)
        # "t * * <file>"            constrain by file
        # For consistency with the f/g/b commands, the fileglob is the t command's last arg.

        try:
            caseSens = ("c" in commandWords[0])
            regex    = ("r" in commandWords[0])
            exact    = ("e" in commandWords[0])

            tokenList = []
            tagName = commandWords[1]

            # Optionally constrain the search by tag type.
            try:
                tagType = commandWords[2]    # user may omit
                if tagType == "*":           # '*' means any type, used as syntactic placeholder
                    tagType = None           # to constrain by fileglob only
            except:
                tagType = None

            # Optionally constrain the search by tag file (fileglob).
            try:
                tagFile = commandWords[3]    # user may omit
            except:
                tagFile = None

            # To tell user if search was constrained.
            constraints = ""
            if caseSens:
                constraints = AddIntoParentheses( constraints, "case-sensitive" )
            if tagType:
                constraints = AddIntoParentheses( constraints, "type:" + tagType )
            if tagFile:
                constraints = AddIntoParentheses( constraints, "file:" + tagFile )

            # Find tags.
            if exact:
                tagList = st.tags.FindTag( tagName, tagType, tagFile, caseSens=1, partial=0, regex=0 )
            elif regex:
                tagList = st.tags.FindTag( tagName, tagType, tagFile, caseSens,   partial=0, regex=1 )
            else:
                tagList = st.tags.FindTag( tagName, tagType, tagFile, caseSens,   partial=1, regex=0 )

            # Prepare to load text widget with hyperlinks to matching tags.
            if tagList:
                tokenList.insert( 0, (eTokenText,"Found tags matching `" + tagName + "`" + constraints + ":\n\n") )
                for tag in tagList:
                    # Build a phony tag from the real tag whose name is the elements
                    # of the real tag (name,type,file,line).  The reason is that
                    # phony tags, unlike real tags, when put into a token list,
                    # are shown as hyperlinks whose color changes if visited.
                    phonyTagName = tag[0] + " " + tag[1] + " " + tag[2] + " " + str(tag[3])
                    phonyTagFile = tag[2]
                    phonyTagLine = tag[3]
                    phonyTag = st.tags.PhonyTag( phonyTagName, phonyTagFile, phonyTagLine )
                    tokenList.append( (eTokenTag,phonyTag) )
            else:
                tokenList.insert( 0, (eTokenText,"No tags match `" + tagName + "`" + constraints + ".\n\n") )
                # Does this look like a regex pattern but the user didn't specify the "x" suffix?
                if tagName != "*" and (tagName.find( "*" ) != -1 or tagName.find( "." ) != -1 or tagName.find( "[" ) != -1):
                    if not commandWords[0] in [ "tx", "txc", "tcx" ]:
                        tokenList.append( (eTokenText,"What you're searching for looks like a regex pattern.\n") )
                        tokenList.append( (eTokenText,"Use 'tx' rather than the 't' command to specify a regex pattern.\n") )
                        tokenList.append( (eTokenText,"Just press Up in the Command entry and append 'x' to 't'.\n") )

            # Load text widget with hyperlinks to matching tags.
            fname = "[results: " + command + "]"
            contents = ContentsResults( tokenList=tokenList, fname=fname, command=command )
            s.resultsTextWid.LoadContents( contents )
        except:
            PrintException()

    # PRIVATE
    def __BrowseSrcFile( s, command, commandWords ):
        """ Find a source file to browse. """
        try:
            if len(commandWords) != 2:
                st.gui.statusbar.Print( "One argument is required." )
                return
                    
            sought = commandWords[1]
            finder = s.__Finder( sought, commandWords[0][1:], findFile=True )
            fnameResults = "[results: " + command + "]"
            
            # Search listbox for rows that match the sought src file.
            # Build a list of matches (by index).
            rowTextList = st.gui.GetFileList()
            rowIdx = 0
            matchIdxList = []
            for rowText in rowTextList:
                if finder.IfMatch( rowText ):
                    matchIdxList.append( rowIdx )
                rowIdx += 1

            # Found one match?
            if len(matchIdxList) == 1:
                # Auto-select the matching src file.
                st.gui.DeselectFileList( matchIdxList[0] )
                fname = rowTextList[matchIdxList[0]]
                st.gui.notebookSrcFile.LoadSrcFile( fname )
            # Found multiple matches?
            elif len(matchIdxList) > 1:
                # Build results (hyperlinks to matching src files) in text widget
                # to let the user pick one.
                tokenList = []
                tokenList.insert( 0, (eTokenText,"Found multiple source files:\n\n") )
                for idx in matchIdxList:
                    fname = rowTextList[idx]
                    line = 0
                    phonyTag = st.tags.PhonyTag( fname, fname, line )
                    tokenList.append( (eTokenTag,phonyTag) )
                contents = ContentsResults( tokenList=tokenList, fname=fnameResults, command=command )
                s.resultsTextWid.LoadContents( contents )
            # Found no matches?
            else:
                tokenList = []
                text = "No source files match `" + sought + "`"
                tokenList.insert( 0, (eTokenText,text) )
                contents = ContentsResults( tokenList=tokenList, fname=fnameResults, command=command )
                s.resultsTextWid.LoadContents( contents )
        except:
            PrintException()

    # PRIVATE
    def __Finder( s, pattern, modifiers, negative=False, findFile=False ):
        """ Helper to return a Finder object according to match modifier chars. """
        # The user can enter "cre" as match modifier suffixes (in stand-alone word).
        # "~" for a negative match is special as it prefixes the pattern, not in a stand-alone modifier word.
        caseSens = "c" in modifiers    # case-sensitive matching
        regex    = "r" in modifiers    # regex matching
        exact    = "e" in modifiers    # exact matching
        glob     = False
        # File-globbing is implied by the chars in the pattern, a file being sought,
        # and other higher-precedence modifiers being false.
        if findFile and not (exact or regex):
            for c in pattern:
                if c in "*?[]":
                    glob = True
                    break
        try:
            return Finder( pattern, caseSens=caseSens, regex=regex, exact=exact, glob=glob, negative=negative )
        except:
            if regex:
                st.gui.statusbar.Print( "Bad regex.  Try quoting it (?)." )
            raise
# $Id: Gui.py,v 1.115 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# Notes:
# ==============================================================================

# ==============================================================================
# These are base classes that provide an interface independent of any
# GUI toolkit in order to facilitate possibly porting to a different one.
# And they summarize the Python code.
# ------------------------------------------------------------------------------
# Thread-safety:
# These are the only Gui methods that are thread-safe:
# - Statusbar.Print()
# - Statusbar.Clear()
# The initial thread is dedicated to the Gui.
# Other threads are restricted to calling only thread-safe Gui methods.
# ==============================================================================

#-------------------------------------------------------------------------------
# Class GuiBase.
#-------------------------------------------------------------------------------

class GuiBase:
    """ GUI base class. """

    def __init__( s, argList ):
        st.gui = s    # Gui components need st.gui during construction/bringup
        
        # --- may be overridden by cmd-line args ---
        s.geom            = "1000x700"
        if IfCygwin():
            s.fontName    = "courier"
            s.fontSize    = 10
        else:
            s.fontName    = None
            s.fontSize    = None            
        # --- may be overridden by cmd-line args ---
        
        s.ProcessArgsPre( argList )
        
        s.enabled         = False
        s.nestEnable      = Nest( nestFirst=False )  # to nest Gui.Enable(T/F)
        s.mutex           = threading.RLock()        # reentrant lock
        
        s.ProcessArgsPost( argList )
        
    def ProcessArgsPre( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        try:
            i = 0
            while i < len(argList):
                arg = argList[i]
                if arg is None:
                    i += 1
                elif arg == "-geom":
                    s.geom = argList[i+1]
                    argList[i]   = None
                    argList[i+1] = None
                    i += 2
                elif arg in [ "-font", "-font-name", "-fontname" ]:
                    s.fontName = argList[i+1]
                    argList[i]   = None
                    argList[i+1] = None
                    i += 2
                elif arg in [ "-font-size", "-fontsize" ]:
                    s.fontSize = argList[i+1]
                    argList[i]   = None
                    argList[i+1] = None
                    i += 2
                else:
                    i += 1
        except:
            PrintException()
        return True

    def ProcessArgsPost( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True

    def IfReady( s ):
        """ If GUI is ready/running (not to be confused with enabled). """
        raise Oops( "abstract method" )

    def IfEnabled( s ):
        """ If GUI is sensitive to input. """
        return s.enabled

    def Enable( s, enable=True ):
        """ Make GUI sensitive to input. """
        # Derived class should override and call this.
        s.enabled = enable

    def EnableWidget( s, widget, enable=True ):
        """ Enable widget (or disable if enable=False). """
        raise Oops( "abstract method" )
    
    def DisableWidget( s, widget ):
        """ Unconditionally disable widget.
            Returns old enable state of widget. """
        s.EnableWidget( widget, False )

    def EventLoop( s ):
        raise Oops( "abstract method" )
        
    def SetTitle( s, title ):
        """ Set text of title bar of main window. """
        raise Oops( "abstract method" )

    def SetFileList( s, fileList ):
        """ Set listbox with list of source filenames to browse. """
        raise Oops( "abstract method" )

    def GetFileList( s ):
        """ Return list of source filenames. """
        raise Oops( "abstract method" )

    def DeselectFileList( s, idx ):
        """ Deselect row(s) in the listbox of src files. """
        raise Oops( "abstract method" )

    def GetResultsTextWid( s ):
        """ Return the text widget used to hold results. """
        raise Oops( "abstract method" )
    
    def FlushSrcFiles( s, parseModes=None ):
        """ Flush all source files out of cache. """
        # Without args, flushes entire cache.
        # With parseModes=n, flushes src files filtered in specified parse modes.
        raise Oops( "abstract method" )

    def ReloadSrcFiles( s ):
        """ Reload those text widgets that are showing source code. """
        raise Oops( "abstract method" )
        
    def TagClick( s, clickedTagName, clickedFname, currentPage=False ):
        """ The user clicked a hyperlink in a source file.
            Also called for the [click tag] internal command (for refreshing). """
        raise Oops( "abstract method" )
        
    def ProcessPendingEvents( s ):
        """ Process any pending GUI events. """
        raise Oops( "abstract method" )

    def ScheduleIdleFunc( s, func, *args ):
        """ Schedule a function to be called (once) when GUI is idle. """
        raise Oops( "abstract method" )

    def ScheduleTimerFunc( s, milsec, func, *args ):
        """ Schedule a function to be called (once) in the next n milliseconds. """
        raise Oops( "abstract method" )

#-------------------------------------------------------------------------------
# Class StatusbarBase.
#-------------------------------------------------------------------------------

class StatusbarBase:

    def __init__( s, frame ):
        s.updateCallback     = None  # a client of Statusbar can optionally register a callback
        s.updateCallbackArgs = None

    def Print( s, text, timeout=5000, baton=False ):
        """ Print text to statusbar. """
        # This pends a change to statusbar text (rendering the widget is delayed).
        # The message is automatically cleared after a few seconds.
        # Pass timeout=0 to make the message stay indefinitely.
        # Pass baton=True inside a loop to twirl a baton (the loop drives the twirling).
        # -- don't call any GUI method, this can be run by second thread --
        raise Oops( "abstract method" )

    def PrintError( s, text ):
        """ Print error message. """
        raise Oops( "abstract method" )
        
    def PrintWarning( s, text ):
        """ Print warning message. """
        raise Oops( "abstract method" )
        
    def Clear( s ):
        """ Clear statusbar. """
        raise Oops( "abstract method" )

    def SetUpdateCallback( s, callback, *callbackArgs ):
        # This is to help periodically update the statusbar.
        # As the statusbar widget is updated by a timer,
        # SetUpdateFunc() can be called to specify a function
        # to be when the timer occurs and can call Statusbar.Print().
        # Pass None to set no function.
        st.gui.mutex.acquire()
        s.updateCallback     = callback
        s.updateCallbackArgs = callbackArgs
        st.gui.mutex.release()

#-------------------------------------------------------------------------------
# Class ScrollFrameBase.
#-------------------------------------------------------------------------------

class ScrollFrameBase:
    """ Create a frame and scrollbars to contain a widget. """

    def __init__( s, parent, WidgetClass, labelText, **args ):
        raise Oops( "abstract method" )

    def SetLabelText( s, text ):
        """ Set the text of a label widget in a scroll-frame widget. """
        raise Oops( "abstract method" )
    
    def SetLabelColor( s, color ):
        """ Set the color of a label widget in a scroll-frame widget. """
        raise Oops( "abstract method" )

    def GetScrollbarPos( s ):
        """ Returns (x,y) tuple of position of horizontal and vertical scrollbars. """
        raise Oops( "abstract method" )
    
    def SetScrollbarPos( s, xy ):
        """ Set position of horizontal and vertical scrollbars.  xy must be (x,y) tuple. """
        raise Oops( "abstract method" )
        
#-------------------------------------------------------------------------------
# Class NotebookBase.
#-------------------------------------------------------------------------------

class NotebookBase:
    """ Base class for a notebook widget. """
    # Page names are not the same as its tab text,
    # as pypersrc could browse files with the same names but different paths.

    def __init__( s, parentWidget ):
        pass

    def Enable( s, enable ):
        raise Oops( "abstract method" )

    def IfEmpty( s ):
        """ Return true if no pages. """
        raise Oops( "abstract method" )
    
    def AddPage( s, pageName, tabText, append=True ):
        """ Add a new notebook page.  Given page name must be unique.  tabText is for display.
            Returns widget so that caller can pack a child widget into it. """
        raise Oops( "abstract method" )

    def SelectPage( s, pageName ):
        """ Select a page. """
        raise Oops( "abstract method" )
        
    def RemovePage( s, pageName ):
        """ Remove a notebook page. """
        raise Oops( "abstract method" )

    def GetCurrentPageName( s ):
        """ Return name of current page or None if no pages exist. """
        raise Oops( "abstract method" )
        
    def CloseCurrentPage( s ):
        """ Close the current page. """
        raise Oops( "abstract method" )

#-------------------------------------------------------------------------------
# Class NotebookTextWid.
#-------------------------------------------------------------------------------

class NotebookTextWidBase(NotebookBase):

    def __init__( s ):
        pass

    def LoadContents( s, contents, line=0, hiliteLine=False, currentPage=False ):
        """ Load text widget (contained by notebook page) with Contents object. """
        # - line : Lines are counted from 1.  If line > 0, text widget is scrolled to line.
        raise Oops( "abstract method" )
    
    def GetCurrentTextWid( s ):
        """ Return text widget whose page is currently selected. """
        raise Oops( "abstract method" )

    def GetTextWidList( s ):
        """ Return list of text widgets in notebook. """
        raise Oops( "abstract method" )
    
#-------------------------------------------------------------------------------
# Class NotebookSrcFileBase.
#-------------------------------------------------------------------------------

class NotebookSrcFileBase(NotebookTextWidBase):

    def __init__( s ):
        pass
    
    def LoadSrcFile( s, fname, line=0 ):
        """ Load src file into notebook page. """
        raise Oops( "abstract method" )
        
    def Reload( s ):
        """ Reload src files in the notebook. """
        raise Oops( "abstract method" )
# $Id: GuiText.py,v 1.51 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# Notes:
# ==============================================================================

#-------------------------------------------------------------------------------
# These are base and fully-implemented classes, indepedent of any GUI toolkit,
# providing an interface to text widgets and their contents.
#
# fname
# -----
# A fname is associated with and identifies Contents objects and text widgets.
# fname can be a real filename or a phony one such as "[results: ]".
#
# Class Contents:
# ---------------
# The contents and state of a text widget is organized as a class
# so that it can be stored into ResultsHistory and CacheContents.
#
# Class TextWid:
# --------------
# This is derived from ScrollFrame and is very specialized for pypersrc.
# The top one is generally dedicated to holding source code, while the bottom one
# is for being loaded by hyperlinks, results from user commands, showing logs, etc.
#
# Class CacheSrcFile:
# -------------------
# A potentially slow operation is tokenizing a src file (tokens are stored
# as the Contents.tokenList member which is assigned by ContentsSrcFile constructor).
# Since the user can re-open and re-filter the same src files, caching is used
# to avoid retokenizing (file buffers, of course, are already cached by the OS).
# The cache item is keyed by both a fname and parse modes (modes being plural).
#
# Class ResultsHistory:
# ---------------------
# The results of commands are saved/cached in a history.
# Because refreshing can cause results to become stale, ResultsHistory
# has the ability to redo a command to freshen stale results.
# For example, if the user did a global search, then modified a src file,
# the global search command will be redone if the command's results
# are to be pulled from the history.
#
#-------------------------------------------------------------------------------
# Mechanisms:
#
# Caching Contents objects:
# -------------------------
# Contents objects contain a token list.  The process of tokeniziation can be slow.
# Text widgets display Contents objects.  Since Contents objects can migrate
# in and out of text widgets, caching Contents objects speeds reloading text widgets.
# ContentsSrcFile objects are cached by CacheSrcFile (see Class ContentsSrcFile).
# ContentsResults objects are cached by ResultsHistory (see Class ResultsHistory).
#
# Refreshing:
# -----------
# The user is supposed to click the refresh button after changing any source file
# (with an editor, removing, etc).  A goal is to avoid discarding or redoing results
# that are associated with unchanged src files.
#
# Periodic log refreshing:
# ------------------------
# Every second or so, IdleLog() is called.  If the log buffer has grown more than
# the log being shown in the text widget, the difference (the recently logged lines)
# is appended into the results text widget.
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
# Class Contents*.
#-------------------------------------------------------------------------------

class ContentsBase:
    """ The contents and state of a text widget. """

    def __init__( s, tokenList, fname ):
        # The contents stored is a list of tokens derived from the original source.
        s.tokenList = tokenList     # tokens from parsing source code or manufactured for results
        s.fname     = fname         # filename (can be phony)
        s.editable  = False         # whether contents is read-only or editable
        s.error     = False         # error during construction (invalid object)

    def IfEmpty( s ):
        """ Return true if no contents. """
        return not s.tokenList

    def IfError( s ):
        return s.error

class ContentsNone(ContentsBase):
    """ No contents. """
    # Why ContentsNone instead of using Python's None?
    # Client code assumes a TextWid (whether empty or not) has a Contents object.

    def __init__( s, tokenList=[], fname=None ):
        # fname can specified to mean empty file and to be shown in page tab.
        # tokenList can be specified too, but contents is still considered to have nothing important.
        ContentsBase.__init__( s, tokenList=tokenList, fname=fname )

class ContentsSrcFile(ContentsBase):
    """ Contents of a tokenized source file. """

    def __init__( s, fname ):
        ContentsBase.__init__( s, tokenList=[], fname=fname )
        s.fnameSaved = None    # fnameSaved/missing are for if a deleted src file-
        s.missing    = False   # -is reinstated (unusual case)

        # Read src file, tokenize it.
        st.log.Write( "ContentsSrcFile: reading/tokenizing src file " + fname )
        (s.tokenList,ok) = SrcFile2TokenList( fname )
        s.error = not ok
        s.parseModes = st.parser.GetModes()   # used by CacheSrcFile

class ContentsResults(ContentsBase):
    """ Contents from results dynamically computed by pypersrc.
        Appropriate ContentsResults can be updated by refreshing. """

    def __init__( s, tokenList, fname, command ):
        ContentsBase.__init__( s, tokenList, fname )
        s.command       = command    # user command that produced the results (N/A if this is a source file)
        s.refreshCnt    = st.refreshCnt    # to recompute stale results
        s.saveInHistory = True
        
    def UpdateRefreshCnt( s ):
        s.refreshCnt = st.refreshCnt

class ContentsSearch(ContentsResults):
    """ Results of a text search.  Might be affected by refreshing. """
    
    def __init__( s, tokenList, fname, command ):
        ContentsResults.__init__( s, tokenList, fname, command )

class ContentsLog(ContentsResults):
    """ For displaying log.  Automatically updated timer when log is being displayed. """
    
    def __init__( s, tokenList, fname ):
        ContentsResults.__init__( s, tokenList, fname, command="log" )

class ContentsHistory(ContentsResults):
    """ For displaying history of commands. """
    
    def __init__( s, tokenList, fname ):
        ContentsResults.__init__( s, tokenList, fname, command=None )

#-------------------------------------------------------------------------------
# Class TextWidBase.
#-------------------------------------------------------------------------------

class TextWidBase(ScrollFrameBase):
    """ Text widget in a scroll-frame widget. """
    
    def __init__( s ):
        # ContentsNone provides the required empty Contents.tokenList/.fname members so that
        # clearing a text widget can be done by passing ContentsNone to LoadContents().
        s.contents = ContentsNone()

    # INTERFACE
    def Enable( s, enable=True ):
        """ Enable text widget. """
        raise Oops( "abstract method" )

    # INTERFACE
    def IfEmpty( s ):
        """ Return True if text widget has no contents. """
        return isinstance( s.contents, ContentsNone )

    # INTERFACE
    def GetFname( s ):
        """ Return fname that identifies text widget. """
        return s.contents.fname

    # INTERFACE
    def LoadContents( s, contents, line=0, hiliteLine=False, scrollbarPos=None ):
        """ Load text widget with a Contents object. """
        # - contents : To implicitly force a reload, pass a different Contents object.
        # - scrollbarPos : (x,y) tuple (or None) of positions of horizontal,vertical scrollbars
        raise Oops( "abstract method" )
        
    # INTERFACE
    def GotoLine( s, line, hiliteLine=False ):
        raise Oops( "abstract method" )
        
    # INTERFACE
    def Refresh( s, change, addedList, modifiedList, removedList ):
        """ Refresh text widget after source files/dirs have changed. """
        raise Oops( "abstract method" )
        
#-------------------------------------------------------------------------------
# Class CacheSrcFile.
#-------------------------------------------------------------------------------

# Parse modes and filter modes are almost synonymous.
# Src files are filtered by setting the parse modes of Parser object.
    
class CacheSrcFile:
    """ Cache specialized for ContentsSrcFile objects. """
    # Src files can be filtered in different ways (parse modes).
    # Conceptually, ContentsSrcFile objects should be keyed by fname,parseModes.
    # How this is implemented in Python is that every fname is a dict key
    # which maps to a list of ContentsSrcFiles object which have the same fname
    # but have different parse modes.  This mapping scheme allows flushing cache
    # by fnames or parse modes.

    def __init__( s ):
        s.dictContentsList = {}    # fname : [ ContentsSrcFile ]

    # INTERFACE
    def Get( s, fname, parseModes ):
        """ Try to get ContentsSrcFile object keyed by fname,parseModes.
            If not cached, reads src file, caches and returns new ContentsSrcFile. """
        # Is it in the cache?
        (contentsFound,i) = s.__Get( fname, parseModes )
        if contentsFound is not None:
            # In the cache.
            st.log.Write( "CacheSrcFile: found " + s.__LogItem(fname,parseModes) + " in cache" )
            return contentsFound
        else:
            # Not in the cache.  Read src file from filesystem.
            st.log.Write( "CacheSrcFile: " + s.__LogItem(fname,parseModes) + " isn't cached" )
            contents = ContentsSrcFile( fname )
            if contents.IfError():
                raise IOError    # file read error
            else:
                # Store it in the cache.
                s.Put( contents, parseModes )
                return contents

    # PRIVATE
    def __Get( s, fname, parseModes ):
        """ Find ContentsSrcFile object keyed by fname,parseModes.
            Returns (contents,contentsListIndex) or (None,None). """
        if s.dictContentsList.has_key( fname ):
            # Ok, found fname in dict.
            # Now find parseModes in list (the ContentsSrcFile in desired filter mode).
            i = 0
            contentsList = s.dictContentsList[fname]
            for contents in contentsList:
                if parseModes == contents.parseModes:
                    return (contents,i)
                i += 1
        return (None,None)

    # INTERFACE
    def Put( s, contents, parseModes ):
        """ Put a ContentsSrcFile object into cache. """
        # This cache is only for ContentsSrcFiles objects.
        if isinstance(contents,ContentsSrcFile):
            st.log.Write( "CacheSrcFile: caching " + s.__LogItem(contents.fname,parseModes) )
            s.__Put( contents, parseModes )
        elif isinstance(contents,ContentsNone):
            return    # silently ignore, normally happens
        else:
            raise Oops( "CacheSrcFile.Put: wrong Contents object to cache" )

    # PRIVATE
    def __Put( s, contents, parseModes ):
        """ Put a ContentsSrcFile into cache (directly, no checks). """
        fname = contents.fname
        (contentsFound,i) = s.__Get( fname, parseModes )
        if contentsFound:
            # Found.
            # Does cached contents differ from passed contents?
            if contents is contentsFound:
                return                                      # same contents, was already in cache
            else:
                contentsList = s.dictContentsList[fname]    # different contents, update cache
                contentsList[i] = contents                  # replace old contents
                return
        else:
            # Mismatch.
            if s.dictContentsList.has_key( fname ):
                # Partial match:  fname is in the dict but has no matching parse modes.
                # Append this contents to list.
                contentsList = s.dictContentsList[fname] 
                contentsList.append( contents )
                return
            else:
                # Total mismatch.  This fname was never encountered before.
                # Create new contents list for fname.
                s.dictContentsList[fname] = [ contents ]
                return

    # INTERFACE
    def Flush( s, fname=None, parseModes=None ):
        """ Flush by either fname or parseModes.
            Flush entire cache if neither fname nor parseModes. """
        # If cache is empty, nothing to flush.
        if s.dictContentsList:
            if (fname is None) and (parseModes is None):    # "not parseModes" is wrong 
                # Flush everything.                         # as parseModes=0 is possible
                st.log.Write( "CacheSrcFile: flushing entire cache" )
                #del s.dictContentsList    # wrong FYI
                s.dictContentsList = {}
            elif (fname is not None) and (parseModes is None):
                # Flush all ContentsSrcFile objects matching fname (all of its parse modes).
                if s.dictContentsList.has_key( fname ):
                    st.log.Write( "CacheSrcFile: flushing " + fname + " (all modes)" )
                    del s.dictContentsList[fname]
            elif (fname is None) and (parseModes is not None):
                # Flush all ContentsSrcFile objects matching parse modes (any fname).
                flushed = False
                for contentsList in s.dictContentsList.values():
                    i = 0
                    for contents in contentsList:
                        if contents.parseModes == parseModes:
                            del contentsList[i]   # ok as break follows deletion
                            flushed = True
                            break
                        i += 1
                if flushed:
                    st.log.Write( "CacheSrcFile: flushed src files in " + st.parser.GetModesString(parseModes) )
            else:
                raise Oops( "CacheSrcFile.Flush: unimplemented case" )    # no need for this case

    # PRIVATE
    def __LogItem( s, fname, parseModes ):
        """ Helper to abbrev logging.  Returns string. """
        return fname + " " + st.parser.GetModesString(parseModes)

#-------------------------------------------------------------------------------
# Class ResultsHistory.
#-------------------------------------------------------------------------------

# textWid args should be a TextWidResults object (not the base TextWid).

class ResultsHistory:
    """ Maintains a history of results. """

    def __init__( s ):
        s.contentsList        = []     # maintains order
        s.contentsDict        = {}     # fname : Contents
        s.scrollbarPosDict    = {}     # fname : scrollbar position (x,y) tuple
        s.resultsHistoryFname = "[results: history of command results]"

    # INTERFACE
    def Save( s, textWid ):
        """ Store the results (being displayed in text widget) into history. """
        # Handles duplicates (because of how Show() works).
        # The command history itself shouldn't be saved.
        # Yes, ContentsLog should be saved in history.
        contents = textWid.contents
        if isinstance(contents,ContentsResults) and not isinstance(contents,ContentsHistory) and contents.saveInHistory:
            # Save contents and scrollbar position.
            # Don't save duplicates.
            fname = contents.fname
            if not s.contentsDict.has_key( fname ):
                s.contentsList.append( contents )
                s.contentsDict[fname] = contents
            scrollbarPos = textWid.GetScrollbarPos()  # returns (x,y) tuple
            s.scrollbarPosDict[fname] = scrollbarPos

    # INTERFACE
    def Load( s, textWid, fname ):
        """ Load text widget with results from history. """
        # Has results become stale because of recent refreshing?
        contents = s.contentsDict[fname]
        if contents.refreshCnt < st.refreshCnt and contents.command:
            # Redo command to refresh results.
            contents.refreshCnt = st.refreshCnt
            st.log.Write( "ResultsHistory: stale results, redoing command '" + contents.command + "'" )
            st.command.Do( contents.command )
            # Command.Do() has again loaded the results text widget.
        else:
            # Load text widget with saved results, restore scrollbar position.
            scrollbarPos = s.scrollbarPosDict[fname]
            textWid.LoadContents( contents=contents, scrollbarPos=scrollbarPos )

    # INTERFACE
    def Show( s, textWid ):
        """ Show the history of command results. """
        # The results history is shown as hyperlinked fnames to saved results.
        # The most recent results is shown at the top.

        # Save the current results right now to show it also.
        s.Save( textWid )

        # Build fname list.
        fnameList = []
        for contents in s.contentsList:
            fnameList.insert( 0, contents.fname )

        # Build token list from fname list.
        tokenList = []
        tokenList.append( (eTokenText,"History of command results:\n\n") )
        for fname in fnameList:
            phonyTag = st.tags.PhonyTag( fname, fname, lineNum=0 )    # hyperlink
            tokenList.append( (eTokenTag,phonyTag) )

        # Build contents from token list, then load it into the text widget.
        # Indeed, ResultsHistory.Save() will be called twice as TextWidResults.LoadContents()
        # also calls it.  But Save() handles being passed duplicate results.
        contents = ContentsHistory( tokenList=tokenList, fname=s.resultsHistoryFname )
        textWid.LoadContents( contents=contents )

    # INTERFACE
    def ShowLast( s, textWid ):
        """ Show the results from the last command. """
        # textWid is the text widget to show results in.
        if s.contentsList:
            fname = s.contentsList[-1].fname
            s.Load( textWid, fname )

    # INTERFACE
    def Refresh( s, change, addedList, modifiedList, removedList ):
        """ Refresh history after source files have changed. """
        # This has become a NOP as ResultsHistory.Load() refreshes stale results
        # (this is kept for logical completeness in HandlerButtonRefresh()).
        pass
# $Id: GuiTk.py,v 1.50 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# Notes:    On cygwin, event.keysym isn't assigned correctly for Fn keys,
#           so they're bound to individual handlers.
# ==============================================================================

#-------------------------------------------------------------------------------
# Load-time checks.
#-------------------------------------------------------------------------------

# Even if configure.sh approved, the environment or Python installation could change.

if IfCygwin() and not os.environ.has_key("TCL_LIBRARY"):
    sys.stderr.write( "###\n" )
    sys.stderr.write( "### TCL_LIBRARY env var must be set, for example: \n" )
    sys.stderr.write( "### export TCL_LIBRARY=c:/cygwin/usr/share/tcl8.4 \n" )
    sys.stderr.write( "###\n" )

# Although Tkinter is supposedly a standard part of Python, in practice,
# not all systems have Tk installed (nor Tkinter consequently).
# This is necessary to print a meaningful error for the user.
try:
    import Tkinter
except:
    sys.stderr.write( "### pypersrc depends on Tk/Tkinter which doesn't seem to be installed. \n" )
    sys.stderr.write( "### Or, less likely, the pypersrc executable can't be loaded because \n" )
    sys.stderr.write( "### of a shared lib dependency. \n" )
    if IfDarwin():
        sys.stderr.write( "### If this is Mac OS X older than 10.4 Tiger, try installing \n" )
        sys.stderr.write( "### Fink's version of Python. \n" )
        sys.stderr.write( "### http://fink.sourceforge.net \n" )
TK = Tkinter    # abbrev
import tkFont

# Pmw is an required site-specific module based on Tkinter.
try:
    import Pmw
except:
    sys.stderr.write( "### pypersrc depends on Python Mega Widgets (Pmw). \n" )
    sys.stderr.write( "### Download: http://pmw.sourceforge.net/ \n" )
    sys.stderr.write( "### Pmw needs to be installed in Python's site-packages/ subdir. \n" )
    sys.stderr.write( "### If your system has multiple installations of Python, \n" )
    sys.stderr.write( "### install Pmw at which one pypersrc is linked with. \n" )
    if IfDarwin():
        sys.stderr.write( "### On Mac OS X, pypersrc is linked with Apple's version of Python \n" )
        sys.stderr.write( "### which is at /usr.  Fink's Pmw package, if installed, is at /sw \n" )
        sys.stderr.write( "### ln -s /sw/lib/python2.3/site-packages/Pmw /usr/lib/python2.3/site-packages/Pmw \n" )

#-------------------------------------------------------------------------------
# Class Gui.
#-------------------------------------------------------------------------------

class Gui(GuiBase):
    """ pypersrc Tkinter GUI. """

    def __init__( s, argList ):
        GuiBase.__init__( s, argList )

        s.ready           = False
        s.beenEnabled     = False
        s.keyEventTime    = 0.0         # to mitigate fast keyboard repeat rate
        
        s.TKTAG_DEFAULT   = "df"
        s.TKTAG_TEXT      = "tx"
        s.TKTAG_NUM       = "nu"
        s.TKTAG_SYMBOL    = "sy"
        s.TKTAG_WORD      = "wo"
        s.TKTAG_KEYWORD   = "kw"
        s.TKTAG_TYPE      = "ty"
        s.TKTAG_LOUD      = "lo"
        s.TKTAG_QUOTED    = "qu"
        s.TKTAG_COMMENT   = "co"
        s.TKTAG_TAGDEF    = "td"
        s.TKTAG_TAGREF    = "tr"
        s.TKTAG_TAGREFV   = "trv"  # visited hyperlink
        
        # Lexical tokens are drawn with Tk text widget tags
        # (not to be confused with source code tags).
        s.tktagDict = { eTokenInvalid     : s.TKTAG_DEFAULT,
                        eTokenWs          : s.TKTAG_DEFAULT,
                        eTokenSymbol      : s.TKTAG_SYMBOL,
                        eTokenNum         : s.TKTAG_NUM,
                        eTokenWord        : s.TKTAG_WORD,
                        eTokenKeyword     : s.TKTAG_KEYWORD,
                        eTokenKeywordLoud : s.TKTAG_LOUD,
                        eTokenType        : s.TKTAG_TYPE,
                        eTokenText        : s.TKTAG_TEXT,
                        eTokenTag         : s.TKTAG_TAGDEF,
                        #eTokenTag        : s.TKTAG_TAGREF,  # special case
                        #eTokenTag        : s.TKTAG_TAGREFV, # special case
                        eTokenQuoted      : s.TKTAG_QUOTED,
                        eTokenComment     : s.TKTAG_COMMENT,
                        }
        
        # Set main window geometry.
        s.root = TK.Tk()
        Pmw.initialise( s.root )
        s.root.geometry( s.geom )

        # These handlers must be installed right now.
        # Other handlers are installed by InstallHandlers() when GUI becomes enabled.
        # -- TextWid.__HandlerKey() handles keyboard events in text widgets first --
        s.root.bind( "<KeyPress-Escape>",   s.__HandlerKeyPress )
        s.root.bind( "<KeyRelease-Escape>", s.__HandlerKeyRelease )
        s.root.bind( "<Destroy>", s.__HandlerProgramExit ) # this intercepts ALL destroy events!

        # Create the main frame.
        s.mainFrame = TK.Frame( s.root )
        s.mainFrame.pack( expand=TK.YES, fill=TK.BOTH )

        s.SetTitle( "pypersrc", verbatim=True )

        # Create the statusbar.
        s.statusbar = Statusbar( s.mainFrame )

        # Create Command entry and Filter checkbuttons at top in a frame.
        s.entryFilterFrame = TK.Frame( s.mainFrame, relief=TK.RAISED, borderwidth=2 )
        s.entryFilterFrame.pack( side=TK.TOP, expand=TK.NO, fill=TK.BOTH, anchor=TK.W )

        s.entryFrame = TK.Frame( s.entryFilterFrame )
        s.entryFrame.pack( side=TK.LEFT, expand=TK.YES, fill=TK.BOTH, anchor=TK.W )

        s.filterFrame = TK.Frame( s.entryFilterFrame )
        s.filterFrame.pack( side=TK.RIGHT, anchor=TK.E )

        s.labelCommand = TK.Label( s.entryFrame )
        s.labelCommand.config( text="Command: " )
        s.labelCommand.pack( side=TK.LEFT, anchor=TK.W )
        
        s.colorFocusOut = s.labelCommand["bg"]
        s.colorFocusIn  = "#a0a0f0"

        s.entryCommand = TK.Entry( s.entryFrame )
        s.entryCommand.pack( side=TK.LEFT, anchor=TK.W )
        st.gui.DisableWidget( s.entryCommand )

        s.labelFilter = TK.Label( s.filterFrame )
        s.labelFilter.config( text="Filter: " )
        s.labelFilter.pack( side=TK.LEFT, anchor=TK.E )

        s.__ResetFilterButtons()
        
        s.checkButtonComments  = TK.Checkbutton( s.filterFrame, text="comments",  command=s.__HandlerFilterComments )
        s.checkButtonFunctions = TK.Checkbutton( s.filterFrame, text="functions", command=s.__HandlerFilterFunctions )
        s.checkButtonData      = TK.Checkbutton( s.filterFrame, text="data",      command=s.__HandlerFilterData )
        
        st.gui.DisableWidget( s.checkButtonComments )
        st.gui.DisableWidget( s.checkButtonFunctions )
        st.gui.DisableWidget( s.checkButtonData )
        
        s.checkButtonComments.pack(  side=TK.LEFT, anchor=TK.E )
        s.checkButtonFunctions.pack( side=TK.LEFT, anchor=TK.E )
        s.checkButtonData.pack( side=TK.LEFT, anchor=TK.E )

        s.buttonRefresh = TK.Button( s.filterFrame, text="refresh", command=s.__HandlerButtonRefresh )
        s.buttonRefresh.pack( side=TK.LEFT, anchor=TK.E )
        st.gui.DisableWidget( s.buttonRefresh )

        # Yes, the Close (Page) button is separate from the notebook widget,
        # as here is the right time to pack the button.
        # Consequently, the Gui class, rather than the Notebook class, handles this button
        s.buttonClosePage = TK.Button( s.filterFrame, text="close", command=s.__HandlerButtonClosePage )
        s.buttonClosePage.pack( side=TK.LEFT, anchor=TK.E )
        st.gui.DisableWidget( s.buttonClosePage )

        # Create horizontal paned widget for listbox and vertical paned widget.
        panedArgs = { "separatorthickness":4, "separatorrelief":"raised", "handlesize":12 }
        s.panedHorz = Pmw.PanedWidget( s.mainFrame, orient="horizontal", **panedArgs )
        s.panedHorz.pack( expand=1, fill="both" )
        s.panedHorz1 = s.panedHorz.add( "listbox",   size=0.20 )
        s.panedHorz2 = s.panedHorz.add( "panedVert", size=0.80 )
        s.panedHorz1.pack()
        s.panedHorz2.pack()

        # Create vertical paned widget for both text widgets.
        s.panedVert = Pmw.PanedWidget( s.panedHorz2, orient="vertical", **panedArgs )
        s.panedVert.pack( expand=1, fill="both" )
        s.panedVert1 = s.panedVert.add( "text1", size=0.75 )
        s.panedVert2 = s.panedVert.add( "text2", size=0.25 )
        s.panedVert1.pack()
        s.panedVert2.pack()

        # Create source file listbox.
        s.fileWid = ScrollFrame( s.panedHorz1, TK.Listbox, "", selectmode=TK.SINGLE )
        s.fileWid.frame.pack( side=TK.BOTTOM, expand=TK.YES, fill=TK.BOTH )

        # Create notebook to hold text widgets (tabbed browsing of src code).
        s.notebookSrcFile = NotebookSrcFile( s.panedVert1 )
        
        # Create a text widget to show results.
        # -- clients should access this via GetResultsTextWid() --
        s.resultsTextWid = TextWidResults( s.panedVert2 )

        # Create a cache to speed reloading text widgets.
        s.cacheSrcFile = CacheSrcFile()

        # Create history of results.
        s.resultsHistory = ResultsHistory()

        s.ProcessArgsPost( argList )

        # Let's rock.
        s.ready = True
        
    #-------------------------------------------------------------------------------
    # Interface methods.
    #-------------------------------------------------------------------------------

    # INTERFACE
    def IfReady( s ):
        return s.ready

    # INTERFACE
    def Enable( s, enable=True ):
        """ Make GUI sensitive to input. """

        # Allow nesting Gui.Enable(True/False).
        # Only the outer level has an effect.
        if not s.nestEnable.IfProceed( enable ):
            return

        # Proceed (now at outer-level of nesting).
        GuiBase.Enable( s, enable )  # sets s.enabled

        # Install handlers only once.
        if enable and not s.beenEnabled:
            s.beenEnabled = True
            s.__InstallHandlers()

        # Enable all the widgets.
        st.gui.EnableWidget( s.labelCommand, enable )
        st.gui.EnableWidget( s.entryCommand, enable )
        st.gui.EnableWidget( s.labelFilter, enable )
        st.gui.EnableWidget( s.checkButtonComments, enable )
        st.gui.EnableWidget( s.checkButtonFunctions, enable )
        st.gui.EnableWidget( s.checkButtonData, enable )
        st.gui.EnableWidget( s.buttonRefresh, enable )
        s.fileWid.Enable( enable )
        s.resultsTextWid.Enable( enable )
        s.notebookSrcFile.Enable( enable )
        if not st.gui.notebookSrcFile.IfEmpty():
            st.gui.EnableWidget( s.buttonClosePage, enable ) # Gui class manages this button
        else:
            st.gui.DisableWidget( s.buttonClosePage ) # should be disabled if no pages
        #if enable:
        #    s.entryCommand.focus()

        # To take effect.
        st.gui.ProcessPendingEvents()

    # INTERFACE
    def EnableWidget( s, widget, enable=True ):
        """ Enable widget (or disable if enable=False). """
        # Should apply to Pmw widgets as they're based on Tk widgets.
        # FYI: This was supposed to return the old Tk widget state as bool,
        #      but determining that proved unreliable as some widgets have
        #      multiple enable states and some widgets don't implement querying "state".

        # Convert new enablement to a Tk "state" config arg.
        if enable:
            stateNew = TK.NORMAL
        else:
            stateNew = TK.DISABLED
        widget.config( state=stateNew )

    # INTERFACE
    def EventLoop( s ):
        s.root.mainloop()

    # INTERFACE
    def SetTitle( s, title, verbatim=False ):
        """ Set text of title bar of main window. """
        if verbatim:
            pre = ""
        else:
            pre = "pypersrc  "
        s.root.title( pre + title )

    # INTERFACE
    def SetFileList( s, fileList ):
        """ Set listbox with list of source filenames to browse. """

        listbox = s.fileWid.widget

        # Listbox must be enabled or it'll do nothing.
        enable = st.gui.__EnableWidgetIfNotGui( listbox, True )

        listbox.delete( "0", TK.END )              # clear listbox
        pos = 0
        for fname in fileList:                     # append filenames into listbox
            listbox.insert( pos, fname )
            pos += 1
        
        st.gui.__EnableWidgetIfNotGui( listbox, enable )  # restore enablement

    # INTERFACE
    def GetFileList( s ):
        """ Return list of source filenames. """
        return s.fileWid.widget.get( 0, TK.END )
    
    # INTERFACE
    def DeselectFileList( s, idx ):
        """ Deselect row(s) in the listbox of src files. """
        s.fileWid.widget.selection_clear( 0, TK.END )
    
    # INTERFACE
    def GetResultsTextWid( s ):
        """ Return the text widget used to hold results. """
        return s.resultsTextWid

    # INTERFACE
    def FlushSrcFiles( s, parseModes=None ):
        """ Flush out all source files from cache. """
        # Without args, flushes entire cache.
        # With parseModes=n, flushes src files filtered in specified parse modes.
        s.cacheSrcFile.Flush( parseModes=parseModes )

    # INTERFACE
    def ReloadSrcFiles( s ):
        """ Reload those text widgets that are showing source code. """
        s.notebookSrcFile.Reload()

    # INTERFACE
    def TagClick( s, clickedTagName, clickedFname, currentPage=False ):
        """ The user clicked a hyperlink in a text widget.
            Also called for the [click tag] internal command (for refreshing). """
        try:
            (goodTagList,badTagList) = st.tags.ChooseTag( clickedTagName, clickedFname )
            if not goodTagList and not badTagList:
                s.statusbar.Print( "Can't find the tag '" + clickedTagName + "'" )
            elif len(goodTagList) == 1 and not badTagList:
                # Only one tag has this name
                # Load the source file at the tagged line into the text widget.
                fname = goodTagList[0][TAG_FILE]
                line  = goodTagList[0][TAG_LINE]
                st.gui.ScheduleIdleFunc( s.notebookSrcFile.LoadSrcFile, fname, line, True, currentPage )
            else:
                # Multiple tags have the same name.

                # If a tag was clicked while filtering functions,
                # automatically choose the tag that is in the module being viewed.

                foundTheBestTag = False

                if st.parser.IfFiltering():
                    for tag in goodTagList:
                        if (st.gui.notebookSrcFile.GetCurrentTextWid()) \
                       and (st.gui.notebookSrcFile.GetCurrentTextWid().fname == tag[TAG_FILE]):
                            foundTheBestTag = True
                            fname = tag[TAG_FILE]
                            line  = tag[TAG_LINE]
                            st.gui.ScheduleIdleFunc( s.notebookSrcFile.LoadSrcFile, fname, line )
                        
                if not foundTheBestTag:
                    # Let the user pick the tag.
                    # Show a list hyperlinks to the tags in the results widget,
                    # and try to put the best matching hyperlink at the top.
                    tokenList = []
                    if goodTagList:
                        tokenList.append( (eTokenText,"Found tags with the same name, best match is first:\n\n") )
                    for tag in goodTagList:
                        text = tag[TAG_NAME] + " " + tag[TAG_TYPE] + " " + tag[TAG_FILE] + " " + str(tag[TAG_LINE])
                        phonyTag = st.tags.PhonyTag( text, tag[TAG_FILE], tag[TAG_LINE] )
                        tokenList.append( (eTokenTag,phonyTag) )
                    if badTagList:
                        if not goodTagList:
                            tokenList.append( (eTokenText,"Found tags with the same name, unlikely matches: \n\n") )
                        else:
                            tokenList.append( (eTokenText,"Unlikely matches: \n\n") )
                        for tag in badTagList:
                            text = tag[TAG_NAME] + " " + tag[TAG_TYPE] + " " + tag[TAG_FILE] + " " + str(tag[TAG_LINE])
                            phonyTag = st.tags.PhonyTag( text, tag[TAG_FILE], tag[TAG_LINE] )
                            tokenList.append( (eTokenTag,phonyTag) )
                    fname = "[results: click tag '" + clickedTagName + "' in " + clickedFname + "]"
                    # Compose internal command it needs to be run to refresh this later.
                    command = Command.INTERNAL_COMMAND_CLICK_TAG + " " + clickedTagName + " " + clickedFname
                    contents = ContentsResults( tokenList=tokenList, fname=fname, command=command )
                    s.GetResultsTextWid().LoadContents( contents )
        except:
            PrintException()

    # INTERFACE
    def ProcessPendingEvents( s ):
        """ Process any pending Tk events. """
        # Prevent Tk exception caused by a pending Quit event (program exit).
        if not st.exiting:
            s.root.update()
        if not st.exiting:
            s.root.update_idletasks()   # prevent Tk exception at this

    # INTERFACE
    def ScheduleIdleFunc( s, func, *args ):
        """ Schedule a function to be called (once) when GUI is idle. """
        s.root.after_idle( func, *args )

    # INTERFACE
    def ScheduleTimerFunc( s, milsec, func, *args ):
        """ Schedule a function to be called (once) in the next n milliseconds. """
        s.root.after( milsec, func, *args )

    #-------------------------------------------------------------------------------
    # Misc private methods.
    #-------------------------------------------------------------------------------

    # PRIVATE
    def __EnableWidgetIfNotGui( s, widget, enable=True ):
        """ If GUI is disabled, enable or disable widget.
            If GUI is enabled, NOP.
            Returns old enable state of widget. """
        # Tk-specific.  The purpose is to temporarily enable some widgets
        # while the entire GUI is supposed to be disabled, as otherwise
        # some widgets can't be changed when disabled.
        stateOld = widget.cget( "state" )
        if not s.IfEnabled():
            s.EnableWidget( widget, enable )  # enable or disable
        return stateOld

    # PRIVATE
    def __ResetFilterButtons( s ):
        s.checkButtonFlagComments  = False
        s.checkButtonFlagFunctions = False
        s.checkButtonFlagData      = False

    #-------------------------------------------------------------------------------
    # Event handlers.
    #-------------------------------------------------------------------------------

    # PRIVATE
    def __InstallHandlers( s ):
        s.beenEnabled = True
        # -- TextWid.__HandlerKey() handles keyboard events in text widgets first --
        s.root.bind( "<F1>",                 s.__HandlerKeyPress )
        s.root.bind( "<F2>",                 s.__HandlerKeyPress )
        s.root.bind( "<Control-w>",          s.__HandlerKeyPress )
       #s.fileWid.widget.bind( "<Button-1>", s.__HandlerFileListbox ) # doesn't work right
        s.fileWid.widget.bind( "<Double-1>", s.__HandlerFileListbox )
        s.entryCommand.bind( "<FocusIn>",    s.__HandlerFocusIn )
        s.entryCommand.bind( "<FocusOut>",   s.__HandlerFocusOut )
        s.entryCommand.bind( "<Return>",     s.__HandlerEntryCommand )
        s.entryCommand.bind( "<Up>",         s.__HandlerEntryKey )
        s.entryCommand.bind( "<Down>",       s.__HandlerEntryKey )

    # PRIVATE
    def __HandlerButtonRefresh( s ):
        """ Handle Refresh button.  Refresh according to which source files changed. """
        st.threads.Spawn( ThreadRefresh )
        
    # PRIVATE
    def __HandlerButtonClosePage( s ):
        """ Handle Close (Page) button.  Close current page in notebook """
        s.notebookSrcFile.CloseCurrentPage()
    
    # PRIVATE
    def __HandlerFilterSub( s, eParse, checkButton ):
        """ Subroutine to handle clicking a Filter checkbutton. """

        # Change parser mode (the parser does the filtering).
        if checkButton:
            st.parser.AddMode( eParse )
        else:
            st.parser.RemoveMode( eParse )
        st.log.Write( "filtering " + st.parser.GetModesString() )

        # Skip re-parsing/loading if notebook is empty.
        if not st.gui.notebookSrcFile.IfEmpty():
            st.gui.Enable( False )
            try:
                st.gui.statusbar.Print( "Filtering..." )
                # ReloadSrcFiles() leads to asking CacheSrcFile for
                # a ContentsSrcFile matching both fname and parserModes.
                # If not yet cached, the src file is read, tokenized, and cached.
                s.ReloadSrcFiles()
            except:
                PrintException()
            st.gui.Enable( True )

    # PRIVATE
    def __HandlerFilterComments( s ):
        """ Handle clicking a Filter checkbutton. """
        s.checkButtonFlagComments = not s.checkButtonFlagComments
        s.__HandlerFilterSub( eParseComments, s.checkButtonFlagComments )

    # PRIVATE
    def __HandlerFilterFunctions( s ):
        """ Handle clicking a Filter checkbutton. """
        s.checkButtonFlagFunctions = not s.checkButtonFlagFunctions
        s.__HandlerFilterSub( eParseFunctions, s.checkButtonFlagFunctions )

    # PRIVATE
    def __HandlerFilterData( s ):
        """ Handle clicking a Filter checkbutton. """
        s.checkButtonFlagData = not s.checkButtonFlagData
        s.__HandlerFilterSub( eParseData, s.checkButtonFlagData )

    # PRIVATE
    def __HandlerFileListbox( s, event ):
        """ Handle selection of a source file in listbox. """
        # Get item select in file listbox.
        try:
            fname = event.widget.get( "active" )
            st.gui.ScheduleIdleFunc( s.notebookSrcFile.LoadSrcFile, fname, 1 )
        except:
            PrintException()

    # PRIVATE
    def __HandlerFocusIn( s, event ):
        """ Handler for when a widget receives the focus. """
        try:
            if event.widget is s.entryCommand:
                s.labelCommand.config( bg=s.colorFocusIn )
        except:
            PrintException()

    # PRIVATE
    def __HandlerFocusOut( s, event ):
        """ Handler for when a widget loses the focus. """
        try:
            if event.widget is s.entryCommand:
                s.labelCommand.config( bg=s.colorFocusOut )
        except:
            PrintException()

    # PRIVATE
    def __HandlerEntryCommand( s, event ):
        """ Handle a command given by the user in entry widget. """
        try:
            command = s.entryCommand.get()
            s.entryCommand.delete( "0", TK.END )
            st.command.Do( command )
        except:
            PrintException()

    # PRIVATE
    def __HandlerProgramExit( s, event ):
        """ Handle a Tk Destroy (program exit) event. """
        # NOTE: All destroyed widgets will hit this point!
        # Therefore, ignore all of them except the root widget.
        if event.widget is st.gui.root:
            st.exiting = True

    # PRIVATE
    def __HandlerKeyPress( s, event ):
        """ Handle a key press in the main window (any window). """
        # -- TextWid.__HandlerKey() handles keyboard events in text widgets first --
        # This works only if pending Tk events are processed frequently enough.

        # Ignore overly rapid key events caused by keyboard repeat.
        if s.__IgnoreKeyEvent():
            return

        # Esc
        if event.keysym == "Escape":
            st.escKey = True
        # Ctrl-W
        elif event.keysym == "w":
            s.notebookSrcFile.CloseCurrentPage()
        # F1
        elif event.keysym == "F1":
            s.resultsHistory.Show( s.GetResultsTextWid() )
        # F2
        elif event.keysym == "F2":
            s.resultsHistory.ShowLast( s.GetResultsTextWid() )
        else:
            PrintWarning( "Gui: unhandled key press" )
                
    # PRIVATE
    def __HandlerKeyRelease( s, event ):
        """ Handle a key press in the main window (any window). """
        # -- TextWid.__HandlerKey() handles keyboard events in text widgets first --
        # This works only if pending Tk events are processed frequently enough.

        # Ignore overly rapid key events caused by keyboard repeat.
        if s.__IgnoreKeyEvent():
            return

        # Esc
        if event.keysym == "Escape":
            st.escKey = False

    # PRIVATE
    def __HandlerEntryKey( s, event ):
        """ Handle a key press in the command entry widget. """
        # Insert prior or next command in command history
        # back into the command entry widget.

        # (don't call __IgnoreKeyEvent(), let user type at a fast speed)
        
        try:
            if event.keysym in [ "Up", "Down" ]:
                if event.keysym == "Up":
                    command = st.command.Recall( prev=True )
                else:
                    command = st.command.Recall( next=True )
                if command:
                    s.entryCommand.delete( "0", TK.END )
                    s.entryCommand.insert( TK.END, command )
        except:
            PrintException()

    # PRIVATE
    def __IgnoreKeyEvent( s ):
        """ Return False which means keyboard events are happening too fast. """
        now  = time.time()
        last = s.keyEventTime
        if now - last < 1.0:            # one second delay between responses
            return True
        else:
            st.keyEventTime = now       # remember time of this allowed key event
            return False
                
#-------------------------------------------------------------------------------
# Class Statusbar.
#-------------------------------------------------------------------------------

# Updating of statusbar and threads:
# ----------------------------------
# The statusbar (as all widgets are) is rendered by the main thread.
# Long operations that run in the main thread will block updating the statusbar.
# If a long operation can be executed by another thread, the thread is allowed
# to call Statusbar.Print() which only changes its underlying string that
# is serialized by a mutex.  An idle time, run by the main thread, will update
# the statusbar widget with the changed string (statusbar text).
#
# Showing error/warnings on statusbar:
# ------------------------------------
# The statusbar is implemented as two Tk label widgets.
# Normal text is shown in one, and errors and warnings both are shown in the other.
# The Print*() methods (as explained above) just change a string
# and a timer function periodically updates the label widgets.
# Normal statusbar text times-out after a few seconds,
# but errors/warnings persist for a minute.

class Statusbar(StatusbarBase):

    # Thread-safe methods: (see Gui.py)

    def __init__( s, frame ):
        # Statusbar widget is lazy-updated by a timer.
        # Tk operations are slow.  Therefore, most calls to Statusbar.Print()
        # only affects a Python string.  Only once every timer is the Tk widget rendered.

        StatusbarBase.__init__( s, frame )

        s.text    = ""           # statusbar text
        s.textError = ""         # error/warning text
        s.textErrorPrev = ""     # previous
        s.textErrorPadding = ""
        if IfDarwin():           # Mac OS X: prevent bottom-right widget obscuring error
            s.textErrorPadding = "    "
        s.timeout = 0            # timeout of statusbar text
        s.timeoutError = 0       # timeout for error/warning text
        s.milsec  = 250          # frequency of timer
        s.changed = False        # set when Print() is called, cleared by timer
        s.batonChars = "-\|/"
        s.batonIdx = 0

        # Tk specific.  Don't confuse frame/s.frame.
        # Subtle Tk behavior: By creating the error label before the normal label,
        # if their text overlaps, error text will overlay normal text (as desired).
        s.frame = TK.Frame( frame, relief=TK.SUNKEN, borderwidth=2 )
        s.frame.pack( side=TK.BOTTOM, expand=TK.NO, fill=TK.BOTH )
        s.labelError = TK.Label( s.frame )
        s.labelError.pack( side=TK.RIGHT, expand=TK.NO, fill=TK.BOTH, anchor=TK.E )
        s.labelErrorColorOff = s.labelError["bg"]
        s.labelErrorColorOn  = "#d00000"
        s.label = TK.Label( s.frame )
        s.label.pack( side=TK.LEFT, expand=TK.NO, fill=TK.BOTH, anchor=TK.W )

        s.__RechargeTimer()

    # INTERFACE
    def Clear( s ):
        # This pends a change to statusbar text (rendering the widget is delayed).
        # -- don't call any GUI method, this can be run by second thread --
        try:
            st.gui.mutex.acquire()
            s.__Clear()
        finally:
            st.gui.mutex.release()

    # PRIVATE
    def __Clear( s ):
        # Clear statusbar text (not the actual widget itself).
        # -- don't call any GUI method, this can be run by second thread --
        s.changed = True
        s.text = ""

    # INTERFACE
    def Print( s, text, timeout=7000, baton=False, now=False ):
        """ Print text to statusbar. """
        # (see base)
        try:
            st.gui.mutex.acquire()
            # Allow only the main thread to execute in the GUI toolkit.
            if now and st.IfMainThread():
                # -- is NOT thread-safe: executes in GUI toolkit --
                s.text = Chomp(text)
                s.timeout = timeout
                s.__ChangeWidget( text )    # limited to main thread
                st.gui.ProcessPendingEvents()
            else:
                # -- is thread-safe --
                # A zero timeout arg means no timeout: change it to a a huge value.
                if timeout == 0:
                    timeout = 100*1000*1000
                s.text = Chomp(text)
                s.timeout = timeout
                s.changed = True
                if baton:
                    s.text += " " + s.batonChars[s.batonIdx]
                    s.batonIdx = (s.batonIdx + 1) % len(s.batonChars)
        finally:
            st.gui.mutex.release()

    # PRIVATE
    def __PrintError( s, text ):
        # Timer function will update the label widget.
        try:
            st.gui.mutex.acquire()
            # Change newlines otherwise label widget will expand vertically.
            s.textError = text.replace("\n"," ") + s.textErrorPadding
            s.timeoutError = 1000 * 60 * 1    # 1 minute as milliseconds
        finally:
            st.gui.mutex.release()

    # INTERFACE
    def PrintError( s, text ):
        """ Print error message. """
        s.__PrintError( "ERROR: " + text )

    # INTERFACE
    def PrintWarning( s, text ):
        """ Print warning message. """
        s.__PrintError( "WARNING: " + text )

    # PRIVATE
    def __Update( s ):
        """ Called by a timer (in main thread) to update/render statusbar widget. """

        st.gui.ProcessPendingEvents()

        # Update label widget showing any error/warning.
        s.__UpdateError()

        # Call an optional registered callback which is expected
        # to call Statusbar.Print() to set the text.
        # Can be used for periodically updating the statusbar every timer.
        if s.updateCallback is not None:
            s.updateCallback( *s.updateCallbackArgs )

        # Was Print() called recently?
        if s.changed:
            s.changed = False
            s.__ChangeWidget( s.text )
        # Clear only once when timeout becomes negative.
        elif s.timeout < 0 and s.text:
            s.__Clear()
            s.__ChangeWidget( "" )
        # Age statusbar text.
        s.timeout -= s.milsec
        s.__RechargeTimer()

    # PRIVATE
    def __UpdateError( s ):
        """ Update label widget with error/warning text. """
        if s.textError:
            # Age the text.
            s.timeoutError -= s.milsec
            # Clear the text after time-out.
            if s.timeoutError < 0:
                s.textError = ""
            # Change label widget if text changed.
            if s.textError != s.textErrorPrev:
                s.textErrorPrev = s.textError[:]    # copy
                # Change text.
                s.labelError.config( text=s.textError )
                # Change color.
                if s.textError:
                    s.labelError.config( bg=s.labelErrorColorOn )
                else:
                    s.labelError.config( bg=s.labelErrorColorOff )

    # PRIVATE
    def __ChangeWidget( s, text ):
        """ Directly set text of statusbar widget. """
        s.text = text
        s.label.config( text=text )

    # PRIVATE
    def __RechargeTimer( s ):
        st.gui.ScheduleTimerFunc( s.milsec, s.__Update )

#-------------------------------------------------------------------------------
# Class ScrollFrame.
#-------------------------------------------------------------------------------

class ScrollFrame(ScrollFrameBase):
    """ Create a frame and scrollbars for the given widget. """

    def __init__( s, parentWidget, WidgetClass, labelText, **args ):
        s.enabled = False
        s.frame = TK.Frame( parentWidget )

        if labelText:
            s.labelFrame = TK.Frame( s.frame )
            s.label = TK.Label( s.labelFrame, text=labelText )
            s.labelOrigColor = s.label["bg"]
        else:
            s.label = None
            s.labelOrigColor = None
        s.widget     = WidgetClass( s.frame, args )
        s.widget.ScrollFrame = s  # store reference back to its ScrollFrame container
        s.xscrollbar = TK.Scrollbar( s.frame, orient=TK.HORIZONTAL )
        s.yscrollbar = TK.Scrollbar( s.frame, orient=TK.VERTICAL   )
        
        if labelText:
            s.labelFrame.pack( side=TK.TOP, expand=TK.NO, fill=TK.BOTH, anchor=TK.W )
            s.label.pack( side=TK.LEFT, fill=TK.Y )
        s.yscrollbar.pack( side=TK.RIGHT,  fill=TK.Y )
        s.xscrollbar.pack( side=TK.BOTTOM, fill=TK.X )
        s.widget.pack( expand=TK.YES, fill=TK.BOTH )

        s.widget.config( xscrollcommand=s.xscrollbar.set, yscrollcommand=s.yscrollbar.set )
        s.xscrollbar.config( command=s.widget.xview )
        s.yscrollbar.config( command=s.widget.yview )

    # INTERFACE
    def Enable( s, enable=True ):
        s.enabled = enable
        st.gui.EnableWidget( s.widget, enable )
        if s.label is not None:
            st.gui.EnableWidget( s.label, enable )

    # INTERFACE
    def SetLabelText( s, text ):
        """ Set the text of a label widget in a scroll-frame widget. """
        try:
            if s.label:
                s.label.config( text=text )
        except:
            PrintException()

    # INTERFACE
    def SetLabelColor( s, color ):
        """ Set the color of a label widget in a scroll-frame widget. """
        # Pass None to restore original color.
        try:
            if not color:
                color = s.labelOrigColor
            s.label.config( bg=color )
        except:
            PrintException()

    # INTERFACE
    def GetScrollbarPos( s ):
        """ Returns (x,y) tuple of horizontal and vertical scrollbars. """
        (x,dontcare) = s.xscrollbar.get()
        (y,dontcare) = s.yscrollbar.get()
        return (x,y)

    # INTERFACE
    def SetScrollbarPos( s, xy ):
        """ Set position of horizontal and vertical scrollbars (xy is a tuple). """
        if xy is not None:
            #x = xy[0]   # ignore horizonal position (unimportant)
            y = xy[1]
            if y is not None:  
                s.widget.yview_moveto( y )

#-------------------------------------------------------------------------------
# Class NotebookPage.
#-------------------------------------------------------------------------------

# Page names in Pmw notebooks:
# - Outside code gives a name to a page.
#   Given names are associated with a unique name (a stringified int count).
#   The unique name, not the given name, is passed down to Pmw when the page is created.
#   The reason is Pmw page names must be unique and can't be renamed
#   Notebook.__RenamePage() effectively renames a page can changing the given name
#   (re-associating the same unique name but with a different given name).
# - Page names are not the same as its tab text (tab text is for display)
#   as pypersrc could load files with the same names but different paths.
#   The Pmw NoteBook class also treats page names and tab text as different.

class NotebookPage:

    nextPageId = 0

    def __init__( s, pageName, tabText, notebookWidget, append=True ):
        # append arg is ignored (always appends)
        s.pageName       = pageName         # given page name
        s.notebookWidget = notebookWidget
        s.pageId         = s.__NewPageId()  # allocate a unique ID for Pmw page widget.
        s.pageWidget     = s.__NewPageWidget( s.pageId, tabText )
        s.SetTabText( tabText )

    # INTERFACE
    def GetPageName( s ):
        """ Return page's given name. """
        # This is the name of the page given by pypersrc (as opposed to page ID used by Pmw).
        return s.pageName
    
    # INTERFACE
    def GetPageId( s ):
        """ Return page ID. """
        # pageId is how the page is known to Pmw (pypersrc knows pages by given names).
        return s.pageId
    
    # INTERFACE
    def GetPageWidget( s ):
        """ Return page widget. """
        return s.pageWidget
        
    # INTERFACE
    def GetTabWidget( s ):
        """ Return tab widget (part of this page). """
        # tab() is a Pmw Notebook method.
        tabWidget = s.notebookWidget.tab( s.pageId )
        return tabWidget

    # INTERFACE
    def RenamePage( s, newPageName ):
        """ Change given name of page (page ID unchanged). """
        s.pageName = newPageName
        s.SetTabText( s.pageName )

    # INTERFACE
    def SetTabText( s, tabText ):
        """ Set text in tab widget of page. """
        # Pmw has a known bug where changing tab text doesn't change tab width
        # (not much of a problem for pypersrc).
        tabWidget = s.GetTabWidget()
        tabWidget["text"] = tabText      # implicit call to Pmw/Tkinter method
        
    # PRIVATE
    def __NewPageId( s ):
        """ Allocate a unique page ID for a new page. """
        NotebookPage.nextPageId += 1
        # Must stringify the page ID to pass to Pmw.
        return str(NotebookPage.nextPageId)

    # PRIVATE
    def __NewPageWidget( s, pageId, tabText ):
        # Return new page widget appended to Pmw notebook widget.
        # Pmw NoteBook.insert()/add() returns a page widget.
        return s.notebookWidget.add( pageId, tab_text=tabText )

#-------------------------------------------------------------------------------
# Class Notebook (Pmw).
#-------------------------------------------------------------------------------

# Pmw widgets inherit many methods from the Tkinter widgets.

class Notebook(NotebookBase):
    """ Generalized (but Tk-specific) class for a notebook widget. """
    # Functionality deferred to derived class:
    # - loading the contents of a page
    # - enable widgets contained in pages

    def __init__( s, parentWidget ):
        # Create a Pmw notebook widget.
        s.notebookWidget = Pmw.NoteBook( parentWidget, raisecommand=s.__HandlerPageSelected )
        s.notebookWidget.pack( fill="both", expand=1, padx=0, pady=0 )

        # Maps given page names to NotebookPage objects.
        s.pageDict = {}    # givenPageName : NotebookPage object

    # INTERFACE
    def Enable( s, enable ):
        """ This enables only the notebook and pages themselves.
            A derived class must enable whatever widgets contained in its pages. """
        # Enable notebook widget.
        st.gui.EnableWidget( s.notebookWidget, enable )

        # Enable all pages.
        #for pageWidget in s.GetListPageWidget():         # doesn't work as Pmw page widget
        #    st.gui.EnableWidget( pageWidget, enable )    # evidently doesn't implement "state"
        
        # Enable all tab widgets of each page.
        for tabWidget in s.GetListTabWidget():
            st.gui.EnableWidget( tabWidget, enable )

    # INTERFACE
    def IfEmpty( s ):
        """ Return true if no pages. """
        return len(s.pageDict) == 0

    # INTERFACE
    def AddPage( s, pageName, tabText, append=True ):
        """ Add a new notebook page.  Given page name must be unique.  tabText is for display.
            Returns widget so that caller can pack a child widget into it. """
        page = NotebookPage( pageName, tabText, s.notebookWidget, append )
        s.pageDict[pageName] = page        # associate NotebookPage object
        s.__UpdateClosePageButton()        # to enable button when there's at least one page
        return page.GetPageWidget()        # return Pmw page widget

    # INTERFACE
    def RemovePage( s, pageName ):
        """ Remove a notebook page. """
        page = s.pageDict[pageName]        # get NotebookPage object
        pageId = page.GetPageId()          # pageId is how page is known to Pmw
        s.notebookWidget.delete( pageId )  # call Pmw method
        del s.pageDict[pageName]
        s.__UpdateClosePageButton()        # to enable button when there's at least one page

    # PROTECTED
    def RenamePage( s, oldPageName, newPageName ):
        page = s.pageDict[oldPageName]
        s.page.RenamePage( newPageName )

    # INTERFACE
    def SelectPage( s, pageName ):
        """ Select a page. """
        page = s.pageDict[pageName]
        pageId = page.GetPageId()
        s.notebookWidget.selectpage( pageId )

    # INTERFACE
    def GetCurrentPageName( s ):
        """ Return name of current page or None if no pages exist. """
        try:
            pageId = s.notebookWidget.getcurselection()
            pageName = s.__PageId2PageName( pageId )
            return pageName
        except:
            return None

    # INTERFACE
    def CloseCurrentPage( s ):
        """ Close the current page. """
        pageName = s.GetCurrentPageName()
        if pageName:
            s.RemovePage( pageName )

    # PROTECTED
    def SetTabText( s, pageName, tabText ):
        """ Set text in tab widget of page. """
        page = s.pageDict[pageName]
        page.SetTabText( tabText )
        
    # PROTECTED
    def GetListPageWidget( s ):
        """ Return list of page widgets. """
        l = []
        for page in s.pageDict.values():
            l.append( page.GetPageWidget() )
        return l

    # PROTECTED
    def GetListTabWidget( s ):
        l = []
        """ Return list of tab widgets (part of page). """
        for page in s.pageDict.values():
            l.append( page.GetTabWidget() )
        return l
    
    # PRIVATE
    def __HandlerPageSelected( s, pageId ):
        # Reflect selected notepage page in titlebar.
        try:
            pageName = s.__PageId2PageName( pageId )
            st.gui.SetTitle( pageName )
        except:
            # An exception will normally occur while a page is being added because
            # this point is hit before pypersrc has mapped page IDs with page names.
            pass

    # PRIVATE
    def __PageId2PageName( s, pageId ):
        """ Convert page ID to a given page name. """
        # Converts a Pmw page name to a pypersrc page name.
        for page in s.pageDict.values():
            if pageId == page.GetPageId():
                return page.GetPageName()
        raise IndexError

    # PRIVATE
    def __UpdateClosePageButton( s ):
        # Gray-out "Close (Page)" button when the notebook is empty.
        if s.IfEmpty():
            st.gui.DisableWidget( st.gui.buttonClosePage )
            st.gui.SetTitle( "" ) # and clear titlebar
        else:
            st.gui.EnableWidget( st.gui.buttonClosePage )

#-------------------------------------------------------------------------------
# Class NotebookTextWid (Pmw).
#-------------------------------------------------------------------------------

class NotebookTextWid(Notebook,NotebookTextWidBase):

    def __init__( s, parentWidget ):
        Notebook.__init__( s, parentWidget )
        s.textWidDict = {}   # givenPageName : textWid

    # INTERFACE
    def Enable( s, enable ):
        """ This enables the notebook, pages, and the text widgets contained in the pages. """
        # Base method enables all widgets except the widgets contained in pages.
        Notebook.Enable( s, enable )
        # Enable all the text widgets contained in pages.
        textWidList = s.GetTextWidList()
        for textWid in textWidList:
            textWid.Enable( enable )
            
    # INTERFACE
    def LoadContents( s, contents, line=0, hiliteLine=False, currentPage=False, selectPage=True ):
        """ Load notebook page with contents.  Updates title bar. """
        # currentPage can be True even if notebook is empty
        # in which case the first notebook page will be added now.

        pageName = contents.fname
        tabText  = Basename(contents.fname)               # abbrev fname

        # Determine which text widget to load.
        # Replace current page?
        if currentPage and s.GetCurrentTextWid():         # unless notebook has no page yet
            oldPageName = s.CurrentPageName()
            newPageName = pageName
            s.RenamePage( oldPageName, newPageName )
            textWid = s.GetCurrentTextWid()
        # Is a notebook page holding it?
        elif s.textWidDict.has_key( pageName ):
            textWid = s.textWidDict[pageName]
        else: # load in new page
            s.AddPage( pageName, tabText )                # creates notebook page and text widget
            textWid = s.textWidDict[pageName]
            
        # Load text widget.
        textWid.LoadContents( contents=contents, line=line, hiliteLine=hiliteLine )
        if selectPage:
            s.SelectPage( pageName )
            st.gui.SetTitle( pageName )

    # INTERFACE
    def AddPage( s, pageName, tabText, append=True ):
        """ Add a new notebook page.  Given page name must be unique.  tabText is for display.
            Returns widget so that caller can pack a child widget into it. """
        # This derived method additionally manages textWidDict.
        pageWidget = Notebook.AddPage( s, pageName, tabText ) # create page widget
        textWid = TextWidEditor( pageWidget ) # create text widget (inside page)
        s.textWidDict[pageName] = textWid     # remember new page in dict

    # INTERFACE
    def RemovePage( s, pageName ):
        """ Remove a notebook page. """
        # This derived method additionally manages textWidDict.
        Notebook.RemovePage( s, pageName )
        del s.textWidDict[pageName]           # forget destroyed text widget

    # INTERFACE
    def GetCurrentTextWid( s ):
        """ Return current text widget or None if no pages exist. """
        try:
            name = s.GetCurrentPageName()
            return s.textWidDict[name]
        except:
            return None    # else notebook is empty
        
    # INTERFACE
    def GetTextWidList( s ):
        """ Return list of text widgets in notebook. """
        textWidList = s.textWidDict.values()
        return textWidList
    
#-------------------------------------------------------------------------------
# Class NotebookSrcFile (Pmw).
#-------------------------------------------------------------------------------

class NotebookSrcFile(NotebookTextWid,NotebookSrcFileBase):
    """ Notebook class specialized for src files. """

    def __init__( s, parentWidget ):
        NotebookTextWid.__init__( s, parentWidget )

    # INTERFACE
    def LoadSrcFile( s, fname, line=0, hiliteLine=False, currentPage=False, selectPage=True ):
        """ Load src file into notebook page. """
        # Subtle: If the same fname is passed, different contents could be loaded
        # if CacheSrcFile was flushed or the parser is now in a different mode (filtering changed).
        try:
            st.gui.statusbar.Print( "Loading " + fname + "...", now=True, timeout=0 )
            st.gui.Enable( False )    # producing and loading contents may take a while
            try:
                # CacheSrcFile will read src file if it isn't cached.
                contents = st.gui.cacheSrcFile.Get( fname, st.parser.GetModes() )
                # Load notebook page with src file.
                s.LoadContents( contents, line, hiliteLine, currentPage, selectPage )
            except IOError:
                PrintWarning( "could not read source file '" + fname + "'" )
                return
        finally:
            st.gui.Enable( True )
        
    # INTERFACE
    def Reload( s ):
        """ Reload src files (shown in every notebook page). """
        textWidList = s.GetTextWidList()
        for textWid in textWidList:    # for each text widget in notebook
            # Subtle: Although the same fname is fed-back, different contents could be loaded
            # if CacheSrcFile was flushed or the parser is now in a different mode (filtering changed).
            fname = textWid.contents.fname
            s.LoadSrcFile( fname=fname, selectPage=False )
# $Id: GuiTkText.py,v 1.25 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# Notes:    - Tk Text widgets are inserted with "tags" (not to be confused with src code tags).
#             tktag is written to disambiguate.
#           - Tk will combine similar tags.
#           - Disabling a Tk Text widget isn't the desired way to turn in read-only.
#             When disabled, it will ignore keyboard input (read-only), but
#             disabling causes methods, such as insert() and delete(), to become NOPs.
#           - Read-only Tk Text widgets are simulated using the trick of installing
#             a keyboard handler that returns "break" to stop propagation of events.
#             This trick allows the widget effectively read-only but still be enabled
#             (state=normal), so that insert()/delete() will work.
#           - Historical note: Text widgets used to be disabled to turn them read-only.
#             But a disabled text widget causes several problems: methods such as insert()
#             become NOPs (requiring the kludge of temporarily enabling the widget),
#             scrollbar widgets ignore Page Up/Down keys, and can't receive the focus.
#           - See GuiText.py.
# ==============================================================================

#-------------------------------------------------------------------------------
# Class TextWid.
#-------------------------------------------------------------------------------

class TextWid(ScrollFrame,TextWidBase):
    """ Text widget contained in a scroll-frame widget. """

    def __init__( s, frame ): # frame (not fname)

        # Constructor args are distributed to multiple-inherited base classes.
        TextWidBase.__init__( s )
        ScrollFrame.__init__( s, frame, TK.Text, " ", wrap=TK.NONE )
        s.frame.pack( side=TK.TOP, expand=TK.YES, fill=TK.BOTH )
        s.__InstallHandlers()
        s.widget.TextWid = s            # let widget have a reference back to its TextWid class
        s.enabled        = False

        # Fonts.
        if st.gui.fontName and st.gui.fontSize:
            font = tkFont.Font( family=st.gui.fontName, size=st.gui.fontSize )
        elif st.gui.fontName:
            font = tkFont.Font( family=st.gui.fontName )
        elif st.gui.fontSize:
            font = tkFont.Font( size=st.gui.fontSize )
        else:
            font = None
        if font:
            s.widget.tag_config( st.gui.TKTAG_DEFAULT, font=font, foreground="black"  )
            s.widget.tag_config( st.gui.TKTAG_TEXT,    font=font, foreground="black"  )
            s.widget.tag_config( st.gui.TKTAG_NUM,     font=font, foreground="black"  )
            s.widget.tag_config( st.gui.TKTAG_SYMBOL,  font=font, foreground="darkred" )
            s.widget.tag_config( st.gui.TKTAG_WORD,    font=font, foreground="black"  )
            s.widget.tag_config( st.gui.TKTAG_KEYWORD, font=font, foreground="darkblue" )
            s.widget.tag_config( st.gui.TKTAG_TYPE,    font=font, foreground="purple" )
            s.widget.tag_config( st.gui.TKTAG_LOUD,    font=font, foreground="gold2" )
            s.widget.tag_config( st.gui.TKTAG_QUOTED,  font=font, foreground="#804040" )
            s.widget.tag_config( st.gui.TKTAG_COMMENT, font=font, foreground="darkgreen" )
            s.widget.tag_config( st.gui.TKTAG_TAGDEF,  font=font, foreground="darkcyan" )
            s.widget.tag_config( st.gui.TKTAG_TAGREF,  font=font, foreground="darkcyan", underline=1 )
            s.widget.tag_config( st.gui.TKTAG_TAGREFV, font=font, foreground="#a000a0", underline=1 )
        else:
            s.widget.tag_config( st.gui.TKTAG_DEFAULT,            foreground="black"  )
            s.widget.tag_config( st.gui.TKTAG_TEXT,               foreground="black"  )
            s.widget.tag_config( st.gui.TKTAG_NUM,                foreground="black"  )
            s.widget.tag_config( st.gui.TKTAG_SYMBOL,             foreground="darkred" )
            s.widget.tag_config( st.gui.TKTAG_WORD,               foreground="black"  )
            s.widget.tag_config( st.gui.TKTAG_KEYWORD,            foreground="darkblue" )
            s.widget.tag_config( st.gui.TKTAG_TYPE,               foreground="purple" )
            s.widget.tag_config( st.gui.TKTAG_LOUD,               foreground="gold2" )
            s.widget.tag_config( st.gui.TKTAG_QUOTED,             foreground="#804040" )
            s.widget.tag_config( st.gui.TKTAG_COMMENT,            foreground="darkgreen" )
            s.widget.tag_config( st.gui.TKTAG_TAGDEF,             foreground="darkcyan" )
            s.widget.tag_config( st.gui.TKTAG_TAGREF,             foreground="darkcyan", underline=1 )
            s.widget.tag_config( st.gui.TKTAG_TAGREFV,            foreground="#a000a0", underline=1 )

    # PRIVATE
    def __InstallHandlers( s ):
        """ Install handlers for TextWid. """
        s.widget.tag_bind( st.gui.TKTAG_TAGREF,  "<Button-1>", s.__HandlerHyperlink1 )
        s.widget.tag_bind( st.gui.TKTAG_TAGREF,  "<Button-2>", s.__HandlerHyperlink2 )
        s.widget.tag_bind( st.gui.TKTAG_TAGREF,  "<Button-3>", s.__HandlerHyperlink3 )
        s.widget.tag_bind( st.gui.TKTAG_TAGREFV, "<Button-1>", s.__HandlerHyperlinkVisited1 )
        s.widget.tag_bind( st.gui.TKTAG_TAGREFV, "<Button-2>", s.__HandlerHyperlinkVisited2 )
        s.widget.tag_bind( st.gui.TKTAG_TAGREFV, "<Button-3>", s.__HandlerHyperlinkVisited3 )
        s.widget.bind( "<FocusIn>",  s.__HandlerFocusIn )
        s.widget.bind( "<FocusOut>", s.__HandlerFocusOut )
        s.widget.bind( "<Key>", s.__HandlerKey )

    # INTERFACE
    def Enable( s, enable=True ):
        """ Enable TextWid. """
        ScrollFrame.Enable( s, enable )
        st.gui.EnableWidget( s.widget, enable )

    # INTERFACE
    def LoadContents( s, contents, line=0, hiliteLine=False, scrollbarPos=None ):
        """ Load text widget with a Contents object. """
        # This wrapper temporarily disables the GUI while loading Tk text widget which can be slow.
        st.gui.statusbar.Print( "Loading " + contents.fname + "...", now=True, timeout=0 )
        st.gui.Enable( False )      # disable entire GUI
        s.Enable()                  # but enable text widget or it can't be changed
        s.__LoadContents( contents, line, hiliteLine, scrollbarPos )
        st.gui.Enable( True )       # enable GUI (and text widget)
        st.gui.statusbar.Clear()

    # PRIVATE
    def __LoadContents( s, contents, line=0, hiliteLine=False, scrollbarPos=None ):
        # PRECOND: Text widget is enabled (otherwise insert()/delete() would be NOPs).
        # - contents : To implicitly force a reload, pass a different Contents object.
        # - line : Lines are counted from 1.  If line > 0, text widget is scrolled to line.
        # - scrollbarPos : (x,y) tuple (or None) of positions of horizontal,vertical scrollbars

        s.contents = contents

        s.widget.delete( s.__Idx(1), TK.END )    # this can be slow, widget state must be enabled prior

        # Load text widget from tokens using Tk's tag mechanism.
        # Tk tag lists are cached as a member of the Contents class,
        # but initially the member doesn't exist.
        try:
            # Executes if before Tk tag list was computed.
            tktagList = s.contents.tktagList    # possible exception
            st.log.Write( "TextWid: loading " + s.contents.fname + " (have Tk tags)" )
        except:
            # Compute Tk tag list, store/cache it as a class member of Contents.
            tktagList = s.Contents2TktagList( s.contents )
            s.contents.tktagList = tktagList
            st.log.Write( "TextWid: loading " + s.contents.fname + " (built Tk tags)" )

        # Insert Tk tags into text widget.
        for tktag in tktagList:
            text   = tktag[0]
            tktag2 = tktag[1]
            s.widget.insert( TK.END, text, (tktag2) )   # widget state must be enabled prior

        # Change position of scrollbars.
        if scrollbarPos is not None:
            s.SetScrollbarPos( scrollbarPos )

        s.SetLabelText( s.contents.fname )
        s.GotoLine( line, hiliteLine )   # NOP if line=0
        
    # PROTECTED/ABSTRACT
    def Contents2TktagList( s, contents ):
        # If a source file is being loaded, all token tuples are one type:
        # (eToken*,text)
        # Otherwise, there are two types of token tuples.
        # The phony tag tuple is used to show hyperlinks in the results
        # of user commands such as find text.
        # (eTokenTag,(PhonyTag))
        # (eTokenText,text)
        raise Oops( "abstract method, derived TextWid must define this!" )

    # INTERFACE
    def GotoLine( s, line, hiliteLine=False ):
        # Ignore if line=0.
        if line >= 1:
            # Scroll widget to line.
            s.widget.see( s.__Idx(line) )
            s.widget.mark_set( TK.INSERT, s.__Idx(line) )

            # Hilite line.
            if hiliteLine:
                try:
                    s.widget.tag_delete( "PYMATCH" )    # forget previous matching lines
                except:
                    PrintException()
                s.widget.tag_add( "PYMATCH", s.__Idx(line), s.__Idx(line,999) )    # column 0...999
                s.widget.tag_config( "PYMATCH", background="cyan" )
        
            # Prevent an unintentional text selection during scrolling
            # (can happen if user has mouse button down while this is executing).
            s.widget.tag_remove( TK.SEL, s.__Idx(1), TK.END )

    # PRIVATE
    def __Idx( s, line, col=0 ):
        # Lines are counted from 0.
        # Tk-specific: Return an Tk Text widget index for location at line and column.
        # The return value can be passed to Tk Text.insert() eg.
        return "%d.%d" % (line,col)

    # PROTECTED
    def Clear( s, tokenList=[], fname=None ):
        """ Clear text widget. """
        # fname can be used as the title (for display).
        contents = ContentsNone( tokenList=tokenList, fname=fname )
        s.LoadContents( contents=contents )

    # PRIVATE
    def __SetTktagList( s, key, val ):
        i = 0
        for tktag in s.contents.tktagList:
            if tktag[0] == key:
                s.contents.tktagList[i] = (key,val)
            i += 1

    # PRIVATE
    def __HandlerFocusIn( s, event ):
        # If text widget loaded, change label to a brighter color.
        textWid = event.widget.TextWid
        if not textWid.IfEmpty():
            textWid.SetLabelColor( st.gui.colorFocusIn )

    # PRIVATE
    def __HandlerFocusOut( s, event ):
        # Focus has left, change label back to default color.
        textWid = event.widget.TextWid
        textWid.SetLabelColor( None )

    # PRIVATE
    def __HandlerKey( s, event ):
        # Trick to simulate a read-only-but-enabled Tk text widget.
        # Discard keys that would modify the text widget by returning "break".
        # But allow movement keys events to propagate scrollbar widget.
        allowed = event.keysym in [ "Escape", "Next", "Prior", \
                                    "Up", "Down", "Left", "Right", \
                                    "Home", "End", "Control_L", "Control_R", \
                                    "F1", "F2", "F3", "F4", "F5", "F6", \
                                    "F7", "F8", "F9", "F10", "F11", "F12" ]
        allowed = allowed or event.char in [ "\x17" ]    # Ctrl-W = 0x17
        # If not a movement key, assume the key would modify the text widget.
        if not allowed and not s.contents.editable:
            st.gui.statusbar.Print( "Text widget is read-only (pypersrc is not an editor)." )
            return "break"    # stop event
        else:
            return            # let event propagate

    # PRIVATE
    def __HandlerHyperlink1( s, event ):
        s.__HandlerHyperlinkSub( event, st.gui.TKTAG_TAGREF, 1 ) # button 1

    # PRIVATE
    def __HandlerHyperlink2( s, event ):
        s.__HandlerHyperlinkSub( event, st.gui.TKTAG_TAGREF, 2 ) # button 2

    # PRIVATE
    def __HandlerHyperlink3( s, event ):
        s.__HandlerHyperlinkSub( event, st.gui.TKTAG_TAGREF, 3 ) # button 3

    # PRIVATE
    def __HandlerHyperlinkVisited1( s, event ):
        s.__HandlerHyperlinkSub( event, st.gui.TKTAG_TAGREFV, 1 ) # button 1

    # PRIVATE
    def __HandlerHyperlinkVisited2( s, event ):
        s.__HandlerHyperlinkSub( event, st.gui.TKTAG_TAGREFV, 2 ) # button 2

    # PRIVATE
    def __HandlerHyperlinkVisited3( s, event ):
        s.__HandlerHyperlinkSub( event, st.gui.TKTAG_TAGREFV, 3 ) # button 3

    # PRIVATE
    def __HandlerHyperlinkSub( s, event, tktag, button ):
        """ Handle clicking hyperlink in a text widget. """
        # See GuiText.py for mouse button behavior.
        # tkag is TKTAG_TAGREF or TKTAG_TAGREFV.
        # .................................................
        # Button 1 means to load into a new notebook page.
        # Button 2 means to load the current notebook page.
        widget = event.widget
        currentPage = (button != 1)
        try:
            textWid = widget.TextWid
            # This might be slow on older TK versions (eg Redhat 8).
            st.gui.statusbar.Print( "Locating hyperlink..." )
            # Find the nearest Tk tag to where the user clicked.
            pos = '@%d,%d' % (event.x+1, event.y)
            (pos1,pos2) = widget.tag_prevrange( tktag,  pos )
            tag = widget.get( pos1, pos2 ) # this a Tk tag but may also be a source code tag (confusing)
            if isinstance(textWid.contents,ContentsResults):
                # The user clicked a hyperlink in a results text widget.
                # Search the results tokenList for what was clicked.
                # token[1] holds a phony tag (special case).
                for token in textWid.contents.tokenList:
                    if st.tags.IfPhonyTagInToken( token ):
                        phonyTag = token[1]
                        if tag == phonyTag[TAG_NAME]:
                            # Immediately change the color of the hyperlink
                            # to indicate that it was visited (dark purple).
                            widget.delete( pos1, pos2 )
                            widget.insert( pos1, tag, (st.gui.TKTAG_TAGREFV) )

                            # -- This affects cached ContentsResults --
                            # Update Tk tag list.
                            s.__SetTktagList( tag, st.gui.TKTAG_TAGREFV )

                            # -- This affects non-cached ContentsResults (eg ContentsHistory) --
                            # Remember that this hyperlink was visited by the user.
                            # This causes the TextWid.LoadText() to color the hyperlink
                            # as dark purple if the results are reloaded from the cache.
                            st.visited[tag] = True

                            fname = phonyTag[TAG_FILE]
                            line  = phonyTag[TAG_LINE]
                            if isinstance(textWid.contents,ContentsHistory):
                                # Hyperlink clicked in the history of command results.
                                st.gui.resultsHistory.Load( st.gui.GetResultsTextWid(), fname )
                            else:
                                # Load into current or different notebook page (text widget).
                                st.gui.notebookSrcFile.LoadSrcFile( fname, line=line, hiliteLine=True, currentPage=currentPage )
                            break
            else:
                st.gui.TagClick( clickedTagName=tag, clickedFname=textWid.contents.fname )
        except:
            PrintException()
        st.gui.statusbar.Clear()
                
#-------------------------------------------------------------------------------
# Class TextWidEditor.
#-------------------------------------------------------------------------------

# TextWidEditor doesn't mean a text widget that is an editor.
# Rather, it has buttons for running an external editor (emacs/vi).

class TextWidEditor(TextWid):
    """ Specialized text widget with buttons for invoking external editor (emacs/vi). """

    def __init__( s, frame ): # frame (not fname)
        TextWid.__init__( s, frame )
        # Create editor buttons.
        if st.editorXterm:
            s.buttonEditorXterm = TK.Button( s.labelFrame, text=Basename(st.editorXterm), command=s.__HandlerButtonEditorXterm )
            s.buttonEditorXterm.pack( side=TK.RIGHT, anchor=TK.E )
            st.gui.DisableWidget( s.buttonEditorXterm )
        if st.editorGui:
            s.buttonEditorGui = TK.Button( s.labelFrame, text=Basename(st.editorGui), command=s.__HandlerButtonEditorGui )
            s.buttonEditorGui.pack( side=TK.RIGHT, anchor=TK.E )
            st.gui.DisableWidget( s.buttonEditorGui )

    # INTERFACE
    def Enable( s, enable=True ):
        """ Enable TextWid. """
        TextWid.Enable( s, enable )
        if st.editorGui:
            st.gui.EnableWidget( s.buttonEditorGui, enable )
        if st.editorXterm:
            st.gui.EnableWidget( s.buttonEditorXterm, enable )

    # INTERFACE
    def LoadContents( s, contents, line=0, hiliteLine=False, scrollbarPos=None ):
        """ Load text widget with a Contents object. """
        # Jump to another line in the same contents?
        if (contents is s.contents) and (not scrollbarPos):
            s.GotoLine( line, hiliteLine )
        else:
            TextWid.LoadContents( s, contents, line, hiliteLine, scrollbarPos )

    # PROTECTED
    def Contents2TktagList( s, contents ):
        """ Convert a Contents object to a list of Tk text tags. """
        # (see base method for comments)
        tktagList  = []
        line       = 1
        fname      = contents.fname
        tokenList  = contents.tokenList
        for token in tokenList:
            eToken = token[0]    # token enum
            text   = token[1]
            # Underline a reference to source code tags
            # to make a HTML-like hyperlink but not for
            # a source tag on the line where it's defined.
            if eToken == eTokenTag:
                if st.tags.IfTagRef( text, fname, line ):
                    tktag = st.gui.TKTAG_TAGREF
                else:
                    tktag = st.gui.TKTAG_TAGDEF
            else:
                tktag = st.gui.tktagDict[eToken]
            tktagList.append( (text,tktag) )
            line += text.count( "\n" )
        return tktagList
    
    # INTERFACE
    def Refresh( s, change, addedList, modifiedList, removedList ):
        # Pass change=False if no source files/dirs really changed,
        # in which case only the log will be refreshed if it's showing.

        # Return if nothing needs refreshing.
        if not change:
            return

        def RefreshReloadTextwid( textwid ):
            # PRECOND: The src file was flushed out of the cache.
            # But still ask the cache for the src file in order to re-cache it.
            fname = textwid.contents.fname
            contents = st.gui.cacheSrcFile.Get( fname, st.parser.GetModes() )
            textwid.LoadContents( contents=contents )    # assigns s.contents

        # Was the src file (being shown) modified?
        if s.contents.fname in modifiedList:
            st.log.Write( "TextWid: refreshing " + s.contents.fname )
            RefreshReloadTextwid( s )
            return
            
        # Was the src file (being shown) deleted?
        if s.contents.fname in removedList:
            # Set .fnamedSaved so that the file can be reloaded
            # by refresh if later reinstated by the user.
            tokenList = [ (eTokenText,"("+s.contents.fname+" no longer exists)") ]
            s.Clear( tokenList=tokenList, fname=s.contents.fname )
            s.contents.missing    = True     # remember that it was removed externally
            s.contents.fnameSaved = s.contents.fname
            return
            
        # Consequently, was the source file deleted but later reinstated?
        if s.contents.fname in addedList and s.contents.missing:
            st.log.Write( "TextWid: refreshing reinstated '" + s.contents.fnameSaved + "'" )
            RefreshReloadTextwid( s )
            s.contents.missing    = False
            s.contents.fnameSaved = None
            return
            
        # No reason to refresh this text widget.
        return

    # PRIVATE
    def __HandlerButtonEditorGui( s ):
        """ Handler for starting a native GUI editor (eg xemacs,gedit). """
        if st.editorGui:
            prog      = st.editorGui
            argsTuple = (st.editorGui, s.contents.fname)
            s.__HandlerButtonEditor( prog, st.editorGui, argsTuple )

    # PRIVATE
    def __HandlerButtonEditorXterm( s ):
        """ Handler for starting an editor that must run in an xterm (eg vi/vim). """
        if st.editorXterm:
            prog      = "xterm"
            argsTuple = ("xterm", "-e", st.editorXterm, s.contents.fname)
            s.__HandlerButtonEditor( prog, st.editorXterm, argsTuple )

    # PRIVATE
    def __HandlerButtonEditor( s, prog, editor, argTuple ):
        # prog:   program to run (could be the editor itself or xterm)
        # editor: name of editor
        try:
            if prog:
                st.log.Write( "running editor " + editor + " on " + s.contents.fname )
                RunProgram( prog, True, argTuple )
            else:
                PrintWarning( "Gui: editor not specified" )
        except:
            PrintException()
    
#-------------------------------------------------------------------------------
# Class TextWidResults.
#-------------------------------------------------------------------------------

class TextWidResults(TextWid):
    """ Specialized text widget for holding results. """

    def __init__( s, frame ): # frame (not fname)
        TextWid.__init__( s, frame )
        # To keep log refreshed if it is being shown.
        s.logLen        = 0
        s.milsecIdleLog = 2000
        st.gui.ScheduleTimerFunc( s.milsecIdleLog, s.__IdleLog )
        
    # INTERFACE
    def LoadContents( s, contents, line=0, hiliteLine=False, scrollbarPos=None ):
        """ Load text widget with a Contents object. """
        # Store outgoing results in history.
        st.gui.resultsHistory.Save( s )
        TextWid.LoadContents( s, contents, line, hiliteLine, scrollbarPos )

    # PROTECTED
    def Contents2TktagList( s, contents ):
        """ Convert a Contents object to a list of Tk text tags. """
        # (see base method for comments)
        tktagList  = []
        line       = 1
        fname      = contents.fname
        tokenList  = contents.tokenList
        for token in tokenList:
            # Is this a (eTokenTag,(PhonyTag)) ?
            if st.tags.IfPhonyTagInToken( token ):
                phonyTag = token[1]
                text = phonyTag[TAG_NAME]
                # Display visited and not-yet-visited hyperlinks
                # in different colors (like a web browser would do).
                # These two Tk tags are associated with different colors.
                if st.visited.has_key( text ):
                    tktag = st.gui.TKTAG_TAGREFV
                else:
                    tktag = st.gui.TKTAG_TAGREF
                tktagList.append( (text,tktag) )
                # Insert different tag to prevent Tk from combining like tags.
                # An empty line between matches makes clicking easier for the user.
                tktagList.append( ("\n\n",st.gui.TKTAG_DEFAULT) )
            else:
                # No, this is a usual (eToken*,text).
                text = token[1]
                tktag = st.gui.TKTAG_DEFAULT
                tktagList.append( (text,tktag) )
        return tktagList
    
    # INTERFACE
    def Refresh( s, change, addedList, modifiedList, removedList ):
        # Pass change=False if no source files/dirs really changed,
        # in which case only the log will be refreshed if it's showing.

        # Is this text widget showing results?
        # If so, rerun its command if it is affected by the source file change.
        # But always refresh the log if it's being shown.
        if change and (not isinstance(s.contents,ContentsNone)) and s.contents.command: # order-dependent
            st.log.Write( "refresh: redoing command '" + s.contents.command + "'" )
            s.contents.UpdateRefreshCnt()
            st.command.Do( s.contents.command )
        else:
            # Refresh unnecessary.
            return

    # INTERFACE (not part of base class)
    def LoadLog( s ):
        """ Load log into text widget. """
        log = st.log.Read()                  # returns a copy (not a reference)
        s.logLen = len( log )                # used by TextWid.__IdleLog()
        tokenList = [ (eTokenText,log) ]
        fname = "[results: log]"
        contents = ContentsLog( tokenList=tokenList, fname=fname )
        line = log.count( "\n" ) - 1         # find last line to show end of log
        s.LoadContents( contents=contents, line=line )

    # PRIVATE
    def __IdleLog( s ):
        # If log is loaded in a text widget, automatically refresh it.
        if isinstance(s.contents,ContentsLog):
            log    = st.log.Read()    # returns a copy (not a reference)
            logLen = len(log)
            if s.logLen < logLen:
                
                # Append new log lines directly to Tk text widget.
                text = log[s.logLen:]     # new log lines
                s.widget.insert( TK.END, text, (st.gui.TKTAG_DEFAULT) )
                s.logLen = logLen         # update log length
                s.widget.see( TK.END )    # scroll down
                
        # Again.
        st.gui.ScheduleTimerFunc( s.milsecIdleLog, s.__IdleLog )
# $Id: State.py,v 1.32 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Class State.
#-------------------------------------------------------------------------------

class State:

    def __init__( s, argList ):
        """ Global state. """

        s.log             = Log()    # first of all
        s.mainThreadId    = thread.get_ident()

        s.ProcessArgsPre( argList )
        
        # ---- PERSISTENT STATE (TODO) ------------------------------
        s.visited         = {}    # visited hyperlinks
        # ---- PERSISTENT STATE (TODO) ------------------------------
        
        # s.langList is the default list of programming language files to gather.
        # The -l arg can append to it.
        s.version         = ""    # assigned by C++ module passing -setversion internal arg
        s.exiting         = False   # set when GUI destroy event (quit/program exit) is dequeued
        s.error           = False   # set by PrintError()
        s.warning         = False   # set by PrintWarning()
        s.refreshCnt      = 0       # to recompute stale results
        s.langList        = [ "c", "c++", "python" ]
        s.verbose         = False
        s.ignoreDirList   = [ "CVS", "RCS", "bak", "out" ]
        s.tabWidth        = 4
        s.escKey          = False
        # Two kinds of editors are supported: console and GUI (eg vi and xemacs).
        s.editorGui       = "emacs"
        s.editorXterm     = "vi"
        # Prefer vim over vi.
        for pname in [ "/usr/bin/vim", "/usr/local/bin/vim" ]:
            if os.access( pname, os.F_OK ):
                s.editorXterm = "vim"
                break
        if IfCygwin():
            # Must be an absolute pathname (limited by Python/Windows spawn()).
            s.editorGui   = "c:\\program files\\windows nt\\accessories\\wordpad.exe"
            if not os.path.isfile( s.editorGui ):
                s.log.Write( "Startup: can't find editor '" + s.editorGui + "'" )
                s.editorGui  = None
            s.editorXterm = None
        elif IfDarwin():
            s.editorGui   = "xemacs"
            s.editorXterm = "vi"
        
        s.ProcessArgsPost( argList )

    # INTERFACE
    def ProcessArgsPre( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True
    
    # INTERFACE
    def ProcessArgsPost( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        i = 0
        while i < len(argList):
            arg = argList[i]
            if arg is None:
                i += 1
            elif arg == "-v":
                s.verbose = True
                s.log.SetLogLevel( Log.VERBOSE )
                argList[i] = None
                i += 1
            else:
                i += 1
        return True

    # INTERFACE
    def IfMainThread( s ):
        """ Return True if this is the main thread. """
        return s.mainThreadId == thread.get_ident()

    # INTERFACE
    def GetSplashText( s ):
        return "pypersrc " + s.version + "  (c) Jim E. Brooks 2005-2006  http://www.jimbrooks.org/hypersrc/  GNU GPL no warranty"
# $Id: Startup.py,v 1.43 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Startup.
#-------------------------------------------------------------------------------

# This appears at the top of the catenated Python script.
def Startup( args ):
    """ Startup function. """
    try:
        # Convention for arg processing is to assign None to elements
        # in argList after processing the corresponding args.
        argList = list(args)[1:] # skip argv[0]
        
        global st
        st = State( argList )

        if not ProcessArgs( argList ):
            return

        st.log.SetSplashLine( "pypersrc log (" + st.version + ") \n" )

        # Create the major objects of pypersrc.
        st.gather  = Gather(  argList )
        st.tags    = Tags(    argList )
        st.lex     = Lex(     argList )
        st.parser  = Parse(   argList )
        st.command = Command( argList )
        st.gui     = Gui(     argList )
        st.threads = Threads( argList )

        # Gather src files, tag their symbols.
        st.threads.Spawn( ThreadGatherTag, argList )
        
        # Enter GUI event loop.
        st.gui.Enable()
        st.command.Do( Command.INTERNAL_COMMAND_SPLASH )
        st.gui.statusbar.Print( st.GetSplashText() )
        st.gui.EventLoop()

        # Return to C++ code (main.cc).
        return
    
    except:
        PrintException()

def ProcessArgs( argList ):
    """ The first pass of cmd-line arg processing. """
    """ Returns False if program should return/exit. """
    try:
        i = 0
        while i < len(argList):
            arg = argList[i]
            if arg is None:
                i += 1
            elif arg == "-setversion":
                # Internal arg passed by main.cc.
                st.version = argList[i+1]
                argList[i]   = None
                argList[i+1] = None
                i += 2
            elif arg == "-l":
                # Programming language arg.
                # Design note: -l <lang> must be followed by exactly one sub-arg.
                # Allowing multiple sub-args to follow would be convenient,
                # but that would open the door to misinterpreting file/dir args.
                # Eg "pypersrc -l c++ ." means to browse the CWD ('.' isn't a language arg).
                argList[i] = None
                i += 1
                st.langList = []
                if i >= len(argList):
                    PrintWarning( "-l missing sub-arg" )
                elif not argList[i] in progLangList:
                    PrintWarning( "'" + argList[i] + "' is an unsupported programming language" )
                else:
                    st.langList.append( argList[i] )
                    argList[i] = None
                    i += 1
            elif arg in [ "-h", "-help", "--help" ]:
                print "pypersrc by Jim Brooks  http://www.jimbrooks.org/hypersrc/"
                print "pypersrc is a source code browser written in Python."
                print "License: GNU GPL"
                print "Command-line arguments:"
                print "file(s)     source files to browse"
                print "dir(s)      pypersrc will recurse into directories"
                print "-x glob     exclude files that match the fileglob pattern"
                print "-l lang     programming language source files to browse, any of:"
                print "            c c++ java python"
                print "            example: -l c++ -l java  (default: c c++ python)"
                print "-geom       window size"
                print "            example: -geom 800x600"
                print "-ed-gui     GUI editor to use (default: emacs)"
                print "            example: -ed-gui xemacs"
                print "-ed-xterm   console editor to use (default: vi/vim)"
                print "            example: -ed-xterm joe"
                print "-hilite-lo[hi] less[more] syntax hiliting (affects speed of loading files)"
                return False
            elif arg in [ "-ed-gui", "-editor-gui" ]:
                st.editorGui = argList[i+1]
                argList[i]   = None
                argList[i+1] = None
                i += 2
            elif arg in [ "-ed-xterm", "-editor-xterm", "-ed-console" ]:
                st.editorXterm = argList[i+1]
                argList[i]   = None
                argList[i+1] = None
                i += 2
            else:
                i += 1
    except:
        PrintException()
    
    return True
